/* Lex definitions */

%{
#include <y.tab.h>

unsigned int len;	/* so we can share packet size content with parser */

unsigned int bytesread=0;	/* total bytes read by parser */

%}

ANY [\x00-\xFF]

/* LEAD ("\x00")*  The concept here is that the leading number of 
zero' can be any number... and only the last 4 bytes should
be used to match the start code, at present there seems to be
a bug if we use these. So don't use this unless it can be determine
when this is appropriate  */

MINBITS "\x00\x00\x01"


/* PREFIX {LEAD}{MINBITS} */
PREFIX {MINBITS} 

%s start_pack
%s expect_packets
%s opt_ids
%s expect_stream_id
%s expect_pkt_len
%s expect_scale
%x data

%%
{PREFIX}"\xBA"{ANY}{8} { /* PACK */
		BEGIN(start_pack);

		bytesread += yyleng;

		/* Its a pack so suck in the data for the pack */
		/* there is a fixed amount of data here, 8bytes */
		return(PACK_START_CODE);
		}

{PREFIX}"\xB9"      { /* ISO_11172_END_CODE */
		BEGIN(INITIAL);
		bytesread += yyleng;
		return(ISO_11172_END_CODE);
		}

<start_pack>{PREFIX}"\xBB"{ANY}{8} {
		/* SYSTEM_HEADER */
		/* headers are followed by 8 bytes of data then */
		/* it can be followed by an optional number of */
		/* 3 bytes identifying  the stream, scale bounds etc */
		/* this is then terminated by either a packet */
		/* another pack or iso_11172_end_code */
		BEGIN(opt_ids);
		bytesread += yyleng;
		return(SYSTEM_HEADER_START_CODE);
		}

<opt_ids>[\x80-\xFF]{ANY}{2} {
		/* system header contains optional stream ID's */
		/* with bound scale / size bound info */
		bytesread += yyleng;
		return(STREAM_ID_PLUS);
		}

<opt_ids>[^\x80-\xFF] {
		/* no stream ID's in this one */
		/* return this byte to input */
		unput(yytext[0]);
		BEGIN(expect_packets);
		}

<start_pack,expect_packets>{PREFIX} {
		/* PACKET */
		BEGIN(expect_stream_id);
		bytesread += yyleng;
		return(PACKET_START_CODE_PREFIX);
		}

<expect_stream_id>[\xBC-\xFF] {
		BEGIN(expect_pkt_len);
		bytesread += yyleng;
		return(STREAM_ID);
		}

<expect_stream_id>"\xBA" { /* no more packets in pack, we hit new pack */
		/* this can happen when we finish a packet */
		BEGIN(start_pack);
		bytesread += yyleng;
		return(PACK_START_CODE);
		}

<expect_stream_id>"xB9" { /* ISO_11172_END_CODE */
		BEGIN(INITIAL);
		bytesread += yyleng;
		return(ISO_11172_END_CODE);
		}

<expect_pkt_len>{ANY}{2} {
		BEGIN(expect_scale);
		bytesread += yyleng;
		return(PACKETLEN);
		}

<expect_scale>("\xFF")+ { /* just stuffing bytes 1..16 */
		bytesread += yyleng;
		return(PACKETSTUFF);
		}

<expect_scale>[\x40-\x7F]{ANY} { /* optional buffer scale/size appears */
		bytesread += yyleng;
		return(PACKETSCALE);
		}

<expect_scale>[\x20-\x2F]{ANY}{4} { /* PTS in packet */
		if (len > 5)
			BEGIN(data);
		else
			BEGIN(expect_packets);
		bytesread += yyleng;
		return(PTSONLY);
		}

<expect_scale>[\x30-\x3F]{ANY}{9} { /* PTS and DTS in packet */
		if (len > 10)
			BEGIN(data);
		else
			BEGIN(expect_packets);
		bytesread += yyleng;
		return(PTSDTS);
		}

<expect_scale>"\x0F" { /* no PTS or DTS in packet */
		if (len > 1)
			BEGIN(data);
		else
			BEGIN(expect_packets);
		bytesread += yyleng;
		return(NOPTSORDTS);
		}

<data>{ANY}	{ /* packet contents */
		if (yyleng < len) { 
			/* shrink length yet to read */
			/* retain yytext on next match */
			yymore(); 
			}
		else {
			BEGIN(expect_packets);
			bytesread += yyleng;
			return(COMPRESSED_DATA);
			}
		}

{ANY}		{ /* unknown */
		fprintf(stderr,"\n* * * BOMB = 0x%02x * * *\n",
			yytext[0]);

		bytesread += yyleng;
		return(BOMB);
		}		


<INITIAL><<EOF>> { /* if we are in start state, this is a good time to end */
		yyterminate();
		}

<data><<EOF>> { /* end of input mid-stream! */
		fprintf(stderr,"\n*** EOF before termination!\n");
		fprintf(stdout,"\n<<EOF>>\n");
		bytesread += yyleng;
		return(COMPRESSED_DATA);
		}

<<EOF>>		{ /* EOF for all remaining conditions */
		fprintf(stderr,"\n* * * EOF BOMB * * *\n",
			yytext[0]);
		bytesread += yyleng;
		return(BOMB);
		}

%%







