%{
/*
{*NAME*}
	__FILE__

{*SYNOPSIS*}
	__FILE__

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}
	byacc or bison source for system stream parser

{*ENVIRONMENT*}

{*SYSTEM REQUIREMENTS*}

{*ALGORITHM*}

{*COMPILATION INSTRUCTIONS*}
	Needs sysdec2.lex to build sysdec2.

{*RESOURCE CONTROL*}
	$Id: sysdec2.y,v 1.11 1995/07/18 23:52:17 quandt Exp quandt $
	$Log: sysdec2.y,v $
 * Revision 1.11  1995/07/18  23:52:17  quandt
 * func:misc
 * Added byte counting routines.
 *
 * Revision 1.10  1995/02/09  19:37:18  roy
 * func:cont
 * corrected stream id module for proper
 * reporting of multiple audio headers.
 *
 * Revision 1.9  1994/09/30  00:23:45  roy
 * bug:misc
 * omitted place marker for "no size" stats report after scale marker
 *
 * Revision 1.8  1994/09/29  00:09:02  roy
 * bug:misc
 * forgot to remove debug code from previous error
 *
 * Revision 1.7  1994/09/29  00:07:52  roy
 * bug:misc
 * SCR conversions suffering from bad type conversion.
 *
 * Revision 1.6  1994/09/28  21:04:26  roy
 * bug:misc
 * used hex value 0x60 to mask video lock, when it should
 * have been bit mask 0x40.
 *
 * Revision 1.5  1994/09/28  20:50:55  roy
 * func:cont
 *
 * Revision 1.4  1994/09/23  23:49:36  roy
 * func:cont
 *
 * Revision 1.3  1994/09/23  22:06:30  roy
 * func:cont
 *
 * Revision 1.2  1994/09/23  18:13:42  roy
 * *** empty log message ***
 *
 * Revision 1.1  1994/09/23  18:08:59  roy
 * Initial revision
 *

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}
Brian A. Roy
Brian A. Quandt

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "err.h"		/* our error defines */
#include "sysdec2.p"		/* prototypes */

extern char * yytext;
extern int yyleng;
extern unsigned int bytesread;

/* so we can share packet lengths; flex semantic value deps. */
extern unsigned int len; 

#define YYERROR_VERBOSE
#define DFLT_VERBS 2	/* output control */
#define MAX_NAME_SIZE 256 /* output file name limit */

/* a nod to readability */
#define RESERVED	0
#define PRIVATE1	1
#define PADDING		2
#define PRIVATE2	3
#define AUDIO		4
#define VIDEO		5
#define DATA		6

#define STYPES		7

int compdata = 0;

/* pack counter */
int packcount = 0;

/* last packet type for data handling */
unsigned int lastpkt_type = 7;
unsigned int lastpkt_num = 0;

/* this is here so we can control ALL printf()'s */
int verbose = DFLT_VERBS;

%}

%token PACK_START_CODE
%token ISO_11172_END_CODE
%token SYSTEM_HEADER_START_CODE
%token PACKET_START_CODE_PREFIX
%token STREAM_ID
%token STREAM_ID_PLUS
%token COMPRESSED_DATA
%token PACKETLEN
%token PACKETSCALE
%token PACKETSTUFF
%token PTSONLY
%token PTSDTS
%token NOPTSORDTS
%token BOMB

%%

start:	maybestream

maybestream: 	/* empty */
		| maybestream iso11172_stream
;

iso11172_stream:	packs
			ISO_11172_END_CODE { 
				if (verbose) printf("\n * ISO_11172_END_CODE *\n");
				SysTotal(); }
;

packs:	pack
	| packs pack
;

pack:   	PACK_START_CODE { /* token consumes SCR and MUX */  
			++packcount;
			if (verbose > 1) 
				printf("\n\n* Pack %d (Byte Offset 0x%x) *\n", packcount,bytesread - 0x0c); 
			else if (verbose)
				printf("\nPACK: %01lu (0x%x) ", packcount,bytesread - 0x0c);
			outfTS(&yytext[4],((verbose > 1)?"\tSCR = %0.0f"
				:"%0.0f ")); 
			muxrate(&yytext[9]); }
		pack_contents
;

pack_contents:	system_header /* only one optional system header */
		packets
		| packets
;

/* we start with empty in case only system header or empty pack */
packets:	/* empty */
		| packets packet
;

packet:		PACKET_START_CODE_PREFIX { 
			if (verbose > 1) printf("\n\tpacket"); }
		STREAM_ID { 
			StreamID(1); }
		PACKETLEN {
			len = yytexttoint();
			if (verbose) printf(((verbose > 1)?"\t\tLength = %d\n"
				:"%01d "),len); } 
		packetstuff 
		packetscale
		timestamp  
		data {
			if (verbose > 1) printf("\t\t(End of packet data)\n");}
;

packetstuff: 	/* no stuffing */ { 
			if (verbose > 1) printf("\t\t(no stuffing)\n"); 
			else if (verbose) printf("* "); }
		| PACKETSTUFF { 
			if (verbose) printf(
				((verbose > 1) ? "\t\t%d stuffing bytes\n"
				: "%01d "),yyleng); 
			len -= yyleng; }
;

packetscale: 	/* no scale */ { 
			if (verbose > 1) printf("\t\t(no scale)\n"); 
			else if (verbose) printf("* * "); }
		| PACKETSCALE { 
			PacketScale(); 
			len -= yyleng; }
;

timestamp:	NOPTSORDTS {
			if (verbose > 1) printf("\t\t(no PTS or DTS)\n"); 
			else if (verbose) printf("* * ");
			len -= yyleng; }
		| PTSDTS { 
			outfTS(yytext,
				((verbose > 1)?"\t\tPTS = %0.0f\n":"%0.0f ")); 
			outfTS(&yytext[5],
				((verbose > 1)?"\t\tDTS = %0.0f\n":"%0.0f ")); 
			len -= yyleng; } 
		| PTSONLY { 
			outfTS(yytext,
				((verbose > 1)?"\t\tPTS = %0.0f\n":"%0.0f * ")); 
			len -= yyleng; }
;

data:		/* no data */
		| data
		COMPRESSED_DATA {
			DataHandler(); }
;

system_header: 	SYSTEM_HEADER_START_CODE { /* consumes all but stream ID's */
			SystemHeader(); }
		optional_streamids
;

optional_streamids:  /* no stream ID's */
		| optional_streamids 
		STREAM_ID_PLUS { 
			StreamIDplus(); }
;

%%

/* count of packets for each stream, by type and number */
int mixpkts[4] = { 0 };
int audpkts[0x1F] = { 0 };
int vidpkts[0xF] = { 0 };
int datpkts[0xF] = { 0 };

/* this set of global strings shared by a couple of functions */
const char * stypes[STYPES] = {
	"(reserved)",
	"Private 1",
	"Padding",
	"Private 2",
	"Audio",
	"Video",
	"(res. data)",
	};

/* output file stuff */
FILE *files[STYPES] = { NULL };
int snums[STYPES]; /* set these to invalid stream numbers */

/* end global declarations */
/*
{*NAME*}
	main -- entry point for sysdec2 program

{*SYNOPSIS*}
	int main(int argc, char * argv[]) \* normal args *\

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}

{*ENVIRONMENT*}

{*ALGORITHM*}
	Parse args, if any.
	Call yyparse().

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
int main(int argc, char *argv[])
{
	int i;
	
	for ( i = 0; i < STYPES; i++)
	{
		snums[i] = -1;
		files[i] = NULL;
	}

	if (parse_args(argc, argv, files, snums, &verbose) == ERR_NONE)
		return(yyparse());

	for ( i = 0; i < STYPES; i++)
		if (files[i])
			fclose(files[i]);

} /* end main() */
/*
{*NAME*}
	yyerror -- parser error handling for sysdec2

{*SYNOPSIS*}
	int yyerror(char *s) \* error string *\

{*CLASS*}
	c:io

{*DESCRIPTION*}

{*ENVIRONMENT*}

{*ALGORITHM*}

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
int yyerror(char *s)
{
	/* yerrok() action?  Can we identify cases where this is likely? */
	if (verbose) printf("%s\n",s);
	outyytext();
	SysTotal();
} /* end yyerror() */
/*
{*NAME*}
	SysTotal -- print summary information for sysdec2

{*SYNOPSIS*}
	int SysTotal(void) \* no args *\

{*CLASS*}
	c:i/o

{*DESCRIPTION*}

{*ENVIRONMENT*}

{*ALGORITHM*}

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
int SysTotal(void) /* describe and reset figures for pack and packet counts */
{
	int i;
	if (verbose) printf("\n\tTotal packs: %d\n", packcount);

	if (verbose) printf("\tPacket Counts\n");
	if (verbose) printf("\t---------------\n");
	for (i = 0; i < 4; i++) if (mixpkts[i]) {
		if (verbose) printf("\t%s :  %d  (0x%01x)\n",
			stypes[i],mixpkts[i],mixpkts[i]);
		mixpkts[i] = 0;
		}
	for ( i = 0; i < 0x1F ; i++) if (audpkts[i]) {
		if (verbose) printf("\tAudio %01d :  %d  (0x%01x)\n",
			i,audpkts[i],audpkts[i]);
		audpkts[i] = 0;
		}
	for ( i = 0; i < 0xF ; i++) if (vidpkts[i]) {
		if (verbose) printf("\tVideo %01d :  %d  (0x%01x)\n",
			i,vidpkts[i],vidpkts[i]);
		vidpkts[i] = 0;
		}
	for ( i = 0; i < 0xF ; i++) if (datpkts[i]) {
		if (verbose) printf("\tData %01d :  %d  (0x%01x)\n",
				i,datpkts[i],datpkts[i]);
		datpkts[i] = 0;
		}
	packcount = 0; 
} /* end SysTotal() */
/*
{*NAME*}
	outyytext -- format yytext for output as hex digits

{*SYNOPSIS*}
	int outyytext(void) \* no args *\

{*CLASS*}
	c:i/o

{*DESCRIPTION*}

{*ENVIRONMENT*}

{*ALGORITHM*}

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
int outyytext(void)
{
	int i;
	for(i=0;i<yyleng;i++)
		if (verbose) printf("%2.2x ", 0x00FF & yytext[i]);
} /* end outyytext() */
/*
{*NAME*}
	outfTS -- format and print time stamp info for sysdec2

{*SYNOPSIS*}
	void outfTS(char *ptr, \* addr of bytes containing time stamp *\
		char *format) \* format string for output *\

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}

{*ENVIRONMENT*}

{*ALGORITHM*}

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
void outfTS(char *ptr,char *format) /* Out timestamp; SCR's PTS, DTS, etc */
{
	unsigned long tmp, TShi, TSlo;
	float TS33bit;

	/* TShi is a bit!  
		We do nothing with it for now. */

	/* get first three bits */
	tmp = ptr[0]; 
	TSlo = tmp & 0xE;
	/* retain highest bit */
	TShi = (TSlo && 8);
	TSlo <<= 7;

	/* next 8 bits */
	tmp = (ptr[1] & 0xFF);
	TSlo <<= 8L;
	TSlo |= tmp;
	
	/* next 7 bits */
	tmp = (ptr[2] & 0xFE);
	TSlo <<= 7L;
	TSlo |= tmp >> 1L;

	/* next 8 */
	tmp =  (ptr[3] & 0xFF);
	TSlo <<= 8;
	TSlo |= tmp;

	/* last 7 */
	tmp =  (ptr[4] & 0xFE);
	TSlo <<= 7;
	TSlo |= tmp >> 1L;
	
	TS33bit = (float)TSlo;
	if (TShi) TS33bit *= 2.0;
	if (verbose) printf(format,TS33bit);

} /* end outfTS() */
/*
{*NAME*}
	PacketScale -- decode scale/size info for sysdec2

{*SYNOPSIS*}
	void PacketScale(void)	\* no args *\
	
{*CLASS*}
	c:mpeg

{*DESCRIPTION*}

{*ENVIRONMENT*}

{*ALGORITHM*}

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
void PacketScale(void)	/* decode scale/size info */
{
	int scale;
	int size;

	scale = (((yytext[0] & 0xFF) >> 5) & 1);
	size = (yytext[0] & 0x1F);
	size <<= 8;
	size |= (yytext[1] & 0xFF);
	size *= (scale ? 1024 : 128);

	if (verbose > 1) 
	{
		printf("\t\tSTD_buffer_scale %01d\n", scale);
		printf("\t\tSTD_buffer_size %d\n", size);
	}
	else if (verbose)
	{
		printf("%01d %01d ",scale,size);
	}
} /* end PacketScale() */
/*
{*NAME*}
	StreamID -- decode stream ID info from yytext[0] for sysdec2

{*SYNOPSIS*}
	void StreamID(int bumpcount) \* change counter flag *\

{*CLASS*}
	c:i/o

{*DESCRIPTION*}

{*ENVIRONMENT*}

{*ALGORITHM*}

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
void StreamID(int bumpcount) /* decode stream ID info from yytext[0] */
{
	int strnum = 0;
	int strmtype = 0;
	int base = (yytext[0] & 0xFF);

	/* kludge - make any value 1 to save heartache later */
	if (bumpcount)
		bumpcount = 1;

	/* identify type of stream */
	/* and get number, if relevant */
	if ( base >= 0xF0 ) {
		strmtype = DATA;
		strnum = base & 0xF;
		datpkts[strnum] += bumpcount;
		}
	else if ( base >= 0xE0 ) {
		strmtype = VIDEO;
		strnum = base & 0xF;
		vidpkts[strnum] += bumpcount;
		}
	else if ( base >= 0xC0 ) {
		strmtype = AUDIO;
		strnum = base & 0x1F;
		audpkts[strnum] += bumpcount;
		}
	else if ( base == 0xB8 ) {
		/* this may be a header used to describe */
		/* all audio streams in mix */
		strmtype = AUDIO;
		strnum = -1;
		/* audpkts[strnum] += bumpcount; */
		}
	else {
		/* base must be in range BC-BF */
		/* our corresponding shorthand is 0-3 */
		strmtype = base - 0xBC;
		mixpkts[strmtype] += bumpcount;
		}

	if (verbose > 1) 
	{
		printf("\n\t\tStream ID = %02x ( %s ",
			(yytext[0] & 0xFF),
			stypes[strmtype]);
		if (strmtype >= 4)
			if (strnum >= 0)
				printf("# %01d ",strnum);
			else
				printf(" - all ");
		printf(")\n");
	}
	else if (verbose)
	{
		if (bumpcount) /* real packets, give own line */
			printf("\n");
		printf("%02x: ",(yytext[0] & 0xff));
	}
	if (bumpcount) 
	{
		lastpkt_type = strmtype;
		lastpkt_num = strnum;
	}
} /* end StreamID() */
/*
{*NAME*}
	StreamIDplus -- decode system header stream ID info for sysdec2

{*SYNOPSIS*}
	void StreamIDplus(void) \* no args *\

{*CLASS*}
	c:i/o

{*DESCRIPTION*}

{*ENVIRONMENT*}

{*ALGORITHM*}

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
void StreamIDplus(void) /* decode system header stream ID info */
{
	int scale;
	int size;

	scale = ( ((yytext[1] & 0xFF) >> 5) & 1 );
	size = (yytext[1] & 0x1F);
	size <<= 8;
	size |= (yytext[2] & 0xFF);
	size *= (scale ? 1024 : 128);

	StreamID(0);
	
	if (verbose > 1) 
	{
		printf("\t\t\tSTD_buffer_bound_scale %01d\n", scale);
		printf("\t\t\tSTD_buffer_size_bound %d\n", size);	
	} 
	else if (verbose)
	{
		printf("%01d %01d ",scale,size);
	}
} /* end StreamIDplus() */
/*
{*NAME*}
	yytexttoint -- convert yytext bytes to an integer value

{*SYNOPSIS*}
	int yytexttoint(void) \* no args *\

{*CLASS*}
	c:i/o

{*DESCRIPTION*}

{*ENVIRONMENT*}

{*ALGORITHM*}

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
int yytexttoint(void)
{
	int i, val = 0;

	for( i = 0; i < yyleng; i++) {
		val <<= 8;
		val |= (0x00ff & yytext[i]);
		}
	return(val);
} /* end yytexttoint() */
/*
{*NAME*}
	muxrate -- decode mux rate for sysdec2

{*SYNOPSIS*}
	void muxrate(char *chmux) \* addr of bytes containing encoded rate *\

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}

{*ENVIRONMENT*}

{*ALGORITHM*}

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
void muxrate(char *chmux)
{
	int mux;
	int i;

	/* mux rate is stored in 3 bytes, marker bit mux marker bit 
	so need to mask them off */

	mux = 0;
	for(i=0;i<3;i++)
		{
		mux = mux| ( 0x00ff & chmux[i]);
		mux = mux << 8;
		}
	mux = (mux >> 9)& 0x3fffff;
	if (verbose > 1) printf("  mux = ");
	if (verbose) printf("%01d ",mux);
} /* end muxrate */
/*
{*NAME*}
	DataHandler -- handle packet data for sysdec2

{*SYNOPSIS*}
	void DataHandler(void) \* no args *\

{*CLASS*}
	c:i/o

{*DESCRIPTION*}

{*ENVIRONMENT*}

{*ALGORITHM*}

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
void DataHandler(void) 
{
	if (verbose > 1) 
		printf("\t\tData:  %01d bytes\n", yyleng);
	else if (verbose)
		printf("( %01d )",yyleng);

	if (snums[lastpkt_type] >= 0) /* should we extract? */
		if ((lastpkt_type < 4) || (lastpkt_num == snums[lastpkt_type]))
			fwrite(yytext,1,yyleng,files[lastpkt_type]);
} /* end DataHandler() */
/*
{*NAME*}
	usage -- print usage information for sysdec2

{*SYNOPSIS*}
	void usage(void)

{*CLASS*}
	c:i/o

{*DESCRIPTION*}
	Prints usage information, and forces program to exit.

{*ENVIRONMENT*}
	This function does not return, but forces a program exit with
	value 0.

{*ALGORITHM*}

	Print usage instructions to stderr.
	exit.

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}

{*COPYRIGHT*}
	Copyright 1994 Heuris Logic.
	All rights reserved.
{**}
*/
void usage(void)
{
	fprintf(stderr,"sysdec2 -- Interpret ISO 11172-1 streams.\n");
	fprintf(stderr,"usage: sysdec2 [options]\n");
	fprintf(stderr,"(Use command line redirection for normal IO.)\n");
	fprintf(stderr, "-2 <n> <file>     Extract 11172-2 stream <n> to <file>.\n");
	fprintf(stderr, "-3 <n> <file>     Extract 11172-3 stream <n> to <file>.\n");
	fprintf(stderr, "-p <n> <file>     Extract private stream <n> to <file>.\n");
	fprintf(stderr, "-d <n> <file>     Extract data stream <n> to <file>.\n");
	fprintf(stderr, "-r <file>         Extract reserved stream to <file>.\n");
	fprintf(stderr, "-s <file>         Extract padding to <file>.\n");
	fprintf(stderr, "-v     Debug flag (more flags, more info; Level %d is default).\n",DFLT_VERBS);
	fprintf(stderr, "-q     Quiet flag (opposite effect of -v).\n");
	fprintf(stderr, "-h[elp],-?    Print this message.\n");
	fprintf(stderr, "NOTE: only one of each stream type may be extracted at once.\n");
	fprintf(stderr, "Debug levels: 1 = statistics, 2 = text, 3 = full trace.\n");
	exit(0);
} /* end of usage() */
/*
{*NAME*}
  parse_args -- sort out command line arguments for sysdec2 decoder

{*SYNOPSIS*}
  int parse_args(int argc, \* number of command line arguments provided *\
	char *argv[], \* argument values *\
	FILE *files[], \* address of file handles for extracted data *\
	int snums[], \* address of ID of 11172-2 stream to extract *\
	int *verbsap) \* address of debug flag *\

{*CLASS*}
 c:i/o

{*DESCRIPTION*}

  This routine will deal with setting up all variables that may be influenced
by command line options, as well as opening the required files and/or
initializing communications ports.

{*ENVIRONMENT*}

	Returns ERR_NONE for success, ERR_GENERAL on failure.

{*SYSTEM REQUIREMENTS*}

{*ALGORITHM*}

	While we have arguments but no errors,
		If the argument starts with a "-",
			figure out which option applies, and set accordingly.
	If no errors,
		Attempt to open special output files.
	Display status.
	Return error code.

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}

{*COPYRIGHT*}
	Copyright 1994 Heuris Logic.
	All rights reserved.
{**}
*/
int parse_args(int argc, char *argv[], FILE *files[], int snums[], int *verbsap) 
{
	char name[STYPES][MAX_NAME_SIZE]; /* aux file names */
	int get = -1; /* parm flags */
	int rc = ERR_NONE;
	int h, i;

	for (h = 0; h < STYPES; h++)
		name[h][0] = 0;	

	yydebug = 0;

	/* process */
	i = 0;
	while(++i<argc && (rc == ERR_NONE))
	{
		char c, *optkey, *arg, *nextArg;
		int  argUsed;
		optkey = argv[i];

		if(*optkey++ == '-') /* this is an option - check it out */
		{
			if(i+1 < argc)
				nextArg = argv[i+1];
			else
				nextArg = "";
			argUsed = 0;

		 /* get first char after hyphen */
		 while(c = *optkey++)
		 {
			if(*optkey)
				arg = optkey;
			else
				arg = nextArg;

			switch(c)
			{
				case 'h':
				case '?':
			   		/* print usage */
					usage();
					break;
			 	case 'v':
			   		/* increase verbosity */
					*verbsap += 1;
				  	break;
				case 'q':
			 	/* decrease verbosity */
					*verbsap -= 1;
					break;
			 	case '2':
			 	/* extract a 11172-2 stream */
					get = VIDEO;
					h = atoi(arg);
					if ((h >= 0) && (h <= 0x0F))
						snums[get] = h;
					else 
					{
						if (*verbsap) fprintf(stderr,"Invalid video stream number %d\n",h);
						rc = ERR_GENERAL;
					}
					argUsed = 1;
					break;
			 	case '3':
			 	/* extract a 11172-3 stream */
					get = AUDIO;
					h = atoi(arg);
					if ((h >= 0) && (h <= 0x1F))
						snums[get] = h;
					else 
					{
						if (*verbsap) fprintf(stderr,"Invalid audio stream number %d\n",h);
						rc = ERR_GENERAL;
					}
					argUsed = 1;
					break;
			 	case 'd':
			 	/* extract a reserved data stream */
					get = DATA;
					h = atoi(arg);
					if ((h >= 0) && (h <= 0x0F))
						snums[get] = h;
					else 
					{
						if (*verbsap) fprintf(stderr,"Invalid video stream number %d\n",h);
						rc = ERR_GENERAL;
					}
					argUsed = 1;
					break;
			 	case 'p':
			 	/* extract a private stream */
					h = atoi(arg);
					if ( h == 1 )
						get = PRIVATE1;
					else if ( h == 2 )
						get = PRIVATE2;
					else
					{
						if (*verbsap) fprintf(stderr,"Invalid private stream number %d\n",h);
						rc = ERR_GENERAL;
					}
					if (rc == ERR_NONE)
						snums[get] = h;
					argUsed = 1;
					break;
			 	case 'r':
			 	/* extract reserved stream */
					snums[RESERVED] = 1;
					strcpy(name[RESERVED],arg);
					argUsed = 1;
					break;
			 	case 's':
			 	/* extract padding to file */
					snums[PADDING] = 1;
					strcpy(name[PADDING],arg);
					argUsed = 1;
					break;
				default:
					/* unknown option on "-" */
					if (*verbsap) {
						fprintf(stderr,"bad args: unrecognized option %c\n",c);
						fprintf(stderr,"try \"-h\" for help\n");
						}
					rc = ERR_GENERAL;
			} /* end switch (c) */

			if(argUsed)
			{
				if(arg == optkey)
					optkey = "";
			   	else
					++i;
			   	arg = "";
			   	argUsed = 0;
			} /* end bump args */
		 } /* end while options */
	  } /* end if option keyword */
	  else
	  {	
		if (get >= 0)
		{
			strcpy(name[get],argv[i]);
			get = -1;
		} /* end if get name active */
		else
		{
			/* we have untagged parms */
			if (*verbsap)
			{
				fprintf(stderr,"bad args: excess verbiage in the form of %s\n", argv[i]);
				fprintf(stderr,"try \"-h\" for help\n");
			}
			rc = ERR_GENERAL;
		} /* end if not name */
	  } /* end deal with unlabelled parms */
   } /* end all arg tests */

   /* attempt to open our lovely files, if need be! */
   for (i = 0; (i < STYPES) && (rc == ERR_NONE); i++)
   {
	if (name[i][0] != 0)
	{
		files[i] = fopen(name[i],"wb");
		if (!files[i])
		{
			rc = ERR_GENERAL;
			if (*verbsap)
				fprintf(stderr,"Cannot open %s for output!\n",name[i]);
		}
	 }
   } /* end if okay to init input */

   /* if user wants feedback, start now... basic stats */
   if (*verbsap)
   {
	fprintf(stderr,"Debug messages to level: %d\n",*verbsap);
	if (*verbsap > 2)	
	{
		yydebug = 1;
		fprintf(stderr,"\t* MAXIMUM VERBOSITY *\n");
	} 
	else if (*verbsap > 1)
		fprintf(stderr, "\tFull text descriptions will be used.\n");
	if (*verbsap > 1) 
	{
		for ( h = 0; h < STYPES; h++) if ( snums[h] >= 0 )
		{
			if (h < 4)
				fprintf(stderr,"Extracting %s -> %s\n",stypes[h],name[h]);
			else 
				fprintf(stderr,"Extracting %s stream %01d -> %s\n",stypes[h],name[h]);
		}
		fprintf(stderr,"Status:  %s\n", ((rc == ERR_NONE)?"Okay":"Abort"));
	}
   }
   return rc;
}/* end of parse_args */
/*
{*NAME*}
	SystemHeader -- interpret and output header information for sysdec2

{*SYNOPSIS*}
	void SystemHeader(void) \* no args *\

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}

{*ENVIRONMENT*}

{*ALGORITHM*}

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
void SystemHeader(void)
{
	long ltmp;
	int tmp;
	int txt;

	if (verbose)
	{ 
		txt = verbose - 1; /* full text? */
		printf("\n%s", (txt ? "\tSystem Header" : "SYS: "));
/* bytes 5-6 of token are length */ 
		printf((txt ? "  Length = %0d\n" : "%01d "), 
			(((yytext[4]&0x0ff)<<8)|(yytext[5]&0x0ff)));
/* bytes 7-9 are marker-ratebound-marker */
		ltmp = (yytext[6] & 0x007f);
		ltmp <<= 8;
		ltmp |= (yytext[7] & 0x00ff);
		ltmp <<= 8;
		ltmp |= (yytext[8] & 0x00fe);
		ltmp >>= 1;
		printf((txt ? "\t\tRate bound = %ld\n":"%01ld "),ltmp);
/* 10 is audio bound, fixed_flag, CSPS flag */
/* 11 has audio lock, video lock, marker, and video bound */
		tmp = (yytext[9] & 0x00ff) >> 2;
		printf((txt ? "\t\tAudio bound = %d\n":"%01d "),tmp);
		if (yytext[9] & 0x02)
			printf((txt?"\t\tFixed bitrate indicated.\n":"1 "));
		else
			printf((txt?"\t\tVariable bitrate allowed.\n":"0 "));
		if (txt)
		{
			if (yytext[9] & 0x01) printf("\t\tConstrained system parameter stream.\n");
			if (yytext[10] & 0x80) printf("\t\tAudio lock.\n");
			if (yytext[10] & 0x40) printf("\t\tVideo lock.\n");
		}
		else
		{
			printf("%01d ",(yytext[9] & 0x01));
			printf("%01d ",(yytext[10] & 0x80));
			printf("%01d ",(yytext[10] & 0x40));
		}
		tmp = (yytext[10] & 0x1f);
		printf((txt?"\t\tVideo bound = %d\n":"%01d "),tmp);
	} 
}



