/*
{*NAME*}
	__FILE__ -- Routines for generating CDI mpeg1 ISO11172-1 stream

{*SYNOPSIS*}
	__FILE__

{*CLASS*}
	files:generic

{*DESCRIPTION*}
	This program is intended to create ISO11172-1 compliant system
	streams.  Its orginal incarnation was to create only CD-I compliant
	(video CD) streams, however, the parameter file now allows generic
	streams to also be created.  CD-I (Video CD) really only causes
	one problem in compliance and that is in the 20 byte audio padding.

{*ENVIRONMENT*}

{*SYSTEM REQUIREMENTS*}
	unix -- originally developed under both Sun 4.1.x and Linux 1.x

{*COMPILATION INSTRUCTIONS*}
	make binary

{*RESOURCE CONTROL*}
	$Id: cdisys.c,v 1.12 1996/05/01 22:49:03 quandt Exp quandt $
	$Log: cdisys.c,v $

{*SEE ALSO*}

{*BUGS*}
	EOF is not handled properly at this point

{*CREDITS*}
	HEURIS Logic -- Brian Quandt

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/


#include <stdio.h>
#ifdef __GNUC__
#include <unistd.h>
#else
#define R_OK fix me please
#endif


/* kludge code for dos, so that I have 32 bit ints to work with  */
/* #define int long */

char tas[256];	/* junk string for writing debug messages to */

#include <cdisys.h>
#include <cdisys.p>
#include <strstrn.p>


/* global output handle so that don't have to include file handle to outbits all the time */
extern FILE *outfile;

/*
{*NAME*}
	ProcessStreams -- CDI ISO11172-1 stream encoder.

{*SYNOPSIS*}
	int ProcessStreams(struct inparams *UserParams, \* User parameters *\
		struct calcparams *CalcParams)	\* Calculated Parameters *\

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}
	This is the actual work function of this program.  It loops
	until the end of processing.  It spits out either a audio pack, 
	video pack or padding pack based on buffer fullness.

{*ALGORITHM*}
	If CDI
		output the CDI type headers -- 1 for audio, 1 for video
	else
		output a normal system type header

	While continue processing
		Output a pack based on buffer fullness, priority given to audio.
		Update and check the buffers.

{*CREDITS*}
	HEURIS Logic -- Brian Quandt

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
int ProcessStreams(struct inparams *UserParams, /* User parameters */
		struct calcparams *CalcParams)	/* Calculated Parameters */
	{
	int	i,
		vidsect,	/* num of video sect to spit out before sending audio sect*/
		vidcnt=0,		/* total number of video sectors sent */
		audcnt=0,		/* total audiosectors sent */
		counter;
	float ftmp;
	float currentratio;	/* current relationship of audio:video */
	float padsectcount = 0.0;
	float vidsectcount = 0.0;
	float audsectcount = 0.0;
	float sectcount = 0.0;
	float VBV = 0.0;	/* buffer for video */
	float ABV = 0.0;	/* buffer for audio */
	float VBV1 = 0.0;
	float ABV1 = 0.0;
	int emergency = 0;
	int first_loop = 1;
	int bufferchoice;


	/* copy the output file handle so that the outbits routines can see it */
	outfile = UserParams->outfile;

	/* Just send out the first sectors for audio and video, these
	don't contain any real data they just initialize the decoder */

	if (UserParams->CDI_type)
		{
		cdi_vid_sect0(UserParams,CalcParams);
		sectcount++; 
		CalcParams->SCR += UserParams->dsmpacksize/(UserParams->muxrate * 50)*CLOCKFREQ;


		cdi_aud_sect0(UserParams,CalcParams);
		sectcount++;
		CalcParams->SCR += UserParams->dsmpacksize/(UserParams->muxrate * 50)*CLOCKFREQ;
		}
	else
		{
		cdi_pad_sect0(UserParams,CalcParams);
		sectcount++; 
		CalcParams->SCR += UserParams->dsmpacksize/(UserParams->muxrate * 50)*CLOCKFREQ;
		}


	/* now go into the general process loop */
	while (nextsector(UserParams))
		{
#ifdef DEBUG_BUFF
sprintf(tas,"SCR: %10d   VBV %10.0f ABV %10.0f ",CalcParams->SCR,VBV,ABV);
xuxa_log(tas);
#endif

#ifdef DEBUG_BUFFVID
sprintf(tas,"%8.0f\n",VBV); 
xuxa_log(tas);
#endif
#ifdef DEBUG_BUFFAUD
sprintf("%8.0f\n",ABV); 
xuxa_log(tas);
#endif

		if (first_loop)
			{
			emergency = 0;	/* eliminate any false emergencies on starting up*/
			first_loop = 0;
			}

		/* The control of whether to output a video audio or padding
		is controled by buffer fullness.  Ie if the buffer in question is almost
		empty better put another one out, The minimum is controlled
		by the amount each 1/75.0 of a second that needs to go out, the max
		is controled by current buffer fullness plus amount added (ie don't
		go over )  priority is given to Audio */


		if (MAXAUDBUF > (ABV + UserParams->auddatasize))
			{

#ifdef DEBUG_BUFF
xuxa_log("aud \n");
#endif
			cdi_aud_sect(UserParams,CalcParams);
			bufferchoice = AUDIO;
			}
		else if (MAXVIDBUF > (VBV + UserParams->viddatasize))
			{
#ifdef DEBUG_BUFF
xuxa_log("vid \n");
#endif
			cdi_vid_sect(UserParams,CalcParams);
			bufferchoice = VIDEO;
			}
		else
			{
#ifdef DEBUG_BUFF
xuxa_log("pad \n");
#endif
			cdi_pad_sect(UserParams,CalcParams);
			bufferchoice = 0;
			}

		/* Drain/fill the buffers */
		drainbuffers(UserParams,CalcParams,&ABV,&VBV,bufferchoice);


		/* Advance the SCR by an amount equal to the time a sector goes by */
		CalcParams->SCR += UserParams->dsmpacksize/(UserParams->muxrate * 50)
			*CLOCKFREQ;

		}

	/* Put a end code on the stream, if Video CD then the end codes are */
	/* added earlier */
	if (!UserParams->CDI_type)
		{
		cdi_pad_sect_endcode(UserParams,CalcParams);
		}

	}/* end of process streams */



/*
{*NAME*}
	outputemptypack -- Output an empty pack.

{*SYNOPSIS*}
	static int outputemptypack(int totallength, \* lenght of a pack *\
		struct inparams *UserParams,
		struct calcparams *CalcParams)

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}
	Send a padding PACK to the output.  

{*ALGORITHM*}
	Output all the header information for this pack.
	Output padding packet of the required size.

{*CREDITS*}
	HEURIS Logic -- Brian Quandt

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
static int outputemptypack(int totallength, /* lenght of a pack */
	struct inparams *UserParams,
	struct calcparams *CalcParams)
	{
	outbits(PACKSTARTCODE,0,NULL,32L,31L);		/* pack start */
	outbits(0x02L,0,NULL,4,3);					/* 0010 */


	/* SCR */
	outbits(0x00L,0,NULL,1,0);	/* this is actually the first SCR bit, but 
								since I don't have 33 bit ints... */
	outbits(CalcParams->SCR,0,NULL,2,31);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,29);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,14);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(MARKERBIT,0,NULL,1,0);

	/* MUXRATE */
	outbits(UserParams->muxrate,0,NULL,22,21);	/* muxrate */
	outbits(MARKERBIT,0,NULL,1,0);

	outputpadding(totallength-12,UserParams,CalcParams);
	}

/*
{*NAME*}
	outputpadding -- Output a CDI/ISO11172-1 compliant padding stream.

{*SYNOPSIS*}
	static int outputpadding(int totallength, \* lenght of a padding packinet *\
		struct inparams *UserParams,
		struct calcparams *CalcParams)

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}
	This routine will output a padding PACKET.

{*ALGORITHM*}
	Output packet header stuff.
	Output padding bytes necessary to fill the packet.

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}
	HEURIS Logic -- Brian Quandt

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
static int outputpadding(int totallength, 	/* lenght of a padding stream */
		struct inparams *UserParams,
		struct calcparams *CalcParams)
	{
	int tmp,
		i;
	char 	padblk[4096];

#ifdef DEBUG
xuxa_log("\npadding\n");
#endif


	/* output a padding packet of the exact length specified */
	outbits(PACKETSTARTCODE,0,NULL,24,23);
	outbits(PADDINGSTREAMID,0,NULL,8,7);

	/* The length is calculated as the amount remaining to send after
	the length field... the total so far is 24bits + 8bits + 16bits = 6 bytes */
	tmp = totallength - 6;
	outbits(tmp,0,NULL,16,15);

	/* According to the White book 3.2.2 need to make sure that the data
	bytes is on even boundary... so I'll throw a stuffing byte in 
	if  not padding even number of bytes */

	if (tmp%2 == 0)	
		outbits(STUFFINGBYTE,0,NULL,8,7);

	/* now the else part of the packet */
	outbits(0x0FL,0,NULL,8,7);

	/* now the actual data that is part of a padding packet */
	/* this is a loop to fill out the remaining bytes, so far in this
	packet we have sent 7 bytes
	for(i=0;i<(totallength-7);i++)
		outbits(PADDINGBYTE,0,NULL,8,7);
	For now I'm going to use some memory routines so that things
	are cleaner in debugging and print outs.  In real time systems
	it is probably better to just have a static block of data to pull
	from */

	if (tmp%2 == 0)
		{
		memset(padblk,PADDINGBYTE,(totallength-8));
		outbits(0,1,padblk,(totallength-8)*8,7);
		}
	else
		{
		memset(padblk,PADDINGBYTE,(totallength-7));
		outbits(0,1,padblk,(totallength-7)*8,7);
		}

	}

/*
{*NAME*}
	getpicturetype -- Returns I B or P frame for an MPEG encoded picture.

{*SYNOPSIS*}
	int getpicturetype(char *ptr) \* pointer to picture start code *\

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}
	Given a pointer to  picture startcode determine the type of picture
	that is contained in the following data (I|B|P).  This is 
	needed in determining how much to increment the PTS/DTS values
	in later calculations.


{*ALGORITHM*}
	Pointer is to the beginning of a startcode, use the spec to see
	where the bits for picture type fall.

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}
	HEURIS Logic -- Brian Quandt

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
int getpicturetype(char *ptr) /* pointer to picture start code */
	{
	/* ptr point to a PICTURE START CODE */
	int picttype;
	char tmp;

	/* the picture type should be found as follows */
	/* (32bits for start code ) (10 bits for temporal_reference) 
	(3 bits for picture type ) */

	/* WARNING THIS CODE ASSUMES 8 BIT BYTES..., of if longer chars, 
	that only bits 0..7 have data. */
	tmp = ptr[5] & 0x38;	/* 00111000, mask out unwanted bits */
	picttype = tmp;
	return(picttype);
	}

/*
{*NAME*}
	exitroutine -- CDISYS exit routine (die die die die as in failed).

{*SYNOPSIS*}
	int exitroutine()

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}
	General death routine, if need to exit in the code.

{*ALGORITHM*}
	Just die with some lovely message.

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}
	HEURIS Logic -- Brian Quandt

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
int exitroutine()
	{
	xuxa_log("According to ISO/ITC standards of diplomacy,\n");
	xuxa_log("it has become necessary for this program to fail.\n");
	xuxa_log("I'm sorry, please take any problems you have encountered to your\n");
	xuxa_log("respective representative.\n");
	exit(-1);	/*die die die die die die*/
	}



/*
{*NAME*}
	cdi_pad_sect0 -- Routine to output first sector of system stream

{*SYNOPSIS*}
	int cdi_pad_sect0(struct inparams *UserParams,
		struct calcparams *CalcParams)

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}
	This routine outputs the first pack.  Its sole purpose is to output
	a systems header as defined by ISO11172-1.  Whereas
	vid_sect0 and aud_sect0 put out seperate audio and video system headers
	this one puts out a single header with audio and video defined
	within it.


{*ALGORITHM*}
	Out the pack header
	Out the sytems header defining both audio and video
	Output a padding packet to fill in the remaining.

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}
	HEURIS Logic -- Brian Quandt

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/

int cdi_pad_sect0(struct inparams *UserParams,
	struct calcparams *CalcParams)
	{
	/* Output CDI compliant mpeg padding sector with SYSTEM header */
	/* this one has the audio and video system id's defined */
#ifdef DEBUG
xuxa_log("\ncdi_pad_sect0\n");
#endif

	outbits(PACKSTARTCODE,0,NULL,32L,31L);		/* pack start */
	outbits(0x02L,0,NULL,4,3);					/* 0010 */


	/* SCR */
	outbits(0x00L,0,NULL,1,0);	/* this is actually the first SCR bit, but 
								since I don't have 33 bit ints... */
	outbits(CalcParams->SCR,0,NULL,2,31);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,29);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,14);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(MARKERBIT,0,NULL,1,0);

	/* MUXRATE */
	outbits(UserParams->muxrate,0,NULL,22,21);	/* muxrate */
	outbits(MARKERBIT,0,NULL,1,0);

	/* SYSTEM header as stated in the white book figure IV.10 */
	outbits(SYSTEMHEADERSTARTCODE,0,NULL,32,31);
	outbits(12L,0,NULL,16,15);		/* system header length's always fixed */
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(UserParams->muxrate,0,NULL,22,21);	/*ratebound */
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(1L,0,NULL,6,5);	/*audiobound */
	outbits(UserParams->fixed_flag,0,NULL,1,0);		/* fixed_flag */
	outbits(UserParams->CSPS_flag,0,NULL,1,0);		/* CSPS */
	outbits(UserParams->audio_lock,0,NULL,1,0);		/* audio_lock */
	outbits(UserParams->video_lock,0,NULL,1,0);		/* video_lock */

	outbits(MARKERBIT,0,NULL,1,0);
	outbits(1L,0,NULL,5,4);		/*video bound */
	outbits(0xFFL,0,NULL,8,7);	/* reserved byte */

	/* the std bounds for audio */
	outbits(UserParams->audiostreamid,0,NULL,8,7);	/* audio stream id */
	outbits(0x03L,0,NULL,2,1);
	outbits(0x00L,0,NULL,1,0);	/* STD buffer scale, indicates scale of video */
	outbits(32L,0,NULL,13,12);	/* CSPS is set this in constrained to 4k */

	/* the std bounds for video */
	outbits(UserParams->videostreamid,0,NULL,8,7); /* video stream id */
	outbits(0x03L,0,NULL,2,1);
	outbits(0x01L,0,NULL,1,0);	/* STD buffer scale, indicates scale of video */
	outbits(46L,0,NULL,13,12);	/* since CSPS is set this in constrained to 46*1024 */

	/* Padding packet of 2297 bytes to fill out the rest of the sector */
	outputpadding(UserParams->viddatasize, UserParams, CalcParams);
	}


/*
{*NAME*}
	cdi_vid_sect0 -- Output a proper CD-I video sector 0.

{*SYNOPSIS*}
	int cdi_vid_sect0(struct inparams *UserParams,
		struct calcparams *CalcParams)

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}
	Output a system header as defined by the Video CD spec, this only
	defines the video streams.

{*ALGORITHM*}
	Output a pack header 
	Output a system header defining only the video stream
	Output a padding stream to fill the rest of the packet.

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}
	HEURIS Logic -- Brian Quandt

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
int cdi_vid_sect0(struct inparams *UserParams,
	struct calcparams *CalcParams)
	{
	/* Output CDI compliant mpeg video sector with SYSTEM header */
#ifdef DEBUG
xuxa_log("\ncdi_vid_sect0\n");
#endif

	outbits(PACKSTARTCODE,0,NULL,32L,31L);		/* pack start */
	outbits(0x02L,0,NULL,4,3);					/* 0010 */


	/* SCR */
	outbits(0x00L,0,NULL,1,0);	/* this is actually the first SCR bit, but 
								since I don't have 33 bit ints... */
	outbits(CalcParams->SCR,0,NULL,2,31);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,29);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,14);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(MARKERBIT,0,NULL,1,0);

	/* MUXRATE */
	outbits(UserParams->muxrate,0,NULL,22,21);	/* muxrate */
	outbits(MARKERBIT,0,NULL,1,0);

	/* SYSTEM header as stated in the white book figure IV.10 */
	outbits(SYSTEMHEADERSTARTCODE,0,NULL,32,31);
	outbits(9L,0,NULL,16,15);		/* system header length's always fixed, actually an error needs to be calc'ed from input params */
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(UserParams->muxrate,0,NULL,22,21);	/*ratebound */
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(0L,0,NULL,6,5);	/*audiobound */
	outbits(UserParams->fixed_flag,0,NULL,1,0);		/* fixed_flag */
	outbits(UserParams->CSPS_flag,0,NULL,1,0);		/* CSPS */
	outbits(UserParams->audio_lock,0,NULL,1,0);		/* audio_lock */
	outbits(UserParams->video_lock,0,NULL,1,0);		/* video_lock */

	outbits(MARKERBIT,0,NULL,1,0);
	outbits(1L,0,NULL,5,4);		/*video bound */
	outbits(0xFFL,0,NULL,8,7);	/* reserved byte */

	/* the std bounds for video */
	outbits(UserParams->videostreamid,0,NULL,8,7); /* video stream id */
	outbits(0x03L,0,NULL,2,1);
	outbits(0x01L,0,NULL,1,0);	/* STD buffer scale, indicates scale of video */
	outbits(46L,0,NULL,13,12);	/* since CSPS is set this in constrained 
								to 46*1024 */

	/* Padding packet of 2297 bytes to fill out the rest of the sector */
	outputpadding(2297L, UserParams, CalcParams);
	}

/*
{*NAME*}
	cdi_vid_sect -- Output a proper CD-I sector (used after cdi_vid_sect0)

{*SYNOPSIS*}
	int cdi_vid_sect(struct inparams *UserParams,
		struct calcparams *CalcParams)

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}
	Output a pack that contains video data within its packet.

{*ALGORITHM*}
	Pack header
	Out a video packet

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}
	HEURIS Logic -- Brian Quandt

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
int cdi_vid_sect(struct inparams *UserParams,
	struct calcparams *CalcParams)
	{
	/* Output CDI compliant mpeg video sector (without the SYSTEM header) */
#ifdef DEBUG
xuxa_log("\ncdi_vid_sect\n");
#endif

	outbits(PACKSTARTCODE,0,NULL,32L,31L);		/* pack start */
	outbits(0x02L,0,NULL,4,3);					/* 0010 */

#ifdef DEBUG
sprintf(tas,"\nSCR %u\n",CalcParams->SCR);
xuxa_log(tas);
#endif

	/* SCR */
	outbits(0x00L,0,NULL,1,0);	/* this is actually the first SCR bit, 
								but since I don't have 33 bit ints... */
	outbits(CalcParams->SCR,0,NULL,2,31);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,29);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,14);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(MARKERBIT,0,NULL,1,0);

	/* MUXRATE */
	outbits(UserParams->muxrate,0,NULL,22,21);	/* muxrate */
	outbits(MARKERBIT,0,NULL,1,0);

	/* Now the packet for video */
	cdivideopacket(UserParams,CalcParams);
	}

/*
{*NAME*}
	cdivideopacket -- Output the CD-I packet associated to a sector.

{*SYNOPSIS*}
	static int cdivideopacket(struct inparams *UserParams,
		struct calcparams *CalcParams)

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}
	Routine to setup output for outputting a cdi video compliant packet.


{*ALGORITHM*}
	Read the data from the video stream
	Based on the amount read output the video packet taking care of EOF.
	If EOF output a end code

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}
	HEURIS Logic -- Brian Quandt

{*COPYRIGHT*}
	Copyright 1994 Heuris Logic.
	All rights reserved.
{**}
*/
static int cdivideopacket(struct inparams *UserParams,
		struct calcparams *CalcParams)
	{
	/* WARNING THIS MODULE IS VERY CLOSELY COUPLED TO outvideopacket, sorry */
	unsigned char buf[4048];	/* just allocated enough for now, avoid dynamic
								allocation for potential real time app */
	unsigned char buf2[4];		/* used for finding the AU at the boundary of a packet */
	int bytecount,
		numtostuff,		/* number of stuffing bytes required */
		bytesread,		/* number of bytes read from input */
		tmp,
		i;

#ifdef DEBUG
xuxa_log("\nvideopacket\n");
#endif


	/* Try to read in enough data for one VIDEO packet.  If I can't then
	add padding packets to fill out this pack with another padding packet.
	(This implies that partial video packets need another padding packet
	right next to it) */
	if (!feof(UserParams->video)&&
		(bytesread = fread(buf,1,UserParams->viddatasize,UserParams->video)) == UserParams->viddatasize)
		{
		outvideopacket(UserParams->vidpacketsize,bytesread,UserParams,CalcParams,buf);
		}
	else
		{
#if 1
		if (UserParams->CDI_type)
			{
			/* End the video packs */
			outputpadding(UserParams->vidpacketsize-4,UserParams,CalcParams);
			outbits(ISOENDCODE,0,NULL,32,31);

			/* End the audio packs */
			/* Need to put out the pack header stuff, in the code above */
			/* I did not need this since it was already done */

			CalcParams->SCR += UserParams->dsmpacksize/(UserParams->muxrate * 50)*CLOCKFREQ;

			outbits(PACKSTARTCODE,0,NULL,32L,31L);		/* pack start */
			outbits(0x02L,0,NULL,4,3);					/* 0010 */


			/* SCR value */
			outbits(0x00L,0,NULL,1,0);	/* this is actually the first SCR bit, 
								but since I don't have 33 bit ints... */
			outbits(CalcParams->SCR,0,NULL,2,31);
			outbits(MARKERBIT,0,NULL,1,0);
			outbits(CalcParams->SCR,0,NULL,15,29);
			outbits(MARKERBIT,0,NULL,1,0);
			outbits(CalcParams->SCR,0,NULL,15,14);
			outbits(MARKERBIT,0,NULL,1,0);
			outbits(MARKERBIT,0,NULL,1,0);

			/* MUXRATE */
			outbits(UserParams->muxrate,0,NULL,22,21);	/* muxrate */
			outbits(MARKERBIT,0,NULL,1,0);

			outputpadding(UserParams->audpacketsize-20-4,UserParams,CalcParams);
			outbits(ISOENDCODE,0,NULL,32,31);
			if (UserParams->VCD_nulls)
				for (i=0;i<20;i++)
					outbits(0x00L,0,NULL,8,7);
			}
		else
			{
			outputpadding(UserParams->vidpacketsize,UserParams,CalcParams);
			}

		return(0);

#endif

		/*

		bytes remaining 	What to do
		2291-2294			Drop the extra data and treat as if we have
							a video packet that is 2290 bytes 
							(2290 + 18 bytes = 2308)

		2290				Since want to output 2290 + 18 bytes =2308... do it.

							Now the rest is tough, need to make sure that
							don't use more then 16 stuffing bytes.  The
							amount of stuffing used is based on the AU
							type found in the data.  The smallest
							even byte padding stream is 8 bytes!
							Currently the worse case is that the amount
							of stuffing bytes = 11, this leaves 5 to
							play with.  (outvideopacket could be modified
							to leave 7 stuffing bytes avail if the
							STDbuffer/size are coded in...).  So from
							this the following table arises

		2285-2289			Just like 2290. Since there are enough stuffing
							bytes avail to fill out the video packet.

		2284				Send out a vid packet that totals 2284 + 16 = 2300
							(this says that the outvideopacket code will
							trim off the 2 bytes that is normally pads for the
							STDbuffer/size field).  An 8 byte padding packet
							can then be sent!

		2277-2283			Same as 2284, there are again enough stuffing
							bytes to prevent the 16 byte stuff max.

		0-2276				Okay the only trick here is to make sure
							every packet stays even byte aligned.  Again
							a maximum of 5 bytes for stuffing is allowed.
							( the modulo 2 just adds one to odd bytesread)
							Vidsize = bytesread + bytesread%2 + 18
							paddsize = 2308 - Vidsize

		*/
		/* for now I'll keep this hardcoded, don't know how to tie them
		to the user provided parmeters... in any case this is only
		handling a proper EOF on the stream. */
		if (bytesread > 2290)
			bytesread = 2290; /* drop the uncodeable bytes... this may 
								cause problems */

		if (bytesread >= 2285)
			{
			outvideopacket(2308,bytesread,UserParams,CalcParams,buf);
			}
		else
			{
			if (bytesread >= 2277)
				{
				outvideopacket(2300,bytesread,UserParams,CalcParams,buf);
				outputpadding(8,UserParams,CalcParams);
				}
			else
				{
				int vidsize, paddsize;
				vidsize = bytesread + bytesread%2 + 18;
				paddsize = 2308 - vidsize;
				outvideopacket(vidsize,bytesread,UserParams,CalcParams,buf);
				outputpadding(paddsize,UserParams,CalcParams);
				}
			}

		/* Output an ISO-end code, only if CDI/VCD otherwise it is done elsewhere */
		if ((UserParams->CDI_type) && (!UserParams->runforever))
			outbits(ISOENDCODE,0,NULL,32,31);
		else
			outbits(0L,0,NULL,32,31);

		}


	}/* end cdivideopacket */

/*
{*NAME*}
	outvideopacket -- Actual output routine for CDI video packets

{*SYNOPSIS*}
	static int outvideopacket(int desiredsize,  \* size of packet to output *\
		int bytesread, \* bytes read in input buffer *\
		struct inparams *UserParams,
		struct calcparams *CalcParams, 
		char *buf) \* input buffer associated with bytesread *\

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}
	The actual work routine for outputting a video packet.

{*ALGORITHM*}
	Output the header stuff.
	Count the number of AU within the video data read
	Determine the picture type of the first AU within the packet,
		this is used to determine the amount to increment the PTS/DTS
		values, along with wether or not to include them in the packet
		(rules defined within the spec).
	Output any necessary padding and PTS/DTS values
	Output the video data.

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}
	HEURIS Logic -- Brian Quandt

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
static int outvideopacket(int desiredsize,  /* size of packet to output */
	int bytesread, /* bytes read in input buffer */
	struct inparams *UserParams,
	struct calcparams *CalcParams, 
	char *buf) /* input buffer associated with bytesread */
	{
	/* 	WARNING THIS MODULE IS VERY CLOSE COUPLES WITH cdivideopacket, sorry */
	static int firstpicture = 1;	/* if set then have not read the
									first picture */
	char *AUptr;	/* pointer to a found access unit */
	int tmp,
		bytecount,
		numtostuff,		/* number of stuffing bytes required */
		picttype,
		picsinbuf,		/* controls how much to update timestamps */
		i;
	long	delay;	/* number of clock periods to hold a AU before presenting 
					to user, only valid for I and P frames */

	outbits(PACKETSTARTCODE,0,NULL,24,23);
	outbits(UserParams->videostreamid,0,NULL,8,7);

	/* the packet length is fixed in all cases, stuffing bytes are use to
	accomplish this.  The lenght is that following this field (so -6)
	from total  */
	tmp = desiredsize - 6;
	outbits(tmp, 0,NULL,16,15);

	/* Look into the read in buffer, if there is an AU within this */
	/* buffer then we need to output a PTS value that corresponds to it */
	/* Also need to add the necessary byte stuffing */

	/* this used to be a call to strstrn... */

	AUptr = countPICTS(UserParams,CalcParams,
		buf,bytesread,VIDEOAU,SIZEOFVIDEOAU,&picsinbuf);

	if (AUptr != NULL)
		picttype = getpicturetype(AUptr);
	else
		{
		picttype = NOTPICT;	/* not a picture */
		}


	if ( (picttype == IFRAME)||
		(picttype == BFRAME) ||
		(picttype == PFRAME) )
		{
		/* Now determine what type of picture this is, if I or P then */
		/* need to encode both PTS and DTS */
		/* otherwise only need to one of them (PTS).*/
		/* AUptr points to the beginning of this thing so */
		switch(picttype)
			{
			case IFRAME:	/* I frame */
#ifdef DEBUG
xuxa_log("\nIFRAME\n");
#endif
			case PFRAME:	/* P frame */
#ifdef DEBUG
xuxa_log("\nPFRAME\n");
#endif
				/* First output the stuffing bytes */
				/* since both PTS and DTS go out (10 bytes) */
				/* the header is always 18 bytes */
				/* start + id = 3 bytes + length +5(pts) +5(dts) =16 */
				/* So need to stuff desiredsize - (16 + bytesread) */

				tmp = desiredsize - (16 +2+ bytesread);
				if (tmp > 16)
					xuxa_log("Video stuffing more then 16 bytes!\n");
				for(i=0;i<tmp;i++)
					{
					outbits(STUFFINGBYTE,0,NULL,8,7);
					}

				/* STD buffer scale and size  */
				outbits(0x01L,0,NULL,2,1);
				outbits(0x01L,0,NULL,1,0);	
				outbits(46L,0,NULL,13,12);

				/* Now the PTS and DTS values */
				if (!firstpicture)
					{
					delay = CalcParams->videoPTS +CalcParams->videoPTSdelay;
					}
				else
					{
					delay = CalcParams->videoPTS + CalcParams->videoPTSconst;
					}
#ifdef DEBUG
/* 	              SCR     PTS     DTS */
if (getpicturetype(AUptr) == IFRAME)
	{
	sprintf(tas,"\nI\t%10lu\t%10lu\t%10lu\n",
		CalcParams->SCR,delay,(long)CalcParams->videoPTS);
	xuxa_log(tas);
	}
else
	{
	sprintf(tas,"\nP\t%10lu\t%10lu\t%10lu\n",
		CalcParams->SCR,delay,(long)CalcParams->videoPTS);
	xuxa_log(tas);
	}
#endif

				if ((delay < CalcParams->SCR)||(CalcParams->videoPTS < CalcParams->SCR))
					{
					/* Sanity check type stuff, if this happens well its probably a 
					software bug or math error...? */
					if (delay < CalcParams->SCR)
						{
						sprintf(tas,"Oops SCR %d already greater then video PTS %d (%d)\n", CalcParams->SCR,delay,CalcParams->SCR-delay);
						xuxa_log(tas);
						}
					if (CalcParams->videoPTS < CalcParams->SCR)
						{
						sprintf(tas,"Oops SCR %d already greater then video DTS %d (%d)\n", CalcParams->SCR,(long)CalcParams->videoPTS,CalcParams->SCR - (long)CalcParams->videoPTS);
						xuxa_log(tas);
						}
					}

				outbits(0x03L,0,NULL,4,3);	/* output the first markers */
				outbits(0x00L,0,NULL,1,0);	/* this is actually the first
					PTS bit, but since I don't have 33 bit ints... */
				outbits(delay,0,NULL,2,31);
				outbits(MARKERBIT,0,NULL,1,0);
				outbits(delay,0,NULL,15,29);
				outbits(MARKERBIT,0,NULL,1,0);
				outbits(delay,0,NULL,15,14);
				outbits(MARKERBIT,0,NULL,1,0);

				/* The DTS values */
				outbits(0x01L,0,NULL,4,3);	/* output the first markers */
				outbits(0x00L,0,NULL,1,0);	/* this is actually the first
					PTS bit, but since I don't have 33 bit ints... */
				outbits((long)CalcParams->videoPTS,0,NULL,2,31);
				outbits(MARKERBIT,0,NULL,1,0);
				outbits((long)CalcParams->videoPTS,0,NULL,15,29);
				outbits(MARKERBIT,0,NULL,1,0);
				outbits((long)CalcParams->videoPTS,0,NULL,15,14);
				outbits(MARKERBIT,0,NULL,1,0);

				/* Now the actual data */
				outbits(0L,1,buf,bytesread*8,7);

				break;

			case BFRAME:	/* B frame */
#ifdef DEBUG
xuxa_log("\nBFRAME\n");
#endif
				/* First output the stuffing bytes */
				/* since only PTS is sent out (5 bytes) */
				/* okay send out the stuffing bytes */
				/* Here the header still need to be 18 bytes */
				/* so 3 + 1 + 2 + 5 -> add 7 bytes to stuffing */
				tmp = desiredsize - (bytesread + 11 +2);
				if (tmp > 16)
					xuxa_log("Video stuffing more then 16 bytes!\n");
				for(i=0;i<tmp;i++)
					{
					outbits(STUFFINGBYTE,0,NULL,8,7);
					}

				/* STD buffer scale and size */
				outbits(0x01L,0,NULL,2,1);
				outbits(0x01L,0,NULL,1,0);	
				outbits(46L,0,NULL,13,12);

#ifdef DEBUG
/* 	              SCR     PTS     DTS */
sprintf(tas,"\nB\t%10lu\t%10lu\t%10lu\n",
	CalcParams->SCR,(long)CalcParams->videoPTS,(long)CalcParams->videoPTS);
xuxa_log(tas);
#endif
				if (CalcParams->videoPTS < CalcParams->SCR)
					{
					sprintf(tas,"Oops SCR %d already greater then video PTS %d (%d)\n", CalcParams->SCR,(long)CalcParams->videoPTS,CalcParams->SCR - (long)CalcParams->videoPTS);
					xuxa_log(tas);
					}

				/* PTS = DTS since its a B picture */
				outbits(0x02L,0,NULL,4,3);	/* output the first markers */
				outbits(0x00L,0,NULL,1,0);	/* this is actually the first
					PTS bit, but since I don't have 33 bit ints... */
				outbits((long)CalcParams->videoPTS,0,NULL,2,31);
				outbits(MARKERBIT,0,NULL,1,0);
				outbits((long)CalcParams->videoPTS,0,NULL,15,29);
				outbits(MARKERBIT,0,NULL,1,0);
				outbits((long)CalcParams->videoPTS,0,NULL,15,14);
				outbits(MARKERBIT,0,NULL,1,0);

				/* Now the actual data */
				outbits(0L,1,buf,bytesread*8,7);

				break;

			default:
				for (i=-10;i<50;i++)
					{
					sprintf(tas,"%2.2x ",0x00FF & AUptr[i]);
					xuxa_log(tas);
					}
				xuxa_log("this does not look like a picture !!!\n");
				exitroutine();

			}
		firstpicture = 0;	/* Then have read the very first AU/picture */

		}
	else
		{
#ifdef DEBUG
xuxa_log("\nNO AU\n");
#endif
		/* No AU found within packet just output the marker and the data */
		/* Again the header must be 18 bytes */
		/* so this implies add 11 stuffing bytes */
		tmp = desiredsize - (bytesread + 7 +2);
		if (tmp > 16)
			xuxa_log("Video code stuffing more then 16 bytes!\n");
		for(i=0;i<tmp;i++)
			{
			outbits(STUFFINGBYTE,0,NULL,8,7);
			}

		/* STD buffer scale and size  */
		outbits(0x01L,0,NULL,2,1);
		outbits(0x01L,0,NULL,1,0);	
		outbits(46L,0,NULL,13,12);

		outbits(0x0FL,0,NULL,8,7);	/* else clause in mpeg
										spec (not PTS or DTS) */
		outbits(0L,1,buf,bytesread*8,7);
		}

	/* Update the next PTS value based on how many AU's are within this
	current buffer */
	CalcParams->videoPTS = CalcParams->videoPTS + 
		CalcParams->videoPTSconst* picsinbuf;
#ifdef DEBUG
sprintf(tas,"Video AU's is %d\n",picsinbuf);
xuxa_log(tas);
#endif
	}

/*
{*NAME*}
	cdi_aud_sect0 -- Output a CDI audio sector 0.

{*SYNOPSIS*}
	int cdi_aud_sect0(struct inparams *UserParams,
		struct calcparams *CalcParams)

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}
	Output the first sector of a VIDEO CD compliant mpeg system stream.
	VCD has the oddity of wanting a seperate system header for both
	audio and video, this one takes care of the audio system header.

{*ALGORITHM*}
	Output the pack header
	Output the system header as defined for audio in the VCD spec
	Output a padding stream to fill in the rest of the sector.

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}
	HEURIS Logic -- Brian Quandt

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
int cdi_aud_sect0(struct inparams *UserParams,
	struct calcparams *CalcParams)
	{
	/* Output CDI compliant mpeg audio sector with SYSTEM header */

	int i;

#ifdef DEBUG
	xuxa_log("\ncdi_aud_sect0\n");
#endif

	outbits(PACKSTARTCODE,0,NULL,32L,31L);		/* pack start */
	outbits(0x02L,0,NULL,4,3);					/* 0010 */


	/* SCR value */
	outbits(0x00L,0,NULL,1,0);	/* this is actually the first SCR bit, but 
						since I don't have 33 bit ints... */
	outbits(CalcParams->SCR,0,NULL,2,31);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,29);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,14);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(MARKERBIT,0,NULL,1,0);

	/* MUXRATE */
	outbits(UserParams->muxrate,0,NULL,22,21);	/* muxrate */
	outbits(MARKERBIT,0,NULL,1,0);

	/* SYSTEM header as stated in the white book figure IV.10 */
	outbits(SYSTEMHEADERSTARTCODE,0,NULL,32,31);
	outbits(9L,0,NULL,16,15);		/* system header length's always fixed, 
									since only have 1 only and 1 video) */
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(UserParams->muxrate,0,NULL,22,21);	/* rate bound */
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(1L,0,NULL,6,5);			/* audiobound */
	outbits(UserParams->fixed_flag,0,NULL,1,0);		/* fixed_flag */
	outbits(UserParams->CSPS_flag,0,NULL,1,0);		/* CSPS */
	outbits(UserParams->audio_lock,0,NULL,1,0);		/* audio_lock */
	outbits(UserParams->video_lock,0,NULL,1,0);		/* video_lock */

	outbits(MARKERBIT,0,NULL,1,0);
	outbits(0L,0,NULL,5,4);		/* video bound */
	outbits(0xFFL,0,NULL,8,7);	/* reserved byte */

	/* the std bounds for audio */
	outbits(UserParams->audiostreamid,0,NULL,8,7);	/* audio stream id */
	outbits(0x03L,0,NULL,2,1);
	outbits(0x00L,0,NULL,1,0);	/* STD buffer scale, indicates scale of video */
	outbits(32L,0,NULL,13,12);	/* CSPS is set this in constrained to 4k */

	/* Padding packet of 2277 bytes to fill out the rest of the sector */
	outputpadding(2277L, UserParams, CalcParams);

	/* Now the zeros field, whatever the hell they are... they sure
	aren't MPEG!  Twenty of them */
	if (UserParams->VCD_nulls)
		for (i=0;i<20;i++)
			outbits(0x00L,0,NULL,8,7);
	
	}

/*
{*NAME*}
	cdi_aud_sect -- Output CDI compliant audio sectors, other then first one.

{*SYNOPSIS*}
	int cdi_aud_sect(struct inparams *UserParams,
		struct calcparams *CalcParams)

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}
	Create a systems pack that contains audio data.

{*ALGORITHM*}
	Output a pack header
	Go and output the audio data
	If VCD type stuff pad the end of the thing with 20 NULLS.

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}
	HEURIS Logic -- Brian Quandt

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
int cdi_aud_sect(struct inparams *UserParams,
	struct calcparams *CalcParams)
	{
	/* Output CDI compliant mpeg audio sector (without SYSTEM header) */

	int i;

#ifdef DEBUG
	xuxa_log("\ncdi_aud_sect\n");
#endif


	outbits(PACKSTARTCODE,0,NULL,32L,31L);		/* pack start */
	outbits(0x02L,0,NULL,4,3);					/* 0010 */


	/* SCR value */
	outbits(0x00L,0,NULL,1,0);	/* this is actually the first SCR bit, 
								but since I don't have 33 bit ints... */
	outbits(CalcParams->SCR,0,NULL,2,31);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,29);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,14);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(MARKERBIT,0,NULL,1,0);

	/* MUXRATE */
	outbits(UserParams->muxrate,0,NULL,22,21);	/* muxrate */
	outbits(MARKERBIT,0,NULL,1,0);


	/* Now the actual audio data */
	cdiaudiopacket( UserParams, CalcParams);

	/* Now the zeros field, whatever the hell they are... they sure
	aren't MPEG!  Twenty of them, might as well be a million...
	I believe this can just be ignored in all processing even 
	the effect it has on the buffering? */

	if (UserParams->VCD_nulls)
		{
		for (i=0;i<20;i++)
			outbits(0x00L,0,NULL,8,7);
		}
	}

/*
{*NAME*}
	cdiaudiopacket -- Output a CDI compliant audio packet.

{*SYNOPSIS*}
	static int cdiaudiopacket(struct inparams *UserParams,
		struct calcparams *CalcParams)

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}
	First level routine to actaully create a audiopacket.

{*ALGORITHM*}
	Read the data from the audio stream.
	Based on the amount read go and output it.
	If at EOF output a systems end code.

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}
	HEURIS Logic -- Brian Quandt

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
static int cdiaudiopacket(struct inparams *UserParams,
		struct calcparams *CalcParams)
	{
	unsigned char buf[4048];	/* just allocated enough for now, avoid 
							dynamic allocation for potential real time app */
	int bytecount,
		numtostuff,		/* number of stuffing bytes required */
		bytesread,		/* number of bytes read from input */
		tmp,
		i;

#ifdef DEBUG
xuxa_log("\naudiopacket\n");
#endif


	/* Try to read in enough data for one audio packet.  If I can't then
	add padding packets to fill out this pack with another padding packet.
	(This implies that partial audio packets need another padding packet
	right next to it) */
	if (!feof(UserParams->audio)&&
		(bytesread = fread(buf,1,UserParams->auddatasize,UserParams->audio)) == UserParams->auddatasize)
		{
		if(UserParams->CDI_type)
			outaudiopacket(UserParams->audpacketsize-20,bytesread,UserParams,CalcParams,buf);
		else
			outaudiopacket(UserParams->audpacketsize,bytesread,UserParams,CalcParams,buf);

		}
	else
		{
#if 1
		if (UserParams->CDI_type)
			{
			/* End the video packs */
			outputpadding(UserParams->vidpacketsize-4,UserParams,CalcParams);
			outbits(ISOENDCODE,0,NULL,32,31);

			/* End the audio packs */
			outputpadding(2300,UserParams,CalcParams);
			outbits(ISOENDCODE,0,NULL,32,31);
			if (UserParams->VCD_nulls)
				for (i=0;i<20;i++)
					outbits(0x00L,0,NULL,8,7);

			}
		else
			{
			outputpadding(UserParams->audpacketsize,UserParams,CalcParams);
			}

		return(0);
#endif


		/*  So have read < 2279, the problem here is that an ISOEND
		code must be tagged to this 'partial packet', and we need
		even byte boundaries (white book 3.2.4).
		So here are some senario's:
			Bytesread		What to do
			2276-2278       Don't know, can't fit an end code.  And the
							white book implies only that in sectors
							in which an end code show up can I exceed
							13 bytes for the packet headers (stuffing bytes)
							For now I think I will just send 2275 bytes and then
							and endcode... this probably will cut off any
							audio endcode but at least there's an system
							endcode that I'm tacking on.

			2275			Output the packet, then send the endcode 
							(2275+4 = 2279 + 13 = 2292)

			2259-2274		Output the packet as 2275, the output routine will 
							stuff bytes.  Then just send the endcode.  This 
							range takes into consideration the maximun stuffing
							allowed by MPEG1

			0-2258			Audio packet size = bytesread + 
							13 + (15 - bytesread%16)
							padding packet size = 2288 - audio packet size
							This math will result in at most two packets,
							one audio the other padding that are both even bytes
							aligned and can then be output.  The remaining 4 
							bytes are used for the ISOendcode.  Oh, the only 
							reason that padding packets are done here is that
							I can now do them since 2288-bytesread now gives
							me enough space to create a padding packet
		*/
		/* Again I will leave this alone for now.  Worse case is that
		I have mucked up the EOF of a stream (hardware hopefully will recover) */
		if (bytesread > 2275)
			{
			bytesread = 2275;	/* This is the make shift thing that drops the 
						extra audio data since I can't do anything with it */
			xuxa_log("Dropping some audio...\n");
			}

		if (bytesread >= 2259)
			{
			/* Just the audio packet is sent, the end code is followed
			directly by this packet */
			outaudiopacket(2288,bytesread,UserParams,CalcParams,buf);
			}
		else
			{
			/* Both an audio packet and a padding packet is sent, again
			both must align to even byte boundaries */
			int audsize, paddsize;
			audsize = bytesread + 13 + (15 - bytesread%16);
			paddsize = 2288 - audsize;
			outaudiopacket(audsize,bytesread,UserParams,CalcParams,buf);
			outputpadding(paddsize,UserParams,CalcParams);
			}

		/* Output an ISO-end code */
		/* only if cdi... otherwise it is handled in the main loop */
		if ((UserParams->CDI_type) && (!UserParams->runforever))
			outbits(ISOENDCODE,0,NULL,32,31);
		else
			{
			/* this is moderately safe... since a decoder should ignore
			these leading things anyways */
			outbits(0L,0,NULL,32,31); 
			}

		}


	}



/*
{*NAME*}
	outaudiopacket -- Acutal output routine for CDI compliant audio packets

{*SYNOPSIS*}
	static int outaudiopacket(int desiredsize,  \* acutal size of packet *\
		int bytesread, 	\* bytes available in input buffer, audio bytes *\
		struct inparams *UserParams,
		struct calcparams *CalcParams, 
		char *buf)	\* acutal audio data read, associated with bytes read*\

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}
	Actual work routine for outputting a audio packet (ie one that
	contains audio data).

{*ALGORITHM*}
	Output the packet headers
	Determine the number of AU's within the audio data.
	Increment the PTS values for audio based on the number of AU's contained.
	Add any stuffing necessary to fill out the packet to the size required.
	Ouput the PTS values (audio has no DTS values).
	Output the data.

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}
	HEURIS Logic -- Brian Quandt

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
static int outaudiopacket(int desiredsize,  /* acutal size of packet */
	int bytesread, 	/* bytes available in input buffer, audio bytes */
	struct inparams *UserParams,
	struct calcparams *CalcParams, 
	char *buf)	/* acutal audio data read, associated with bytes read*/
	{
	char *AUptr;
	int 	tmp,i;

	outbits(PACKETSTARTCODE,0,NULL,24,23);
	outbits(UserParams->audiostreamid,0,NULL,8,7);

	/* THE NORMAL case:
	In the audio packets that take up the whole sector the size is
	fixed to be 2292 bytes/packet - 6 = 2286, the 6 comes from the
	startcodelen + idlen + packetlenlen.  Actually I use the desired
	size to control the size  */

	tmp = desiredsize - 6;

	outbits( tmp, 0,NULL,16,15);

	/* Look into the read in buffer, if there is an AU within this */
	/* buffer then we need to output a PTS value that corresponds to it */
	/* Also need to add the necessary byte stuffing */
if (1)
		{
		/* Stuff bytes to make up the difference in the amount of data
		supplied and the amount to send out. WARNING there is a maximum
		allowed number of stuffing bytes allowed.  A serious warning message
		appears.  What to do about it I don't know.  But maybe try to
		touch up the size of the audio file, since this seems to
		be a screw up in the white book specification */
		tmp = desiredsize - (6+5+2+bytesread);
		if (tmp > 16)
			xuxa_log("Audio code is stuffing more then 16 bytes!  MPEG1 violation!\n");
		for(i=0;i<tmp;i++)
			outbits(STUFFINGBYTE,0,NULL,8,7);

		/* STD buffer scale and size */
		outbits(0x01L,0,NULL,2,1);
		outbits(0x00L,0,NULL,1,0);	
		outbits(32L,0,NULL,13,12);


#ifdef DEBUG
sprintf(tas,"\nA\t%10lu\t%10lu\t%10lu\n",
	CalcParams->SCR,(long)CalcParams->audioPTS,(long)CalcParams->audioPTS);
xuxa_log(tas);
#endif

		if (CalcParams->audioPTS < CalcParams->SCR)
			{
			sprintf(tas,"Oops SCR %d already greater then audio PTS %d (%d) \n", CalcParams->SCR,(long)CalcParams->audioPTS, CalcParams->SCR - (long)CalcParams->audioPTS);
			xuxa_log(tas);
			}

		outbits(0x02L,0,NULL,4,3);	/* output the first markers */
		outbits(0x00L,0,NULL,1,0);	/* this is actually the first PTS bit, 
									but since I don't have 33 bit ints... */
		outbits((long)CalcParams->audioPTS,0,NULL,2,31);
		outbits(MARKERBIT,0,NULL,1,0);
		outbits((long)CalcParams->audioPTS,0,NULL,15,29);
		outbits(MARKERBIT,0,NULL,1,0);
		outbits((long)CalcParams->audioPTS,0,NULL,15,14);
		outbits(MARKERBIT,0,NULL,1,0);

		/* Update the next audioPTS value */
		CalcParams->audioPTS = CalcParams->audioPTS + 
			CalcParams->audioPTSconst * 
			(countAudioAUs(UserParams,CalcParams,
				buf,bytesread,AUDIOAU,SIZEOFAUDIOAU));

#ifdef DEBUG
sprintf(tas,"Audio AU's is %d\n", 
countAudioAUs(UserParams,CalcParams, buf,bytesread,AUDIOAU,SIZEOFAUDIOAU));
xuxa_log(tas);
#endif
		}
	else
		{
		/* No AU found within buffer So don't need a PTS Value */

#ifdef DEBUG
xuxa_log("\nAUDIO no AU\n");
#endif

		/* Stuff bytes to make up the difference in the amount of data
		supplied and the amount to send out. Also take into consideration
		that PTS values are not sent
		WARNING there is a maximum
		allowed number of stuffing bytes allowed.  A serious
		warning message results... */
		tmp = desiredsize - (6+2+bytesread+1);
		if (tmp > 16)
			xuxa_log("Audio code is stuffing more then 16 bytes!  MPEG1 violation!\n");
		for(i=0;i<tmp;i++)
			outbits(STUFFINGBYTE,0,NULL,8,7);

		/* okay send out the stuffing bytes */
		/* since each packet header must be 13 bytes and a PTS
		takes up 5 bytes ( startcode + id + len + stuff + 2 + 1 = 13 )
		=> stuff = 4 */
		for(i=0;i<4;i++)
			{
			outbits(STUFFINGBYTE,0,NULL,8,7);
			}

		/* If odd packet length stuff to even */
		if ((desiredsize % 2) == 1)
			outbits(STUFFINGBYTE,0,NULL,8,7);

		/* STD buffer scale and size */
		outbits(0x01L,0,NULL,2,1);
		outbits(0x00L,0,NULL,1,0);	/*STD buffer scale */
		outbits(32L,0,NULL,13,12);	/*STD buffer size */

		/* output the 0000 1111 byte since no PTS values*/
		outbits(0x0FL,0,NULL,8,7);

		}

	/* Now can output the actual data */
	outbits(0L,1,buf,bytesread*8,7);
	}



/*
{*NAME*}
	countPICTS -- Given a buffer count the number of pictures within it.

{*SYNOPSIS*}
	char *countPICTS( struct inparams *UserParams,
				struct calcparams *CalcParams,
				char *buf,	\* Buffer to search *\
				int sizebuf,	\* Size of buffer *\
				char *token,	\* token to find in buf *\
				int sizetok,		\* size of token to search for *\
				int *numpics)	\* number of pictures within buffer *\

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}
	The primary purpose of this routine is to determine how many
	access units occur within a given buffer.  This can then be
	used to determine the amount to increase the PTS/DTS values
	for later time stamps (ie if there are two pictures within
	this buffer then the next timestamp must be two pictures
	later...).

{*ALGORITHM*}
	Zip through the entire buffer, don't forget that according to the
	definition a AU occurs if the first byte is contained within the
	buffer (so need to do some lookahead). This would normally take
	only 3 extra bytes of lookahead but since for confirmation I 
	need to check the picture type I really need 5 bytes (picture type
	fall within bytes 5 after a startcode).

{*SEE ALSO*}

{*BUGS*}
	This routine assumes a minimum size to allow one token to be present.
	At this point in time this is set to 200 bytes.  I do this to make
	life a bit easier on programming (ie quick hack).  Plus this should
	be reasonable assuming the normal MPEG-1 streams that are processed.

{*CREDITS*}
	HEURIS Logic -- Brian Quandt

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
char *countPICTS( struct inparams *UserParams,
			struct calcparams *CalcParams,
			char *buf,	/* Buffer to search */
			int sizebuf,	/* Size of buffer */
			char *token,	/* token to find in buf */
			int sizetok,		/* size of token to search for */
			int *numpics)	/* number of pictures within buffer */
	{
	char *firstpict = NULL;
	char *ptr;
	int ret = 0,
		i,j,pictype;

	
	*numpics = 0;

	/* Add 3 bytes so that can detect boundary condition for buffer */
	/* Okay the code really adds 5, need the other 2 bytes so that
	I can determine the picture type, otherwise I will have notread
	the bytes containing the picture type... this only applies
	when the picture start code spans the buffer boundary, with 1 byte
	in the first and 3 bytes in the second, I don't modify the
	search range since still only need to search until the 3 bytes - tok size */
	if (!feof(UserParams->video)&&
		fread(&buf[sizebuf],1,5,UserParams->video) == 5)
		sizebuf += 3;	/* pass by value so can modify it here (temporarily) */

	for(i=0;i<=sizebuf-sizetok;i++)
		{
		if (buf[i] == token[0])
			if (buf[i+1] == token[1])
				if (buf[i+2] == token[2])
					if (buf[i+3] == token[3])
						{
						pictype = getpicturetype(&buf[i]);
						if ((pictype == IFRAME) |
							(pictype == PFRAME) |
							(pictype == BFRAME))
							{
							(*numpics)++;
							if (firstpict == NULL)
								firstpict = &(buf[i]);	
							}
						else
							{
							/* ERROR condition */
							/* not conviced that this is an error bq 6/14*/
							int k;
							sprintf(tas,"PICTURESTART code emulation\n");
							xuxa_log(tas);
							sprintf(tas,"%2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n\n",
							0x00ff&buf[i],0x00ff&buf[i+1],0x00ff&buf[i+2],
							0x00ff&buf[i+3],0x00ff&buf[i+4],0x00ff&buf[i+5]);
							xuxa_log(tas);

							for(k=0;k<sizebuf;k++)
								{
								sprintf(tas,"%2.2x ",0x00ff&buf[k]);
								xuxa_log(tas);
								if (k%20 == 0)
									xuxa_log("\n");
								}
							xuxa_log("\n");
							}
						}
		}
	/* remove the boundary bytes from the search buffer */
	/* again remember actually read 5 extra bytes */
	if (!feof(UserParams->video)&&fseek(UserParams->video,-5,1)!=0)
		xuxa_log("Error in fseeking back 5 bytes\n");
	
	return(firstpict);
	}


/*
{*NAME*}
	countAudioAUs -- Count audio AU's within a given audio buffer.

{*SYNOPSIS*}
	int countAudioAUs( struct inparams *UserParams,
		struct calcparams *CalcParams,
		unsigned char *buf,	\* Buffer to search *\
		int sizebuf,	\* Size of buffer *\
		char *token,	\* token to find in buf *\
		int sizetok)		\* size of token to search for *\

{*CLASS*}
	c:mpeg

{*DESCRIPTION*}
	This one does the same as countPICTS but for a audio buffer,
	the primary difference is lack of checking for the picture type.

{*ALGORITHM*}
	Similar to countPICTS

{*SEE ALSO*}

{*BUGS*}

{*CREDITS*}
	HEURIS Logic -- Brian Quandt

{*COPYRIGHT*}
  	Copyright 1994 Heuris Logic.
  	All rights reserved.
{**}
*/
int countAudioAUs( struct inparams *UserParams,
	struct calcparams *CalcParams,
	unsigned char *buf,	/* Buffer to search */
			int sizebuf,	/* Size of buffer */
			char *token,	/* token to find in buf */
			int sizetok)		/* size of token to search for */
	{
	unsigned char *ptr;
	int ret = 0,
		i,j,
		numAUs;
	
	numAUs = 0;

	/* Add 3 more bytes from the input stream to the search buffer this
	is to accomodate AU's that may span the boundary */
	if (!feof(UserParams->audio)&&
		fread(&buf[sizebuf],1,3,UserParams->audio) == 3)
		sizebuf += 3;	/* pass by value so can modify it here (temporarily) */


	for(i=0;i<=sizebuf-sizetok;i++)
		{
		/* Okay so I ignore the size of the token and the acutal value,
		this needs to be changed to be modifiable via the proper
		defines, but now this was just a quick hack (as always) */
		/* the pattern that we need to match for an audio AU in the
		case of layer II is  1111 1111 1111 110x  layer I audio
		is of course allowed but not for this CDI type coding */
		if (buf[i] == 0xFF)
			if ((buf[i+1] & 0xFE) == 0xFC)
					{
					numAUs++;
					/* okay this is a hack... to avoid recognizing an AU */
					/* within the subband allocation and scale tables */
					/* just add enough to the counters to skip over them */
					/* I'm hoping I will not see them also within the */
					/* sample data... */
					/* okay advance by what is known as a minimal size to the AU */
					i += 100;
					}
		}

	/* remove the boundary bytes from the search buffer */
	if (!feof(UserParams->audio)&&fseek(UserParams->audio,-3,1)!=0)
		xuxa_log("Error in fseeking back 3 bytes\n");

	return(numAUs);
	}


int cdi_pad_sect(struct inparams *UserParams,
	struct calcparams *CalcParams)
	{

#ifdef DEBUG
	xuxa_log("\ncdi_pad_sect\n");
#endif


	outbits(PACKSTARTCODE,0,NULL,32L,31L);		/* pack start */
	outbits(0x02L,0,NULL,4,3);					/* 0010 */


	/* SCR value */
	outbits(0x00L,0,NULL,1,0);	/* this is actually the first SCR bit, 
								but since I don't have 33 bit ints... */
	outbits(CalcParams->SCR,0,NULL,2,31);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,29);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,14);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(MARKERBIT,0,NULL,1,0);

	/* MUXRATE */
	outbits(UserParams->muxrate,0,NULL,22,21);	/* muxrate */
	outbits(MARKERBIT,0,NULL,1,0);

	/* I use video packet size here, just becuase I think I will then be CDI tolerable?? */
	outputpadding(UserParams->vidpacketsize,UserParams,CalcParams);
	}


/* routine to put out a padding pack with end-code at end of it */
/* this seems to be better then just tagging a end-code onto end of system */
/* stream, at least the philips verifier wants it this way */
int cdi_pad_sect_endcode(struct inparams *UserParams,
	struct calcparams *CalcParams)
	{

#ifdef DEBUG
	xuxa_log("\ncdi_pad_sect\n");
#endif


	outbits(PACKSTARTCODE,0,NULL,32L,31L);		/* pack start */
	outbits(0x02L,0,NULL,4,3);					/* 0010 */


	/* SCR value */
	outbits(0x00L,0,NULL,1,0);	/* this is actually the first SCR bit, 
								but since I don't have 33 bit ints... */
	outbits(CalcParams->SCR,0,NULL,2,31);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,29);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,14);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(MARKERBIT,0,NULL,1,0);

	/* MUXRATE */
	outbits(UserParams->muxrate,0,NULL,22,21);	/* muxrate */
	outbits(MARKERBIT,0,NULL,1,0);

	/* I use video packet size here, just becuase I think I will then be CDI tolerable?? */
	outputpadding(UserParams->vidpacketsize-4,UserParams,CalcParams);

	/* output the end code */
	outbits(ISOENDCODE,0,NULL,32,31);
	}

int nextsector(struct inparams *UserParams)
	{
	UserParams->maxsectors--;

	if (UserParams->runforever)
		return(1);

	if (UserParams->maxsectors)
		{
		if (feof(UserParams->video))
			xuxa_log("eof video\n");
		if (feof(UserParams->audio))
			xuxa_log("eof audio\n");

		if (feof(UserParams->video) || feof(UserParams->audio))
			return(0);
		else
			return(1);
		}
	else
		return(0);

	}


/* hack code to handle the problem of needing to generate a pack */
/* with 2 packets, ie a packet less then */
cdi_aud_sect_emergency(struct inparams *UserParams,
	struct calcparams *CalcParams,
	float *ABV)
	{
	/* Output CDI compliant mpeg audio sector (without SYSTEM header) */
	int i;
	int size;
	int remainder;

#ifdef DEBUG
	xuxa_log("\ncdi_aud_sect\n");
#endif


	outbits(PACKSTARTCODE,0,NULL,32L,31L);		/* pack start */
	outbits(0x02L,0,NULL,4,3);					/* 0010 */


	/* SCR value */
	outbits(0x00L,0,NULL,1,0);	/* this is actually the first SCR bit, 
								but since I don't have 33 bit ints... */
	outbits(CalcParams->SCR,0,NULL,2,31);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,29);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(CalcParams->SCR,0,NULL,15,14);
	outbits(MARKERBIT,0,NULL,1,0);
	outbits(MARKERBIT,0,NULL,1,0);

	/* MUXRATE */
	outbits(UserParams->muxrate,0,NULL,22,21);	/* muxrate */
	outbits(MARKERBIT,0,NULL,1,0);


	/* Calculate the size of this packet */
	size = MAXAUDBUF - *ABV;

	/* size of padding packet to fill sector to follow */
	remainder = UserParams->audpacketsize - size;

	/* take care of problem if padding packet is made too small */
	if ((remainder > 0)&&(remainder < 50))
		{
		size -= 50;
		remainder += 50;
		}

	/* make sure remainder is even number of bytes */
	if (remainder%2 != 0)
		{
		remainder--;
		size++;
		}

	/* limit the size to only output max as set in input params */
	if (remainder < 0)
		{
		cdiaudiopacket( UserParams, CalcParams);
		*ABV += UserParams->auddatasize;
		return(0);
		}

	/* update the ABV, code below will write out data 20 bytes less */
	*ABV += (size -20 );

	/* Now the actual audio data */
	cdiaudiopacket_sized( UserParams, CalcParams, size);

	/* output a padding packet to fill gap */
	if (remainder > 0)
		{
		xuxa_log("padding packet\n");
		outputpadding(remainder,UserParams,CalcParams);
		}

	/* Now the zeros field, whatever the hell they are... they sure
	aren't MPEG!  Twenty of them, might as well be a million...
	I believe this can just be ignored in all processing even 
	the effect it has on the buffering? */

	if (UserParams->VCD_nulls)
		{
		for (i=0;i<20;i++)
			outbits(0x00L,0,NULL,8,7);
		}

	}


cdiaudiopacket_sized(struct inparams *UserParams,
	struct calcparams *CalcParams,
	int size)
	{
	unsigned char buf[4048];	/* just allocated enough for now, avoid 
							dynamic allocation for potential real time app */
	int bytecount,
		numtostuff,		/* number of stuffing bytes required */
		bytesread,		/* number of bytes read from input */
		tmp,
		i;

#ifdef DEBUG
xuxa_log("\naudiopacket\n");
#endif

	/* Try to read in enough data for one audio packet.  If I can't then
	add padding packets to fill out this pack with another padding packet.
	(This implies that partial audio packets need another padding packet
	right next to it) */
	/* the -20 is used to allow enough bytes for packet headers and such */
	if (!feof(UserParams->audio)&&
		(bytesread = fread(buf,1,size-20,UserParams->audio)) == size-20)
		{
		if(UserParams->CDI_type)
			outaudiopacket(size-20,bytesread,UserParams,CalcParams,buf);
		else
			outaudiopacket(size,bytesread,UserParams,CalcParams,buf);

		}
	else
		{
#if 1
		if (UserParams->CDI_type)
			{
			/* End the video packs */
			outputpadding(UserParams->vidpacketsize-4,UserParams,CalcParams);
			outbits(ISOENDCODE,0,NULL,32,31);

			/* End the audio packs */
			outputpadding(2300,UserParams,CalcParams);
			outbits(ISOENDCODE,0,NULL,32,31);
			if (UserParams->VCD_nulls)
				for (i=0;i<20;i++)
					outbits(0x00L,0,NULL,8,7);

			}
		else
			{
			outputpadding(UserParams->audpacketsize,UserParams,CalcParams);
			}

		return(0);
#endif
		}
	}



/* If you want any type of output to the screen, make sure to define DEBUG */
/* Otherwise all info or debug messages are blocked */
xuxa_log(char *s)
	{
#ifdef DEBUG
	fprintf(stderr,s);
#endif
	}
