#include <stdio.h>

#define MAXLINEBUF 150

#define PADDINGSTREAMID 0xBE

#define AUDIOVIDEO 0
#define AUDIOONLY 1
#define VIDEOONLY 2
#define ENDOFINPUT 3
#define NOTPICT -1	/* picture types */
#define IFRAME 0x08	
#define PFRAME 0x10
#define BFRAME 0x18

#define PACKSTARTCODE 0x1BAL
#define MARKERBIT 0x01L
#define PADDINGBYTE 0xFF
#define STUFFINGBYTE 0xFF
#define SYSTEMHEADERSTARTCODE 0x1BBL
#define PACKETSTARTCODE 0x01L
#define ISOENDCODE 0x01B9L

#define MAXSYSBUF (46.0*1024.0 - 0.1)
#define MAXVIDBUF (46.0*1024.0 - 0.1)
#define MAXAUDBUF (4.0*1024.0 - 0.1)
#define MINSYSBUF (0.0 + 0.1)
#define MINVIDBUF (0.0)
#define MINAUDBUF (0.0)


#define CLOCKFREQ 90000.0	/* 90,000 HZ clock */

#define AUDIOAU "\x00\x00\x00\x01"
#define SIZEOFAUDIOAU 4	/* length in bytes */
#define VIDEOAU "\x00\x00\x01\x00"
#define SIZEOFVIDEOAU 4	/* length in bytes */
#define AUDIOSAMPLESPERAU 1152	/* audio samples per AU */
#define VIDEOSAMPLESPERAU 1	/* video samples per AU */

/* defines for pre drain analysis of video/audio files */
#define VIDEO 1234
#define AUDIO 4321


struct inparams
	{
	/* parameters supplied by the user */

	FILE *outfile; /* file pointer open for bitstream output */

	FILE *audio;
	float audiorate;	 /* bits per second */
	unsigned char audiostreamid;
	int audiobound;
	FILE *audioAU;	/* pointer to audio file of last read Access unit */
	long lastaudioSCR;

	FILE *video;
	float videorate; /* bits per second */
	unsigned char videostreamid;
	int videobound;
	FILE *videoAU;	/* pointer to video file of last read Access unit */
	long lastvideoSCR;


	float audiosamplerate;	/* 48000, 44100, 38000, etc */
	int packetheaderlength;
	int packheaderlength;
	int packetlength;
	int video_M;
	float videoframesec;
	float packsize;
	float dsmpacksize;	 /* digital storage media, ie this may be larger  */
						/* then the actual user data, sector headers, etc */
	float vidpacketsize;	/* size of a video packet */ 
	float viddatasize;	/* data contained within a video packet */
	float audpacketsize;	/* size of a audio packet */
	float auddatasize;	/* data contained within an audio packet */
	int packetsperpack;


	int maxsectors;	/* if supplied, the maximum number of sectors to output*/

	int runforever;		/* boolean flag if set continue the encoder 
							even on EOF of all of the inputs, ie use
							padding packets*/
	int muxrate;
	int firstAudioPTS;
	int firstVideoPTS;
	int video_lock;
	int audio_lock;
	int CSPS_flag;
	int fixed_flag;		/* fixed bitrate or not */
	int CDI_type;		/* if set then output CDI type headers and audio padding */
	int VCD_nulls;		/* if set  then include VCD type 20 NULLS after all audio packs */
	};

struct calcparams
	{
	/* calculated parameters from the user supplied ones */
	int Rmux;
	int muxrate;	/* rate at which data arrive to decoder, 50 byte units */
	int ratebound;	 /* just Rmux in 50 byte units, ie, divided by 50 */
	long SCR;		/* this is only 32 bits, that's good enough for now, needs to be 33 for real long sequences */
	int byteSCR;	/* number of bytes output at current SCR value */
	float videoaudioratio;	/* number of video packets output to 1 audio */
	int videostartupdelay;	/* video startup delay */
	int audiostartupdelay;	/* audio startup delay */
	float videoPTSdelay;	/*const value added to I/P frames for reorder delay */
	float videoPTSconst;	/* values added to old PTS value to get new one */
	float audioPTSconst;
	float videoPTS;		/* Last PTS value for respective stream */
	float audioPTS;
	};



int cdi_aud_sect_emergency(struct inparams *,
	struct calcparams *,
	float *);


int newbuffertest(struct inparams *,
	struct calcparams *,
	float *, float *,int );

int testdrain(struct inparams *,
	struct calcparams *, 
	float *, float *,int);
