#include <stdio.h>
#include <cdisys.h>
#include <string.h>

/* temp char for string usage in debug prints */
char tas_drain[256];


/* This routine does a real buffer fill/drain */
drainbuffers(struct inparams *UserParams, 
	struct calcparams *CalcParams,
	float *ABV,	/* current level of the audio verifier buffer */
	float *VBV,	/* currnt level of the video buffer verifier */
	int bufferchoice) /* buffer if any that is being filled */
	{
	int ret;

	/* the actual fill/drain */
	ret = newbuffertest(UserParams,CalcParams,
		ABV, VBV, bufferchoice);

	if (ret != 0)
		{
		xuxa_log("^^^^real buffer error\n");
		}

	}

/* This routine does a non-destructive buffer test of all buffers */
/* that is the buffers and file pointers etc are all reset */
int testdrain( 
	struct inparams *UserParams, 
	struct calcparams *CalcParams,
	float *VBV,
	float *ABV,
	int bufferchoice)
	{
	int ret=0;
	struct inparams uuserparams;
	struct calcparams ccalcparams;
	long aptr, vptr;
	float abv,vbv;


	/* Save the current file pointer positions */
	memcpy(&uuserparams,UserParams,sizeof(struct inparams));
	memcpy(&ccalcparams,CalcParams,sizeof(struct calcparams));
	aptr = ftell(UserParams->audioAU);
	vptr = ftell(UserParams->videoAU);

	/* The the tests */
	ret = newbuffertest(UserParams,CalcParams,
		ABV, VBV, bufferchoice);

	/* restore any screwed up state */
	memcpy(UserParams,&uuserparams,sizeof(struct inparams));
	memcpy(CalcParams,&ccalcparams,sizeof(struct calcparams));
	fseek(UserParams->audioAU ,aptr,SEEK_SET);
	fseek(UserParams->videoAU ,vptr,SEEK_SET);

	return(ret); 
	}


/* BAQ's new stuff to really test the buffering */

int newbuffertest(struct inparams *UserParams,
	struct calcparams *CalcParams,
	float *abv, float *vbv,
	int bufferchoice)
	{
	int i,j;
	float iabv, ivbv;
	int dist;
	int auderr = 0;
	int viderr = 0;
	int begin, end;
	int ret=0;
	int t;


	iabv = *abv;
	ivbv = *vbv;
	
	begin = CalcParams->SCR;
	end = CalcParams->SCR+ UserParams->dsmpacksize/(UserParams->muxrate*50)*CLOCKFREQ;

	sprintf(tas_drain,"%5s [%d:%d] SCR %6d Ascr %6d Vscr %6d ABV %5.1f VBV %5.1f\n",
		bufferchoice==VIDEO?"VIDEO":(bufferchoice==AUDIO?"AUDIO":"PAD"),
			begin,end,
			CalcParams->SCR,
			(long)UserParams->lastaudioSCR,
			(long)UserParams->lastvideoSCR,
			*abv,*vbv);
	xuxa_log(tas_drain);

	/* Loop through each clock and check for overflow underflow */
	for (i=begin,j=0;
		i < end;
		i++,j++)
		{
		if (bufferchoice == AUDIO)
			{
			/* Add 1 clocks worth of audio data to buffer */
			/* Data arrives in the buffer based on muxrate after getting */
			/* to it, ie past all dsm/pack/packet headers */
			t = (((UserParams->dsmpacksize-UserParams->auddatasize+1)/
				(UserParams->muxrate*50))*CLOCKFREQ);
			if (j >= t)
				{
				/* The below equation is the amount that really should
				be added to the buffer (complete accurate) however due
				to discrepancy between iterating in ints (efficiency
				and the fact that the clock is only accurate in these
				units) and floats I use the one right below this one
				to make sure the data is evenly distributed among the
				clocks.  The discrepancy should be very small.  The
				result though after looping through all these clocks
				will be exact (its just that each individual clock
				may deviate slightly from what is stored within muxrate
				*abv += (float)((UserParams->muxrate*50.0)/CLOCKFREQ);
				*/
				*abv +=   (float)UserParams->auddatasize
				                    /
				          (float)((end-begin)-t); /* number clocks actually sending data */

				}
			}

		if (bufferchoice == VIDEO)
			{
			/* Add 1 clocks worth of video data to buffer */
			/* Data arrives in the buffer based on muxrate after getting */
			/* to it, ie past all dsm/pack/packet headers */
			t = (((UserParams->dsmpacksize-UserParams->viddatasize+1)/
				(UserParams->muxrate*50))*CLOCKFREQ);
			if (j >= t)
				{
				/* This the exact same as audio buffer fill see its long
				comment
				*vbv += (float)((UserParams->muxrate*50.0)/CLOCKFREQ);
				*/
				*vbv +=   (float)UserParams->viddatasize
				                    /
				          (float)((end-begin)-t); /* number clocks actually sending data */

				}
			}


		/* Now remove stuff from buffers if proper time */
		if (i == UserParams->lastaudioSCR)
			{
			/* Yank one AU's of audio data from the buffer and set the */
			/* next time to yank another AU */

			dist = bytesbetweenAudioAUs(UserParams->audioAU,UserParams);

			*abv -= dist;

			UserParams->lastaudioSCR += 
				(double)(((double)AUDIOSAMPLESPERAU/UserParams->audiosamplerate)*CLOCKFREQ);
sprintf(tas_drain,"\t\tyank audio %d next %d SCR %d\n",dist,UserParams->lastaudioSCR,i);
xuxa_log(tas_drain);
			}

		if (i == UserParams->lastvideoSCR)
			{
			/* Yank one AU's of video data from the buffer and set the */
			/* next time to yank another AU */

			dist = bytesbetweenVideoAUs(UserParams->videoAU);

			*vbv -= dist;

			UserParams->lastvideoSCR +=
				(double)(((double)VIDEOSAMPLESPERAU/UserParams->videoframesec)*CLOCKFREQ);
sprintf(tas_drain,"\t\tyank video %d next %d SCR %d\n",dist,UserParams->lastvideoSCR,i);
xuxa_log(tas_drain);
			}

		/* Check buffers and report on condition of them */
		if ((*abv < MINAUDBUF)&&(!auderr))
			{
			sprintf(tas_drain,"audio underflow SCR %d buffer %f (%f)\n",i,*abv,iabv);
			xuxa_log(tas_drain);
			auderr = 1;
			ret = AUDIO;
			}
		if ((*abv > MAXAUDBUF)&&(!auderr))
			{
			sprintf(tas_drain,"audio overflow SCR %d buffer %f (%f)\n",i,*abv,iabv);
			xuxa_log(tas_drain);
			auderr = 1;
			ret = AUDIO;
			}
		if ((*vbv < MINVIDBUF)&&(!viderr))
			{
			sprintf(tas_drain,"video underflow SCR %d buffer %f (%f)\n",i,*vbv,ivbv);
			xuxa_log(tas_drain);
			viderr = 1;
			ret = VIDEO;
			}
		if ((*vbv > MAXVIDBUF)&&(!viderr))
			{
			sprintf(tas_drain,"video overflow SCR %d buffer %f (%f)\n",i,*vbv,ivbv);
			xuxa_log(tas_drain);
			viderr = 1;
			ret = VIDEO;
			}
		}

	/* adjust the buffers to integer accuracy */
	/* roundoff errors from the above */
	*abv = (int)(*abv +0.5);
	*vbv = (int)(*vbv +0.5);

	return(ret);
	}



bytesbetweenAudioAUs(FILE *fptr,	/* point to open stream to scan */
		struct inparams *UserParams)
	{
	int i,
		count,
		ch;
	long begin,end;
	int loop, looping;
	int nxtaau;	/* rough number of bytes to next audio AU */

	/* This variable is used to assist in finding the next audio AU */
	/* The problem is that I don't want to fully parse the audio data */
	/* but I still want to find the beginning of the next audio access unit */
	/* If I just search for the syncword, I'm very likely to find a false one */
	/* within either the subband tables or the audio data itself */
	/* To reduce this chance, I just make sure that my search produces */
	/* a lenght of at least something close to what I suspect an audio AU */
	/* to be, this is fixed for layer 2 audio  */
	/* I shorten it by 10 bytes */
	nxtaau = 1152.0/UserParams->audiosamplerate *
		(float)(UserParams->audiorate/8.0) - 10;

	begin = ftell(fptr);

	/* Pointer is always left on byte that is first byte of next AU */
	/* Skip past this one, so can align to next AU */
	ch=fgetc(fptr);

	while (!feof(fptr))
		{
		if ((ch=fgetc(fptr)) == 0xFF)
			{
			if (!feof(fptr))
				{
				end = ftell(fptr);
				if (((ch=(fgetc(fptr)&0xF8)) == 0xF8)&&
				 ((end-begin)>nxtaau))	/* hack!!!, otherwise could id the audio tables as a header */
					{
					/* found the next AU, so back up the number of bytes */
					/* so count is to the start of this next AU */
					fseek(fptr,-2,SEEK_CUR);
					break;
					}
				else
					{
					/* do a reject of the mismatched token */
					fseek(fptr,-1,SEEK_CUR);
					}
				}
			}
		}

	end = ftell(fptr);

	return((int)end-begin);
	}


bytesbetweenVideoAUs(FILE *fptr)	/* point to open stream to scan */
	{
	int i,
		ch;
	int idx;
	int match[] = {0x00,0x00,0x01,0x00};
	int matching;
	int looping, loop;
	long begin,end;

	begin = ftell(fptr);

	/* If this is the begin of the file then have to do this twice */
	/* otherwise will be 1 AU off, ie the first one will be too short */
	if (begin == 0)
		{
		loop = 2;
		}
	else
		loop = 1;

	for (looping = 0;looping<loop;looping++)
		{
		matching = 0;
		idx = 0;
		/* skip first byte, otherwise will loop forever */
		ch = fgetc(fptr);

		/* Pointer is left to read first byte of next AU */
		while (!feof(fptr))
			{
			ch = fgetc(fptr);
			if (match[idx] == ch) 
				matching++;
			else 
				{
				/* rewind buffer by numbr of matches - 1 */
				/* this will accomodate the same as a scanner reject */
				if (matching > 0)
					{
					fseek(fptr,-(matching),SEEK_CUR);
					}
				matching = 0;
				idx = -1;
				}
	
			/* if have matched all 4 then this is a possible picture start code */
			if ((matching == 4)&&(!feof(fptr)))
				{
				fgetc(fptr);	/* trash next value */
				if (!feof(fptr))
					{
					ch = fgetc(fptr)&0x38;
					if ((ch == IFRAME)||
				 	(ch == PFRAME)||
				 	(ch == BFRAME))
				 		{
				 		/* set to the begin of this AU, not its tail */
				 		fseek(fptr,-6,SEEK_CUR);
				 		break;	/* exit since have a picture start code */
				 		}
					else
						{
						fseek(fptr,-5,SEEK_CUR);
						matching = 0;
						idx = -1;
						}
					}
				}
			idx = (idx+1)%4;	/* numbers from 0..3 repeating over and over */
			} /* end while */

		}/* end looping */

	end = ftell(fptr);

	/* Set the number of bytes for this AU */
	return((int)(end-begin));
	}


