
/* File:  FConstr.c */

/*
 *      $Log:   FConstr.c,v $
 * Revision 1.5  93/12/07  12:00:00  sE
 *
 * Revision 1.1.1.1  93/03/29  11:27:54  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:33:17  au
 * Initial revision
 *
 */

#include <stdio.h>
#include "consts.h"
#include "types.h"
#include "decode.h"
#include "decode.pro"

/* construction backwards or forwards macroblock prediction */
void    Construct_Mono_Pred(
	BYTE    *predictY,
	BYTE    *predictU,
	BYTE    *predictV,
	BYTE    *constY,
	BYTE    *constU,
	BYTE    *constV,
	short   MV[]
	)

{
	int     offset, mvx, mvy, index1, index2;
	BYTE    *constY2, *constU2, *constV2;
	BYTE    *frameIndex, *frameIndexU, *frameIndexV;
	int     mv[2];
	int     luminWidth, chromWidth;
	int     f0off2, f0off3, f0off4;

	/*      reconstruct MB  */
	luminWidth = 16 * mb_width;
	chromWidth = 8 * mb_width;

	constY2 = constY;

	/*      lumin   */

	mvx = MV[0] >> 1;
	mvy = MV[1] >> 1;
	offset = (mvy * luminWidth) + mvx;
	f0off2 = MV[0] & 1;
	f0off3 = (MV[1] & 1) * luminWidth;
	f0off4 = f0off2 + f0off3;
	frameIndex = predictY + offset;

	for (index1 = 0; index1 < kMBHeight; index1++,
			constY2 += luminWidth - kMBWidth,
			frameIndex += luminWidth - kMBWidth)
	{
		for (index2 = 0; index2 < kMBWidth; index2++,
				constY2++, frameIndex++)
		{
			/*      add 2 for integer rounding (always non-negative)        */

			*constY2 = ((int) *frameIndex  +
					(int) *(frameIndex + f0off2) +
					(int) *(frameIndex + f0off3) +
					(int) *(frameIndex + f0off4) + 2) / 4;

		}
	}

	/*      chrom   */

	mv[0] = MV[0] / 2;
	mv[1] = MV[1] / 2;
	f0off2 = mv[0] & 1;
	f0off3 = (mv[1] & 1) * chromWidth;
	f0off4 = f0off2 + f0off3;
	mvx = mv[0] >> 1;
	mvy = mv[1] >> 1;
	offset = (mvy * chromWidth) + mvx;
	frameIndexU = predictU + offset;
	frameIndexV = predictV + offset;
	constU2 = constU;
	constV2 = constV;

	for (index1 = 0; index1 < kBHeight; index1++,
			constU2 += chromWidth - kBWidth, constV2 += chromWidth - kBWidth,
			frameIndexU += chromWidth - kBWidth,
			frameIndexV += chromWidth - kBWidth)
	{
		for (index2 = 0; index2 < kBWidth; index2++,
				constU2++, constV2++, frameIndexU++, frameIndexV++)
		{
			*constU2 = ((int) *frameIndexU +
					(int) *(frameIndexU + f0off2) +
					(int) *(frameIndexU + f0off3) +
					(int) *(frameIndexU + f0off4) + 2) / 4;

			*constV2 = ((int) *frameIndexV +
					(int) *(frameIndexV + f0off2) +
					(int) *(frameIndexV + f0off3) +
					(int) *(frameIndexV + f0off4) + 2) / 4;
		}
	}
}


/* Construct bi-directional/interpolated prediction */
void    Construct_Bi_Pred(
	BYTE    *predictY,
	BYTE    *predictU,
	BYTE    *predictV,
	BYTE    *predictYB,
	BYTE    *predictUB,
	BYTE    *predictVB,
	BYTE    *constY,
	BYTE    *constU,
	BYTE    *constV,
	short   MV[],
	short   MVB[]
	)

{
	int             offset, mvx, mvy, index1, index2;
	BYTE    *constY2, *constU2, *constV2;
	BYTE    *frameIndex, *frameIndexU, *frameIndexV;
	BYTE    *frameIndexB, *frameIndexUB, *frameIndexVB;
	int             mv[2];
	int             f0off2, f0off3, f0off4, f0off2B, f0off3B, f0off4B;
	int             pel_for, pel_back;
	int             luminWidth, luminWidthMB, chromWidth, chromWidthB;

	/*      reconstruct MB  */
	luminWidth = 16 * mb_width;
	luminWidthMB = luminWidth - kMBWidth;
	chromWidth = 8 * mb_width;
	chromWidthB = chromWidth - kBWidth;

	/*      lumin   */

	mvx = MV[0] >> 1;
	mvy = MV[1] >> 1;
	offset = (mvy * luminWidth) + mvx;
	f0off2 = MV[0] & 1;
	f0off3 = (MV[1] & 1) * luminWidth;
	f0off4 = f0off2 + f0off3;
	frameIndex = predictY + offset;

	mvx = MVB[0] >> 1;
	mvy = MVB[1] >> 1;
	offset = (mvy * luminWidth) + mvx;
	f0off2B = MVB[0] & 1;
	f0off3B = (MVB[1] & 1) * luminWidth;
	f0off4B = f0off2B + f0off3B;
	frameIndexB = predictYB + offset;

	constY2 = constY;

	for (index1 = 0; index1 < kMBHeight; index1++,
			constY2 += luminWidth - kMBWidth,
			frameIndex += luminWidth - kMBWidth,
			frameIndexB += luminWidth - kMBWidth)
	{
		for (index2 = 0; index2 < kMBWidth; index2++,
			constY2++, frameIndex++, frameIndexB++)
		{
			/*      add 2 for integer rounding (always non-negative)        */

			pel_for = ((int) *frameIndex +
					(int) *(frameIndex + f0off2) +
					(int) *(frameIndex + f0off3) +
					(int) *(frameIndex + f0off4) + 2) / 4;

			pel_back = ((int) *frameIndexB +
					(int) *(frameIndexB + f0off2B) +
					(int) *(frameIndexB + f0off3B) +
					(int) *(frameIndexB + f0off4B) + 2) / 4;

			/*      add 1 for integer rounding (always non-negative)        */

			*constY2 = (pel_for + pel_back + 1) / 2;
		}
	}

	/*      chrom   */

	mv[0] = MV[0] / 2;
	mv[1] = MV[1] / 2;
	f0off2 = mv[0] & 1;
	f0off3 = (mv[1] & 1) * chromWidth;
	f0off4 = f0off2 + f0off3;
	mvx = mv[0] >> 1;
	mvy = mv[1] >> 1;
	offset = (mvy * chromWidth) + mvx;
	frameIndexU = predictU + offset;
	frameIndexV = predictV + offset;

	mv[0] = MVB[0] / 2;
	mv[1] = MVB[1] / 2;
	f0off2B = mv[0] & 1;
	f0off3B = (mv[1] & 1) * chromWidth;
	f0off4B = f0off2B + f0off3B;
	mvx = mv[0] >> 1;
	mvy = mv[1] >> 1;
	offset = (mvy * chromWidth) + mvx;
	frameIndexUB = predictUB + offset;
	frameIndexVB = predictVB + offset;

	constU2 = constU;
	constV2 = constV;

	for (index1 = 0; index1 < kBHeight; index1++,
			constU2 += chromWidth - kBWidth,
			constV2 += chromWidth - kBWidth,
			frameIndexU += chromWidth - kBWidth,
			frameIndexV += chromWidth - kBWidth,
			frameIndexUB += chromWidth - kBWidth,
			frameIndexVB += chromWidth - kBWidth)
	{
		for (index2 = 0; index2 < kBWidth; index2++, constU2++, constV2++,
				frameIndexU++, frameIndexV++, frameIndexUB++, frameIndexVB++)
		{
			pel_for = ((int) *frameIndexU +
					(int) *(frameIndexU + f0off2) +
					(int) *(frameIndexU + f0off3) +
					(int) *(frameIndexU + f0off4) + 2) / 4;

			pel_back = ((int) *frameIndexUB +
					(int) *(frameIndexUB + f0off2B) +
					(int) *(frameIndexUB + f0off3B) +
					(int) *(frameIndexUB + f0off4B) + 2) / 4;

			*constU2 = (pel_for + pel_back + 1) / 2;

			pel_for = ((int) *frameIndexV +
					(int) *(frameIndexV + f0off2) +
					(int) *(frameIndexV + f0off3) +
					(int) *(frameIndexV + f0off4) + 2) / 4;

			pel_back = ((int) *frameIndexVB +
					(int) *(frameIndexVB + f0off2B) +
					(int) *(frameIndexVB + f0off3B) +
					(int) *(frameIndexVB + f0off4B) + 2) / 4;

			*constV2 = (pel_for + pel_back + 1) / 2;
		}
	}
}
