/* file: convert.c */

/*
 *      $Log:   convert.c,v $
 * Revision 1.5  93/12/07  12:00:00  sE
 *
 * Revision 1.1.1.1  93/03/29  11:27:55  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:33:17  au
 * Initial revision
 *
 */

#include <malloc.h>
#include "consts.h"
#include "types.h"
#include "decode.pro"

void    convert420to422(
	BYTE    *chromPicture,
	int             chromWidth,
	int             chromHeight)
{
	int             i;
	int             j;
	int             *a;
	int             *b;

	a = (int *) malloc(chromHeight * sizeof(int));
	b = (int *) malloc(chromHeight * sizeof(int));

	for(j = 0; j < chromWidth; j++)
	{
		for(i = 0; i < chromHeight / 2; i++)
			*(a + i) = *(chromPicture + i * chromWidth + j);

		filter(chromHeight / 2, a, b);

		for(i = 0; i < chromHeight; i++)
			*(chromPicture + i * chromWidth + j) = *(b + i);
	}

	free(a);
	free(b);
}

void    filter(
	int     n,
	int     *F,
	int     *G)
{
	int i, k, M;
	int     tmp;


	M = 2 * n;
	
	F[n] = F[n - 1];  /* for 'interpolation' of last point */

	for(i = 0; i < n; i++)
	{
		k = 2 * i;
		G[k] = F[i];

		tmp = F[i] + F[i + 1];

		/* normalize and round */
		G[k + 1] = (tmp >= 0) ? ((tmp + 1) / 2) : -((-tmp + 1) / 2);
	}
	
	return;
}

void convert420to444(
	BYTE    *chrom420,
	BYTE    *chrom444,
	int             chromWidth,
	int             chromHeight)
{
	int             chromWidth2, i, j;
	BYTE    *src, *dst;

	/* bilinear interpolation filter */

	chromWidth2 = 2 * chromWidth;

	/* horizontal direction */

	for (j = 0; j < chromHeight; j++)
	{
		src = chrom420 + j * chromWidth;
		dst = chrom444 + j * chromWidth2;
		dst[0] = src[0];
		for (i = 1; i < chromWidth; i++)
		{
			dst[2*i-1] = (3 * src[i-1] + src[i] + 2) >> 2;
			dst[2*i]   = (src[i-1] + 3 * src[i] + 2) >> 2;
		}
		dst[chromWidth2-1] = src[chromWidth-1];
	}

	/* vertical direction, (in-place, bottom-up) */

	for (i = 0; i < chromWidth2; i++)
	{
		src = chrom444 + i + (chromHeight-2) * chromWidth2;
		dst = chrom444 + i + (2*(chromHeight-2)+1) * chromWidth2;
		dst[2*chromWidth2] = src[chromWidth2];
		for (j = chromHeight - 2; j >= 0; j--)
		{
			dst[chromWidth2] = (src[0] + 3 * src[chromWidth2] + 2) >> 2;
			dst[0]           = (3 * src[0] + src[chromWidth2] + 2) >> 2;
			dst -= 2 * chromWidth2;
			src -= chromWidth2;
		}
	}
}
