/* getbits.c */

/*
 *      $Log:   GetBits.c,v $
 * Revision 1.5  93/12/07  12:00:00  sE
 *
 * Revision 1.1.1.1  93/03/29  11:27:56  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:34:26  au
 * Initial revision
 *
 */

#include <stdio.h>
#include "consts.h"
#include "types.h"
#include "decode.pro"

#define kBufSize        4096
#define kByteAlign      0x80
#define kWdPtrStart     0x80

FILE    *gBitStream;

static unsigned char    buf[kBufSize + 4];
static unsigned int             wdPtr = kWdPtrStart;
static int                              wdIndex = 0;
static int                              numInBuf = 0;

int     Get_Bits(
	unsigned int    *destBuf,
	int                             numBits)
{
	int     index;

	*destBuf = 0;

	for (index = 0; index < numBits; index++)
	{
		/* refresh buffer if empty */
		if (wdIndex >= numInBuf)
		{
			/* read next kBufSize bytes */
			if ((numInBuf = fread(buf + 4, sizeof(unsigned char), kBufSize,
									gBitStream)) <= 0)
				return(0);              /*      end of bit stream       */
		
			wdIndex = 0;
			wdPtr = kWdPtrStart;
		}

		/* get next bit */

		*destBuf = ((buf[wdIndex + 4] & wdPtr) ?
				(*destBuf << 1) | 0x01 : (*destBuf << 1));

		/* update bit pointer */
		if (wdPtr == 0x01)
		{
			wdIndex++;
			wdPtr = kWdPtrStart;
		}
		else
			wdPtr >>= 1;
	}

	return(1);
}

int     Next_Bits(
	unsigned int    *destBuf,
	int                             numBits)
{
	unsigned int    wdPtrSave;
	int                             wdIndexSave;
	int                             index;

	/* save current buffer state */
	wdPtrSave = wdPtr;
	wdIndexSave = wdIndex;

	*destBuf = 0;

	for (index = 0; index < numBits; index++)
	{
		/* refresh buffer if empty */
		if (wdIndex >= numInBuf)
		{
			/* save last 4 bytes of old content */
			buf[0]=buf[numInBuf];
			buf[1]=buf[numInBuf + 1];
			buf[2]=buf[numInBuf + 2];
			buf[3]=buf[numInBuf + 3];

			wdIndexSave -= numInBuf;

			/* read next kBufSize bytes */
			if ((numInBuf = fread(buf + 4, sizeof(unsigned char), kBufSize,
									gBitStream)) <= 0)
				return(0);              /*      end of bit stream       */
		
			wdIndex = 0;
			wdPtr = kWdPtrStart;
		}

		/* get next bit */

		*destBuf = ((buf[wdIndex + 4] & wdPtr) ?
				(*destBuf << 1) | 0x01 : (*destBuf << 1));

		/* update bit pointer */
		if (wdPtr == 0x01)
		{
			wdIndex++;
			wdPtr = kWdPtrStart;
		}
		else
			wdPtr >>= 1;
	}

	/* restore previous buffer state */
	wdPtr = wdPtrSave;
	wdIndex = wdIndexSave;

	return(1);
}

int Next_Start_Code()
{
	unsigned int buf;

	/* locate next start code */

	if (wdPtr != kByteAlign) /* not byte aligned */
	{
		/* skip stuffed zero bits */
		wdPtr=kByteAlign;
		wdIndex++;
	}

	if (!Next_Bits(&buf, 24))
		return(0); /* end of bitstream */

	while (buf != 0x000001)
	{
		if (!Get_Bits(&buf, 8)) /* zero byte */
			return(0);
		if (!Next_Bits(&buf, 24))
			return(0);
	}

	return(1);
}

int Get(
	int numBits)
{
	unsigned int buf;

	if (! Get_Bits(&buf, numBits))
	{
		printf ("Error (Get - End Of File)\n");
		return(0);
	}

	return(buf);
}
