/* File: Get_Code.c */

/*
 *      $Log:   Get_Code.c,v $
 * Revision 1.5  93/12/07  12:00:00  sE
 *
 * Revision 1.1.1.1  93/03/29  11:27:57  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:34:26  au
 * Initial revision
 *
 */

#include <stdio.h>
#include "consts.h"
#include "types.h"
#include "vlc.h"
#include "decode.pro"

int     Get_AC_Code(
	BYTE                    *bit_count,
	BYTE                    *bit_pattern,
	int                     table_entries,
	int                     *sign_bit,
	int                     *run,
	int                     *level,
	int                     first)
{
	unsigned int    buf, tempInt = 0;
	int                             num_bits = 0, code;

	while (Get_Bits(&buf, 1))
	{
		num_bits++;
		tempInt = (tempInt << 1);

		/*      check for Escape code   */

		if ((num_bits == AC_escape_length) && ((tempInt | buf) == AC_escape_value))
		{
			/*      escape encountered      */

			Get_Bits(&buf, 6);               /*      get run code    */

			*run = buf;

			Get_Bits(&buf, 8);               /*      get level code  */

			if (buf == 128)
			{
				/*      28 bit code - negative level    */

				Get_Bits(&buf, 8);       /*      get level code  */
				if((buf == 0) || (buf > 128))
					printf ("Error (Get_AC_Code) - invalid level %d\n", buf);
				*level = buf - 256;
			}
			else if (buf == 0)
			{
				/*      28 bit code - positive level    */

				Get_Bits(&buf, 8);       /*      get level code  */
				if(buf < 128)
					printf ("Error (Get_AC_Code) - invalid level %d\n", buf);
				*level = buf;
			}
			else
			{
				/*  20 bit code */
				if (buf == 0)
					printf ("Error (Get_AC_Code) - invalid level %d\n", buf);

				*level = (buf<128) ? buf : (buf - 256);
			}

			return(-2);             /*      Escape encounter        */
		}

		if (first && (num_bits == AC_first_length) && (tempInt == AC_first_value))
		{
			/* modified table entry for dct_coeff_first */
			code = 1;
		}
		else
			code = Find_VLC(bit_count, bit_pattern, table_entries, tempInt, num_bits);

		if (code != table_entries)
		{
			/* don't return if code==0 (EOB) but buf==1 (EOB has buf==0) */
			if (code || (! buf))
			{
				*sign_bit = buf;

				return(code);
			}
		}

		tempInt |= buf;
	}

	return(-1);
}


int     Get_Code(
	BYTE                    *bit_count,
	BYTE                    *bit_pattern,
	int                     table_entries)
{
	unsigned int    buf, tempInt = 0;
	int             num_bits = 0, code;

	while (Get_Bits(&buf, 1))
	{
		num_bits++;
		tempInt = (tempInt << 1) | buf;

		code = Find_VLC(bit_count, bit_pattern, table_entries, tempInt, num_bits);

		if (code != table_entries)
		{
			return(code);
		}
	}

	return(-1);
}


int     Find_VLC(
	BYTE                    *bit_count,
	BYTE                    *bit_pattern,
	int                     table_entries,
	unsigned int            buf,
	int                     num_bits)
{
	int             index, maxBits = 0;

	for (index = 0; index < table_entries; index++)
	{
		if ((bit_count[index] == num_bits) && (bit_pattern[index] == buf))
			return(index);

		if (maxBits < bit_count[index])
			maxBits = bit_count[index];
	}

	if (num_bits >= maxBits)
	{
		printf ("Error (Find_VLC) - %d 0x%x\n", num_bits, buf);
		return(-1);                     /*      mismatch table and buf  */
	}

	return(index);
}


/* formerly file: getcoeffs.c  */

int     Get_AC_Coefs(
	int             *currentMBDiff,
	int             currentIndex,
	int             offset)
{
	extern BYTE     AC_length[], AC_value[];
	extern short    AC_run[], AC_level[];

	int             row, code, sign, run, level, index, col;

	col = currentIndex;

	for (row = 0; row < kDCTSize; row++, currentIndex += offset)
	{
		for ( ; col < kDCTSize; col++, currentIndex++)
		{
			/*      get (first or next) DCT coefficient  */

			if ((code = Get_AC_Code(AC_length, AC_value, 
				112, &sign, &run, &level, currentIndex==0)) == -1)
			{
				printf ("Error (Get_AC_Coefs) dct_coef_first - row %d col %d code %x\n", row, col, code);
				return(0);
			}

			if (code == 0)
				return(1);              /*      EOB encountered */

			if (code != -2)         /*      not escape      */
			{
				run = AC_run[code];
				level = sign ? -AC_level[code] : AC_level[code];
			}

			for (index = 0; index < run; index++)
			{
				col++;
				currentIndex++;

				if (col == kDCTSize)
				{
					row++;
					col = 0;
					currentIndex += offset;
				}
			}

			currentMBDiff[currentIndex] = level;

		}

		col = 0;
	}

	/*      remove EOB code */

	if ((code = Get_AC_Code(AC_length, AC_value, 112,
			&sign, &run, &level, 0)) == -1)
	{
		printf ("Error (Get_AC_Coefs) dct_next: row %d col %d code %x\n", row, col, code);
		return(0);
	}

	if (code)
	{
		printf ("Error (Get_AC_Coefs) - remove EOB\n");
		return(0);
	}

	return(1);
}
