/* idct.c */

/*
 *	$Log:	idct.c,v $
 * Revision 1.5  93/12/07  12:00:00  sE
 *
 * Revision 1.1.1.1  93/03/29  11:27:55  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:33:17  au
 * Initial revision
 *
 */

#include <math.h>
#include "consts.h"

#ifndef PI
#ifdef M_PI
#define	PI	M_PI
#else
#define PI	3.14159265358979323846
#endif
#endif

double	TwiddleFor[kDCTSize][kDCTSize];
double	TwiddleInv[kDCTSize][kDCTSize];

/************************************************************************
*
* Module name: Init_DCT
*
* Initializes the DCT scale and twiddle factors
*
************************************************************************/

void Init_DCT(void)
{
	extern double	TwiddleFor[][kDCTSize];
	extern double	TwiddleInv[][kDCTSize];

	/* Local variables */

	double	alpha[kDCTSize];
	short i,j;

	/* Execute */

	alpha[0] = sqrt((double)(1.0 / kDCTSize));

	for(i = 1; i < kDCTSize; i++)
		alpha[i] = sqrt((double)(2.0 / kDCTSize));

	for(i = 0; i < kDCTSize; i++)
	{
		for(j = 0; j < kDCTSize; j++)
		{
			TwiddleFor[i][j] = alpha[i] * cos((double)((2 * j + 1) * i * PI) /
					(2 * kDCTSize));
			TwiddleInv[i][j] = alpha[j] * cos((double)((2 * i + 1) * j * PI) /
					(2 * kDCTSize));
		}
	}
	/*
	alpha[0] = 1.0 / (2.0 * kDCTSize);

	for(i = 1; i < kDCTSize; i++)
		alpha[i] = 1.0 / kDCTSize;

	for(i = 0; i < kDCTSize; i++)
	{
		for(j = 0; j < kDCTSize; j++)
		{
			TwiddleFor[i][j] = 2.0 * cos((double)((2 * j + 1) * i * PI) /
					(2 * kDCTSize));
			TwiddleInv[i][j] = alpha[j] * cos((double)((2 * i + 1) * j * PI) /
					(2 * kDCTSize));
		}
	}
	*/
}

/************************************************************************
*
* Module name: IDCT
*
* Calculates an inverse discrete cosine transform
*
*************************************************************************/

void idct8x8(
	float	dct_img[],
	float	recon_img[])
{
	extern double	TwiddleInv[][kDCTSize];

	/* Local variables */

	short	i, m, n, index, index1;
	float	tmpImg[kDCTLen];

	/* Execute */

	for(i = 0, index = 0; i < kDCTSize; i++, index += kDCTSize)
	{
		for(m = 0; m < kDCTSize; m++)
		{
			tmpImg[index + m] = 0;

			for(n = 0; n < kDCTSize; n++)
			{
				tmpImg[index + m] += dct_img[index + n] * TwiddleInv[m][n];
			}
		}
	}

	for(i = 0; i < kDCTSize; i++)
	{
		for(m = 0, index = i; m < kDCTSize; m++, index += kDCTSize)
		{
			recon_img[index] = 0;

			for(n = 0, index1 = i; n < kDCTSize; n++, index1 += kDCTSize)
			{
				recon_img[index] += tmpImg[index1] * TwiddleInv[m][n];
			}
		}
	}
}
