/* File: iquant.c */

/*
 *      $Log:   iquant.c,v $
 * Revision 1.5  93/12/07  12:00:00  sE
 *
 * Revision 1.1.1.1  93/03/29  11:27:55  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:33:17  au
 * Initial revision
 *
 */

#include <stdio.h>
#include "consts.h"
#include "types.h"
#include "quant.h"
#include "decode.pro"


void    Inverse_Quantize_Intra_MB(
	short   codedBPattern,
	int             *currentMBDiff,
	int             *currentMBDiffU,
	int             *currentMBDiffV,
	int             MQuant)
{
	int             line, pix, row, col, quantIndex;
	int             currentIndex;
	int             tInt;

	/*      lumin   */

	for (line = 0; line < kMBHeight; line += kDCTSize)
	{
		for (pix = 0; pix < kMBWidth; pix += kDCTSize)
		{
			currentIndex = line * kMBWidth + pix;

			currentMBDiff[currentIndex] *= 8;

			if (currentMBDiff[currentIndex] < -2048)
				currentMBDiff[currentIndex] = -2048;
			else if (currentMBDiff[currentIndex] > 2047)
				currentMBDiff[currentIndex] = 2047;

			/*      exclude dc coeff        */

			currentIndex++;

			for (row = 0, col = 1, quantIndex = 1; row < kDCTSize;
					row++, col = 0, currentIndex += kMBWidthDCT)
			{
				for ( ; col < kDCTSize; col++, currentIndex++, quantIndex++)
				{
					if (currentMBDiff[currentIndex] != 0)
					{
						/*      rec = MQuant * 2 * QAC * WI / 16        */

						tInt = (MQuant * currentMBDiff[currentIndex] *
								(int) tQuantIntra[quantIndex]) / 8;

						if (! (tInt % 2))               /*      even    */
						{
							if (tInt < 0)
								tInt += 1;
							else if (tInt > 0)
								tInt -= 1;
						}

						if (tInt <= -2048)
							currentMBDiff[currentIndex] = -2048;
						else if (tInt >= 2047)
							currentMBDiff[currentIndex] = 2047;
						else
							currentMBDiff[currentIndex] = tInt;
					}
				}
			}
		}
	}

	/*      chrom   */

	currentMBDiffU[0] *= 8;

	if (currentMBDiffU[0] < -2048)
		currentMBDiffU[0] = -2048;
	else if (currentMBDiffU[0] > 2047)
		currentMBDiffU[0] = 2047;

	currentIndex = 1;

	for (row = 1; row < kDCTLen; row++, currentIndex++)
	{
		/*      U       */

		if (currentMBDiffU[currentIndex] != 0)
		{
			/*      rec = MQuant * 2 * QAC * WI / 16        */

			tInt = (MQuant * currentMBDiffU[currentIndex] *
					(int) tQuantIntra[currentIndex]) / 8;

			if (! (tInt % 2))               /*      even    */
			{
				if (tInt < 0)
					tInt += 1;
				else if (tInt > 0)
					tInt -= 1;
			}

			if (tInt <= -2048)
				currentMBDiffU[currentIndex] = -2048;
			else if (tInt >= 2047)
				currentMBDiffU[currentIndex] = 2047;
			else
				currentMBDiffU[currentIndex] = tInt;
		}
	}

	currentMBDiffV[0] *= 8;

	if (currentMBDiffV[0] < -2048)
		currentMBDiffV[0] = -2048;
	else if (currentMBDiffV[0] > 2047)
		currentMBDiffV[0] = 2047;

	currentIndex = 1;

	for (row = 1; row < kDCTLen; row++, currentIndex++)
	{
		/*      V       */

		if (currentMBDiffV[currentIndex] != 0)
		{
			/*      rec = MQuant * 2 * QAC * WI / 16        */

			tInt = (MQuant * currentMBDiffV[currentIndex] *
					(int) tQuantIntra[currentIndex]) / 8;

			if (! (tInt % 2))               /*      even    */
			{
				if (tInt < 0)
					tInt += 1;
				else if (tInt > 0)
					tInt -= 1;
			}

			if (tInt <= -2048)
				currentMBDiffV[currentIndex] = -2048;
			else if (tInt >= 2047)
				currentMBDiffV[currentIndex] = 2047;
			else
				currentMBDiffV[currentIndex] = tInt;
		}
	}
}


void    Inverse_Quantize_Non_Intra_MB(
	short   codedBPattern,
	int             *currentMBDiff,
	int             *currentMBDiffU,
	int             *currentMBDiffV,
	int             MQuant)
{
	int             code, line, pix, row, col, quantIndex;
	int             currentIndex;
	int             tInt;

	/*      lumin   */

	code = 0x20;

	for (line = 0; line < kMBHeight; line += kDCTSize)
	{
		for (pix = 0; pix < kMBWidth; pix += kDCTSize)
		{
			if (! (codedBPattern & code))
			{
				code >>= 1;
				continue;
			}

			code >>= 1;

			currentIndex = line * kMBWidth + pix;

			for (row = 0, quantIndex = 0; row < kDCTSize; row++,
					currentIndex += kMBWidthDCT)
			{
				for (col = 0; col < kDCTSize; col++,
						currentIndex++, quantIndex++)
				{
					if (currentMBDiff[currentIndex])
					{
						if (currentMBDiff[currentIndex] > 0)
							tInt = ((2 * currentMBDiff[currentIndex] + 1) *
									MQuant * (int) tQuant[quantIndex]) / 16;
						else    /*      < 0             */
							tInt = ((2 * currentMBDiff[currentIndex] - 1) *
									MQuant * (int) tQuant[quantIndex]) / 16;

						if (! (tInt % 2))               /*      even    */
						{
							if (tInt < 0)
								tInt += 1;
							else if (tInt > 0)
								tInt -= 1;
						}

						if (tInt <= -2048)
							currentMBDiff[currentIndex] = -2048;
						else if (tInt >= 2047)
							currentMBDiff[currentIndex] = 2047;
						else
							currentMBDiff[currentIndex] = tInt;
					}
				}
			}
		}
	}

	/*      chrom   */

	if (codedBPattern & 0x2)
	{
		/*      U       */

		for (row = 0, currentIndex = 0; row < kDCTLen; row++, currentIndex++)
		{
			if (currentMBDiffU[currentIndex])
			{
				if (currentMBDiffU[currentIndex] > 0)
					tInt = ((2 * currentMBDiffU[currentIndex] + 1) *
							MQuant * (int) tQuant[currentIndex]) / 16;
				else
					tInt = ((2 * currentMBDiffU[currentIndex] - 1) *
							MQuant * (int) tQuant[currentIndex]) / 16;

				if (! (tInt % 2))               /*      even    */
				{
					if (tInt < 0)
						tInt += 1;
					else if (tInt > 0)
						tInt -= 1;
				}

				if (tInt <= -2048)
					currentMBDiffU[currentIndex] = -2048;
				else if (tInt >= 2047)
					currentMBDiffU[currentIndex] = 2047;
				else
					currentMBDiffU[currentIndex] = tInt;
			}
		}
	}

	if (codedBPattern & 0x1)
	{
		/*      V       */

		for (row = 0, currentIndex = 0; row < kDCTLen; row++, currentIndex++)
		{
			if (currentMBDiffV[currentIndex])
			{
				if (currentMBDiffV[currentIndex] > 0)
					tInt = ((2 * currentMBDiffV[currentIndex] + 1) *
							MQuant * (int) tQuant[currentIndex]) / 16;
				else
					tInt = ((2 * currentMBDiffV[currentIndex] - 1) *
							MQuant * (int) tQuant[currentIndex]) / 16;

				if (! (tInt % 2))               /*      even    */
				{
					if (tInt < 0)
						tInt += 1;
					else if (tInt > 0)
						tInt -= 1;
				}

				if (tInt <= -2048)
					currentMBDiffV[currentIndex] = -2048;
				else if (tInt >= 2047)
					currentMBDiffV[currentIndex] = 2047;
				else
					currentMBDiffV[currentIndex] = tInt;
			}
		}
	}
}
