/* File: mb.c */

/*
 *      $Log:   procMB.c,v $
 * Revision 1.5  93/12/07  12:00:00  sE
 *
 * Revision 1.1.1.1  93/03/29  11:27:57  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:34:26  au
 * Initial revision
 *
 */

#include <stdio.h>
#include "consts.h"
#include "types.h"
#include "vlc.h"
#include "decode.h"
#include "decode.pro"


int     Process_MB(
	int             *currentMBDiff,
	int             *currentMBDiffU,
	int             *currentMBDiffV,
	int             *dct_dc_y_past,
	int             *dct_dc_cb_past,
	int             *dct_dc_cr_past,
	int             macroblock_type,
	int             coded_block_pattern)
{
	int             block_number, line, pix, mbIntra;

	mbIntra = (macroblock_type & kMbIntra) != 0;
	block_number = 0;

	/* Y blocks */

	for (line = 0; line < kMBHeight; line += kDCTSize)
	{
		for (pix = 0; pix < kMBWidth; pix += kDCTSize)
		{
			if (!Process_Block(block_number, coded_block_pattern, 
				mbIntra, dct_dc_y_past, currentMBDiff + line * kMBWidth + pix))
				return(0);

			block_number++;
		}
	}

	/* Cb */

	if (!Process_Block(4, coded_block_pattern, mbIntra, dct_dc_cb_past, currentMBDiffU))
		return(0);

	/* Cr */

	if (!Process_Block(5, coded_block_pattern, mbIntra, dct_dc_cr_past, currentMBDiffV))
		return(0);

	return(1);
}

int Process_Block(
	int             block_number,
	int             coded_block_pattern,
	int             mbIntra,
	int             *dct_dc_past,
	int             *dct_recon)
{
	extern BYTE     DC_Y_length[];
	extern BYTE     DC_Y_value[];
	extern BYTE     DC_C_length[];
	extern BYTE     DC_C_value[];

	int             currentIndex, size, diff_DC;
	unsigned int    buf;

	if ((coded_block_pattern & (32>>block_number)) || mbIntra)
	{
		currentIndex = 0;
		if (mbIntra)
		{
			/*      get dct_dc_size */

			if (block_number < 4)
				/* luminance */
				size = Get_Code(DC_Y_length, DC_Y_value, 9);
			else
				/* chrominance */
				size = Get_Code(DC_C_length, DC_C_value, 9);

			if (size == -1)
			{
				printf ("Error (Process_Block) - GetCode - %d\n", block_number);
				return(0);
			}

			if (size != 0)
			{
				/* get dct_dc_differential */                   
				if (Get_Bits(&buf, size) == 0)
				{
					printf ("Error (Process_Block) - Get_Bits after fetching DC coef - %d\n", block_number);
					return(0);
				}

				/* compute dct_zz[0] */         
				if (buf & (1 << (size - 1)))
					diff_DC = buf; /* positive */
				else
					diff_DC = (-1 << size) | (buf + 1); /* negative */
			}
			else
				diff_DC = 0; /* zero */
	
			dct_recon[0] = *dct_dc_past + (diff_DC * 1);
			*dct_dc_past = dct_recon[0];
			currentIndex = 1;
		}
	
		/*      get AC coeffs   */
	
		if (! Get_AC_Coefs(dct_recon, currentIndex, 
                        (block_number < 4) ? kMBWidthDCT : 0))
		{
			printf ("Error (Process_Block) - Get_AC_Coefs - %d\n", 
                         block_number);
			return(0);
		}
	}

	return(1);
}
