/* picture.c */


/*
 *      $Log:   procPicture.c,v $
 * Revision 1.5  93/12/07  12:00:00  sE
 *
 * Revision 1.2  93/06/15  15:41:31  oosa
 * 
 * 
 * Revision 1.1  1993/03/10  20:34:26  au
 * Initial revision
 *
 */

#include <stdio.h>
#include "consts.h"
#include "types.h"
#include "flc.h"
#include "decode.h"
#include "decode.pro"

int     Process_Picture(
	int             pictureNum,
	tParameter      *parameter,
	tData           *data)
{
	unsigned int    buf;
	BYTE            *buf_Y, *buf_Cb, *buf_Cr;
	int             picture_coding_type, forward_f_code, backward_f_code;
	int             fullPelForwardV, fullPelBackwardV;
	int             temporal_reference, vbv_delay;
	static char     *typeTab[8]={"forbidden","intra-coded (I)",
				     "predictive-coded (P)",
				     "bidirectionally-predictive-coded (B)",
				     "dc intra-coded (D)", "reserved",
				     "reserved", "reserved"};


	/* picture layer */

	temporal_reference = Get(bTemporalReference);
	picture_coding_type = Get(bPictureCodingType);
	vbv_delay = Get(bVbvDelay);

	if (TraceLevel >= 1)
	{
		fprintf(TraceFile, "Picture Header:\n");
		fprintf(TraceFile, " temporal reference: %d\n", temporal_reference);
		fprintf(TraceFile, " picture coding type: %s\n", typeTab[picture_coding_type]);
		fprintf(TraceFile, " VBV delay: %d\n", vbv_delay);
	}

	if ((picture_coding_type == kPPicture) || (picture_coding_type == kBPicture))
	{
		fullPelForwardV = Get(bFullPelForwardVector);
		forward_f_code = Get(bForwardFCode);

		if (TraceLevel >= 1)
		{
			fprintf(TraceFile, " full pel forward vector: %s\n",
				fullPelForwardV ? "yes" : "no");
			if (fullPelForwardV)
				fprintf(TraceFile,
					" forward_f_code: %d, (range %d to %d)\n", forward_f_code,
					-(8 << forward_f_code), (8 << forward_f_code) - 1);
			else
				fprintf(TraceFile,
					" forward_f_code: %d, (range %.1f to %.1f)\n",
					forward_f_code,
					-0.5*(8 << forward_f_code),
					0.5*((8 << forward_f_code) - 1));
		}
	}

	if (picture_coding_type == kBPicture)
	{
		fullPelBackwardV = Get(bFullPelBackwardVector);
		backward_f_code = Get(bBackwardFCode);

		if (TraceLevel >= 1)
		{
			fprintf(TraceFile, " full pel backward vector: %s\n",
				fullPelBackwardV ? "yes" : "no");
			if (fullPelBackwardV)
				fprintf(TraceFile,
					" backward_f_code: %d, (range %d to %d)\n", backward_f_code,
					-(8 << backward_f_code), (8 << backward_f_code) - 1);
			else
				fprintf(TraceFile,
					" backward_f_code: %d, (range %.1f to %.1f)\n",
					backward_f_code,
					-0.5*(8 << backward_f_code),
					0.5*((8 << backward_f_code) - 1));
		}
	}

	/* skip extra_bit_picture & extra_information_picture */
	Get_Bits(&buf, 1);
	while(buf != 0) {
	    Get_Bits(&buf, 8);
	    Get_Bits(&buf, 1);
	}

	Next_Start_Code();

	Extension_And_User_Data();

	if (picture_coding_type == kBPicture)
	{
		buf_Y = data->predictYR;
		buf_Cb = data->predictUR;
		buf_Cr = data->predictVR;
	}
	else
	{
		buf_Y = data->firstPredictYR;
		buf_Cb = data->firstPredictUR;
		buf_Cr = data->firstPredictVR;

		/* Rotate the prediction buffers */
		data->firstPredictYR = data->secondPredictYR;
		data->firstPredictUR = data->secondPredictUR;
		data->firstPredictVR = data->secondPredictVR;

		data->secondPredictYR = buf_Y;
		data->secondPredictUR = buf_Cb;
		data->secondPredictVR = buf_Cr;

		/* write the previous I/P-picture */
		if(pictureNum != 0){
		    if (Write_Picture(pictureNum,
					data->firstPredictYR,
					data->firstPredictUR,
					data->firstPredictVR,
					parameter))
				exit (-1);
		}
	}

	/* decode a picture */
	do
	{
		if (! Process_Slice(picture_coding_type, buf_Y, buf_Cb, buf_Cr, data,
				forward_f_code, fullPelForwardV,
				backward_f_code, fullPelBackwardV))
		{
			printf ("Error (procPicture) - procSlice\n");
			return(0);
		}
		Next_Bits(&buf, bStartCode);
	}
	while ((buf >= cSliceCodeMin) && (buf <= cSliceCodeMax));

	/* B-pictures are displayed right after they are decoded. */
	if(picture_coding_type == kBPicture){
	    /* write B-picture */
	    if (Write_Picture(pictureNum, buf_Y, buf_Cb, buf_Cr, parameter))
		exit (-1);
	}

	return(1);
}

