/* file: sequence.c */

/*
 *      $Log:   Process_Picture.c,v $
 * Revision 1.5  93/12/07  12:00:00  sE
 *
 * Revision 1.2  93/06/15  15:41:31  oosa
 * 
 * 
 * Revision 1.1  1993/03/10  20:34:26  au
 * Initial revision
 *
 */

#include <stdio.h>
#include "consts.h"
#include "types.h"
#include "flc.h"
#include "decode.h"
#include "decode.pro"

int     Process_Sequence(
	tParameter      *parameter,
	tData           *data)
{
	unsigned int buf;
	int pictureNum;
	int time_code, closed_GOP, broken_link;

	/* video sequence layer */

	pictureNum = 0;

	Next_Start_Code();

	do
	{
		if (! Get_Bits(&buf, bStartCode))
			return(0);

		if (buf != cSequenceHeaderCode)
		{
			printf ("Error (procSequence - sequence header)\n");
			return(0);
		}

		/* handle sequence header */
	
		Process_Sequence_Header();

		/* initialize data structures */
		if (pictureNum == 0)
			Init_Data(data); 

		do
		{
			if (! Get_Bits(&buf, bStartCode))
				return(0);

			if (buf != cGroupStartCode)
			{
				printf ("Error (procSequence - group start code)\n");
				return(0);
			}

			/* group of pictures layer */

			/*      handle GOP Header       */
	
			time_code = Get(bTimeCode);
			closed_GOP = Get(bClosedGOP);
			broken_link = Get(bBrokenLink);

			if (TraceLevel >= 1)
			{
				fprintf(TraceFile, "Group of pictures:\n");
				fprintf(TraceFile,
					" time code (hh:mm:ss:pp): %02d:%02d:%02d:%02d\n",
					(time_code >> 19) & 0x1f, (time_code >> 13) & 0x3f,
					(time_code >> 6) & 0x3f, time_code & 0x3f);
				fprintf(TraceFile, "closed GOP: %s\n",
					closed_GOP ? "yes" : "no");
				fprintf(TraceFile, "broken link: %s\n",
					broken_link ? "yes" : "no");
			}

			Next_Start_Code();

			Extension_And_User_Data();
	
			do
			{
				if (! Get_Bits(&buf, bStartCode))
					return(0);
	
				if (buf != cPictureStartCode)
				{
					printf ("Error (procSequence - picture start code)\n");
					return(0);
				}

				Process_Picture(pictureNum, parameter, data);

				printf("Decoded Picture: %d\n",pictureNum);

				pictureNum++;

				Next_Bits(&buf, bStartCode);
			}
			while (buf == cPictureStartCode);

		}
		while (buf == cGroupStartCode);

	}
	while (buf == cSequenceHeaderCode);

	if (! Get_Bits(&buf, bStartCode))
		return(0);

	if (buf != cSequenceEndCode)
	{
		printf ("Error (procSequence - sequence end code)\n");
		return(0);
	}

    /* write the last I/P-picture */
    if (Write_Picture(pictureNum,
		data->secondPredictYR,
		data->secondPredictUR,
		data->secondPredictVR,
		parameter))
		exit (-1);

	return(0);
}

void    Process_Sequence_Header(void)
{
	extern BYTE     tQuantIntra[], tQuant[];

	int             index;
	int             pel_aspect_ratio, picture_rate, bit_rate, marker_bit;
	int             vbv_buffer_size, constrained_parameters_flag;
	int             load_intra_quantizer_matrix, load_non_intra_quantizer_matrix;
	static char     *aspectTab[16] = {"forbidden", "1.0000 (VGA etc.)", "0.6735",
				"0.7031 (16:9, 625line)", "0.7615", "0.8055",
				"0.8437 (16:9, 525line)", "0.8935",
				"0.9157 (CCIR601, 625line)", "0.9815", "1.0255", "1.0695",
				"1.0950 (CCIR601, 525line)", "1.1575", "1.2015", "reserved"};
	static char *rateTab[9] = {"forbidden", "23.976", "24", "25", "29.97",
				"30", "50", "59.94", "60"};

	/*      process sequence header */

	horizontal_size = Get(bHorizontalSize);
	vertical_size = Get(bVerticalSize);
	pel_aspect_ratio = Get(bPelAspectRatio);
	picture_rate = Get(bPictureRate);
	bit_rate = Get(bBitRate);
	marker_bit = Get(bMarkerBit);
	vbv_buffer_size = Get(bVbvBufferSize);
	constrained_parameters_flag = Get(bConstrainedParameterFlag);

	mb_width = (horizontal_size + 15)/16;
	mb_height = (vertical_size + 15)/16;

	load_intra_quantizer_matrix = Get(bLoadIntraQuantizerMatrix);

	if (load_intra_quantizer_matrix)
		for (index = 0; index < 64; index++)
			tQuantIntra[index] = Get(bIntraQuantizer);

	load_non_intra_quantizer_matrix = Get(bLoadNonIntraQuantizerMatrix);

	if (load_non_intra_quantizer_matrix)
		for (index = 0; index < 64; index++)
			tQuant[index] = Get(bNonIntraQuantizer);

	if(TraceLevel>=1)
	{
		fprintf(TraceFile, "Sequence Header:\n");
		fprintf(TraceFile, " picture size (H x V): %d x %d\n",
			horizontal_size, vertical_size);
		fprintf(TraceFile, " macroblocks  (H x V): %d x %d\n",
			mb_width, mb_height);
		fprintf(TraceFile, " pel aspect ratio: %s\n",
			aspectTab[pel_aspect_ratio]);
		fprintf(TraceFile, " picture rate: %s pictures per second\n",
			(picture_rate<9) ? rateTab[picture_rate] : "reserved");
		if (bit_rate == 0x3ffff)
			fprintf(TraceFile, " bit rate: variable\n");
		else
			fprintf(TraceFile, " bit rate: %.1f kbits/sec\n", 0.4 * bit_rate);
		fprintf(TraceFile, " VBV buffer size: %d Kbyte\n", 2 * vbv_buffer_size);
		fprintf(TraceFile, " constrained parameter stream: %s\n",
			constrained_parameters_flag ? "yes" : "no");
		fprintf(TraceFile, " load intra quantizer matrix: %s\n",
			load_intra_quantizer_matrix ? "yes" : "no");
		if (load_intra_quantizer_matrix && (TraceLevel >= 2))
			for (index = 0; index < 64; index++)
				fprintf(TraceFile,"%3d%c", tQuantIntra[index],
					((index % 8) == 7) ? '\n' : ' ');
		fprintf(TraceFile, " load non intra quantizer matrix: %s\n",
			load_non_intra_quantizer_matrix ? "yes" : "no");
		if (load_non_intra_quantizer_matrix && (TraceLevel >= 2))
			for (index = 0; index < 64; index++)
				fprintf(TraceFile,"%3d%c", tQuant[index],
					((index % 8) == 7) ? '\n' : ' ');
	}

	Next_Start_Code();

	Extension_And_User_Data();
}

void    Extension_And_User_Data()
{
	unsigned int buf;

	Next_Bits(&buf, bStartCode);
	if (buf == cExtensionStartCode)
	{
		/* get extension data */
		Get_Bits(&buf, bStartCode);
		if (TraceLevel>=1)
			fprintf(TraceFile, "Extension Data:\n");
		Next_Bits(&buf, 24);
		while (buf != 0x000001)
		{
			Get_Bits(&buf, 8);
			if (TraceLevel >= 2)
				fprintf(TraceFile, "%02x", buf);
			Next_Bits(&buf, 24);
		}

		if (TraceLevel >= 2)
			putc('\n', TraceFile);

		Next_Start_Code();
	}
	
	Next_Bits(&buf, bStartCode);
	if (buf == cUserDataStartCode)
	{
		/* get user data */
		Get_Bits(&buf, bStartCode);
		if (TraceLevel>=1)
			fprintf(TraceFile, "User Data:\n");
		Next_Bits(&buf, 24);
		while (buf != 0x000001)
		{
			Get_Bits(&buf, 8);
			if (TraceLevel >= 2)
				fprintf(TraceFile, "%02x", buf);
			Next_Bits(&buf, 24);
		}

		if (TraceLevel >= 2)
			putc('\n', TraceFile);

		Next_Start_Code();
	}
}

