/* slice.c */

/*
 *      $Log:   procPicture.c,v $
 * Revision 1.5  93/12/07  12:00:00  sE
 *
 * Revision 1.2  93/06/15  15:41:31  oosa
 * 
 * 
 * Revision 1.1  1993/03/10  20:34:26  au
 * Initial revision
 *
 */

#include <stdio.h>
#include "consts.h"
#include "types.h"
#include "flc.h"
#include "vlc.h"
#include "global.h"
#include "decode.h"
#include "decode.pro"

int     Process_Slice(
	short           picture_coding_type,
	BYTE            *buf_Y,
	BYTE            *buf_Cb,
	BYTE            *buf_Cr,
	tData           *data,
	int                     forward_f_code,
	int                     fullPelForwardV,
	int                     backward_f_code,
	int                     fullPelBackwardV)
{
	unsigned int    buf;
	int     currentMBDiff[kMBLen];
	int     currentMBDiffU[kBLen];
	int     currentMBDiffV[kBLen];
	int     dct_dc_y_past, dct_dc_cb_past, dct_dc_cr_past;
	short   forwardMV[2];
	short   backwardMV[2];
	short   PMVFor[2];
	short   PMVBack[2];
	int     macroblock_type;
	int     bufOffsetY, bufOffsetC;
	int     MQuant, coded_block_pattern, MBA_incr= 0;
	int     index, row, code;
	int     slice_vertical_position, MBA;
	int     MB_row, MB_col;
	int     Skipped;

	/* remove slice start code */

	if (! Get_Bits(&buf, bStartCode))
	{
		printf ("Error (procSlice) -- No slice start code!\n");
		return(0);
	}

	slice_vertical_position = buf & 0x000000ff;
	MBA = (slice_vertical_position - 1) * mb_width - 1;

	/* get quantizer scale */
	MQuant = Get(bQuantizerScale);

	if (TraceLevel >= 2)
	{
		fprintf(TraceFile, "Slice Header: ");
		fprintf(TraceFile, " vertical position:%3d", slice_vertical_position);
		fprintf(TraceFile, " quantizer scale:%2d\n", MQuant);
	}

	/* skip extra_bit_slice & extra_information_slice */

	Get_Bits(&buf, 1);
	while(buf != 0)
	{
	    Get_Bits(&buf, 8);
	    Get_Bits(&buf, 1);
	}

	/* reset all predictors */
	dct_dc_y_past = dct_dc_cb_past = dct_dc_cr_past = 128;
	PMVFor[0] = PMVFor[1] = 0;
	PMVBack[0] = PMVBack[1] = 0;
	
	for (index = 0; ; index++)
	{

		if (MBA_incr == 0)
		{

			Next_Bits(&buf, 23);
			if (buf == 0)
			{
				Next_Start_Code();
				return(1);
			}

			/* get macroblock address increment */

			MBA_incr = Get_MBA_Inc();

			if (MBA_incr < 1)
			{
				printf ("Error (procSlice - MBA) - %d\n", index);
				return(0);
			}

			if (index == 0) /* first macroblock in slice */
			{
				MBA += MBA_incr; /* add horizontal offset */
				if (TraceLevel >= 2)
					fprintf(TraceFile, "               horizontal offset:%3d\n",
						MBA_incr - 1);
				MBA_incr = 1; /* don't skip first MB */
			}
		}

		Skipped = (MBA_incr != 1);

		MB_row = MBA / mb_width;
		MB_col = MBA % mb_width;

		bufOffsetY = MB_row * kMBHeight * (16 * mb_width) + MB_col * kMBWidth;
		bufOffsetC = MB_row * kBHeight  * (8 * mb_width) + MB_col * kBWidth;

		if (TraceLevel >= 3)
			fprintf(TraceFile,
				"  MB(%2d,%2d)", MB_row, MB_col);

		if (! Skipped)
		{

			/* get macroblock_type */

			switch (picture_coding_type)
			{
			case kIPicture:
				code = Get_Code(MB_typeI_length, MB_typeI_value, 2);
				if (code >= 0)
					macroblock_type = MB_typeI_switch[code];
				break;
			case kPPicture:
				code = Get_Code(MB_typeP_length, MB_typeP_value, 7);
				if (code >= 0)
					macroblock_type = MB_typeP_switch[code];
				break;
			case kBPicture:
				code = Get_Code(MB_typeB_length, MB_typeB_value, 11);
				if (code >= 0)
					macroblock_type = MB_typeB_switch[code];
				break;
			default:
				code = -1;
			}

			if (code == -1)
			{
				printf ("Error (procSlice - macroblock_type) - %d\n", index);
				return(0);
			}

			if (macroblock_type & kMbQuant)
			{
				/*      get quantizer_scale     */
				
				Get_Bits(&buf, 5);
				MQuant = buf;
			}

			/* reset motion vector predictors if MB is intra coded */

			if (macroblock_type & kMbIntra)
			{
				PMVFor[0] = 0;
				PMVFor[1] = 0;
				PMVBack[0] = 0;
				PMVBack[1] = 0;
			}

			/* get forward motion vector if present */

			if (macroblock_type & kMbForward)
			{
				if (! Get_MV(forward_f_code, fullPelForwardV,
						forwardMV, PMVFor))
				{
					printf ("Error (procSlice - Get_MV) - %d\n", index);
					return(0);
				}
				if (TraceLevel >= 3)
				{
					fprintf(TraceFile, " MVf(%3d,%3d)",
						forwardMV[0], forwardMV[1]);
				}
			}
			else if (TraceLevel >= 3)
				fprintf(TraceFile, "             ");

			/* get backward motion vector if present */

			if (macroblock_type & kMbBackward)
			{
				if (! Get_MV(backward_f_code, fullPelBackwardV,
						backwardMV, PMVBack))
				{
					printf ("Error (procSlice - Get_MV) - %d\n", index);
					return(0);
				}
				if (TraceLevel >= 3)
				{
					fprintf(TraceFile, " MVb(%3d,%3d)",
						backwardMV[0], backwardMV[1]);
				}
			}
			else if (TraceLevel >= 3)
				fprintf(TraceFile, "             ");

			if (TraceLevel >= 3)
			{
				fprintf(TraceFile, " q:%2d",MQuant);
				fprintf(TraceFile, ", type: ");
				if (macroblock_type & kMbIntra) putc('I', TraceFile);
				if (macroblock_type & kMbForward) putc('F', TraceFile);
				if (macroblock_type & kMbBackward) putc('B', TraceFile);
				if (macroblock_type & kMbPattern) putc('C', TraceFile);
				if (macroblock_type & kMbQuant) putc('Q', TraceFile);
				putc('\n', TraceFile);
			}

		}
		else if (TraceLevel >= 3)
				fprintf(TraceFile, "                                 type: Skipped\n");

		if ((!(macroblock_type & kMbIntra)) || Skipped)
		{
			/* MB is non-intra coded */

			/* reset DC predictors */
			dct_dc_y_past = 128;
			dct_dc_cb_past = 128;
			dct_dc_cr_past = 128;

			/*      reconstruct using MVs   */
			if (Skipped && (picture_coding_type == kPPicture))
				/* Skipped MBs in P picture don't have MVs */
				macroblock_type = 0;

			if ((macroblock_type & kMbForward) && (macroblock_type & kMbBackward))
			{
				/*      interpolated    */

				Construct_Bi_Pred(
						data->firstPredictYR + bufOffsetY,
						data->firstPredictUR + bufOffsetC,
						data->firstPredictVR + bufOffsetC,
						data->secondPredictYR + bufOffsetY,
						data->secondPredictUR + bufOffsetC,
						data->secondPredictVR + bufOffsetC,
						buf_Y + bufOffsetY,
						buf_Cb + bufOffsetC,
						buf_Cr + bufOffsetC,
						forwardMV, backwardMV);
			}
			else if (macroblock_type & kMbForward)
			{
				/*      forward */

				Construct_Mono_Pred( 
						data->firstPredictYR + bufOffsetY,
						data->firstPredictUR + bufOffsetC,
						data->firstPredictVR + bufOffsetC,
						buf_Y + bufOffsetY,
						buf_Cb + bufOffsetC,
						buf_Cr + bufOffsetC, forwardMV);
			}
			else if (macroblock_type & kMbBackward)
			{
				/*      backward        */

				Construct_Mono_Pred(
						data->secondPredictYR + bufOffsetY,
						data->secondPredictUR + bufOffsetC,
						data->secondPredictVR + bufOffsetC,
						buf_Y + bufOffsetY,
						buf_Cb + bufOffsetC,
						buf_Cr + bufOffsetC, backwardMV);
			}
			else if (picture_coding_type == kPPicture)
			{
				/*      no MC - P Picture only  */

				PMVFor[0] = 0;
				PMVFor[1] = 0;

				forwardMV[0] = 0;
				forwardMV[1] = 0;

				Construct_Mono_Pred(data->firstPredictYR + bufOffsetY,
						data->firstPredictUR + bufOffsetC,
						data->firstPredictVR + bufOffsetC,
						buf_Y + bufOffsetY,
						buf_Cb + bufOffsetC,
						buf_Cr + bufOffsetC, forwardMV);
			}
		}

		if (! Skipped)
		{
			/*      get coded block pattern */

			if (macroblock_type & kMbIntra)
				coded_block_pattern = 0x03F;          /*      all blocks      */
			else if (macroblock_type & kMbPattern)
			{
				if ((coded_block_pattern = Get_Code(CBP_length, CBP_value,
						66)) == -1)
				{
					printf ("Error (procSlice - coded_block_pattern) - %d\n",
							index);
					return(0);
				}
			}
			else
				coded_block_pattern = 0;

			/* clear DCT coefficient arrays */

			for (row = 0; row < kMBLen; row++)
				currentMBDiff[row] = 0;
	
			for (row = 0; row < kBLen; row++)
			{
				currentMBDiffU[row] = 0;
				currentMBDiffV[row] = 0;
			}

			if (! Process_MB(currentMBDiff, currentMBDiffU,
					currentMBDiffV, &dct_dc_y_past, &dct_dc_cb_past,
					&dct_dc_cr_past, macroblock_type, coded_block_pattern))
			{
				printf ("Error (procSlice) - after VLDing MB %d\n", index);
				return(0);
			}

			if (macroblock_type & kMbIntra)
				Inverse_Quantize_Intra_MB(coded_block_pattern, 
				 currentMBDiff, currentMBDiffU, currentMBDiffV,
				 MQuant);
			else
				Inverse_Quantize_Non_Intra_MB(coded_block_pattern, currentMBDiff, currentMBDiffU, currentMBDiffV, MQuant);

			Perform_IDCT(macroblock_type, currentMBDiff, currentMBDiffU, currentMBDiffV);

			/*      copy decoded data into picture buf      */

			Copy_From_Buf(macroblock_type, coded_block_pattern, MB_row, MB_col,
					buf_Y, buf_Cb, buf_Cr,
					currentMBDiff, currentMBDiffU, currentMBDiffV);

		}

		MBA_incr--;
		MBA++;
	}

	return(1);
}

int Get_MBA_Inc()
{
	int             code, MBA_incr;

	MBA_incr = 0;

	while (1)
	{
		if ((code = Get_Code(MBA_length, MBA_value, 35)) == -1)
			return(-1);

		if (code == 0)
			MBA_incr += 33;   /* macroblock_escape */
		else if (code != (35 - 1))
		{
			/* not macroblock_stuffing */
			MBA_incr += code;
			break;
		}
	}

	return(MBA_incr);
}

int     Get_MV(
	int     f_code,
	int     fullPelV,
	short   MV[2],
	short   PMV[])
{
	unsigned int    buf;
	int                             delta, sign, motionCodeMag, index;
	int                             f, low, high, range;

	f = 1 << (f_code - 1);
	low = -16*f;
	high = 16*f - 1;
	range = 32*f;

	/*      get diff vectors        */

	for (index = 0; index < 2; index++)
	{
		/*      get vlc_code_magnitude  */

		if ((motionCodeMag = Get_Code(MV_length, MV_value,
				33)) == -1)
		{
			printf ("Error (Get_Code)\n");
			return(0);
		}

		if (motionCodeMag > 16)
		{
			sign = -1;                      /*      negative delta  */
			motionCodeMag -= 16;
		}
		else
			sign = 1;

		/*      get residual    */
		if ((f_code != 1) && (motionCodeMag != 0))
			Get_Bits(&buf, f_code - 1);
		else
			buf = 0;

		if (motionCodeMag == 0)
			delta = 0;
		else
			delta = (motionCodeMag - 1) * f + buf + 1;

		MV[index] = PMV[index] + sign * delta;
	
		/*      adjust out of range MVs */
	
		if (MV[index] < low)
			MV[index] += range;
		else if (MV[index] > high)
			MV[index] -= range;
	
		PMV[index] = MV[index];
		if (fullPelV)
			MV[index] <<= 1;
	}

	return(1);
}
