/* writepic.c for version 1.5a of ISO MPEG-1 decoder  */

/*
 *      $Log:   WritePict.c,v $
 * Revision 1.5  93/12/07  12:00:00  sE
 *
 * Revision 1.1.1.1  93/03/29  11:27:54  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:33:17  au
 * Initial revision
 *
 */

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include "consts.h"
#include "types.h"
#include "decode.h"
#include "decode.pro"


int     Write_Picture(
	int                     aPictureNum,
	BYTE            *y,
	BYTE            *u,
	BYTE            *v,
	tParameter      *aParameter)
{
	int             ret;

	switch (aParameter->outputFormat)
	{
	case 0:
		ret = Write_SIF(aPictureNum, y, u, v, aParameter);
		break;
	case 1:
		ret = Write_TGA(aPictureNum, y, u, v, aParameter);
		break;
	case 2:
		ret = Write_PPM(aPictureNum, y, u, v, aParameter);
		break;
        case 3:
		ret = Write_YUV(aPictureNum, y, u, v, aParameter);
		break;
	}

	return(ret);
}


int     Write_SIF(
	int             aPictureNum,
	BYTE            *luminPictureR,
	BYTE            *chromUPictureR,
	BYTE            *chromVPictureR,
	tParameter      *aParameter)
{
	register        i,
				j;

	FILE    *fptr;
	char    fname[kMaxCharBufLen];
	BYTE    *buffer;
	BYTE    *b;
	BYTE    *y;
	BYTE    *u;
	BYTE    *v;
	BYTE    *tmpuR;
	BYTE    *tmpvR;
	int             numChromPixels, luminWidthHalf, luminWidth2;

	numChromPixels = (8 * mb_width) * (8 * mb_height);
	luminWidthHalf = horizontal_size / 2;
	luminWidth2 = horizontal_size * 2;
	tmpuR = (BYTE *) malloc(numChromPixels * sizeof(BYTE));
	tmpvR = (BYTE *) malloc(numChromPixels * sizeof(BYTE));
	buffer = (BYTE *) malloc(luminWidth2 * sizeof(BYTE));

	y = luminPictureR;
	u = chromUPictureR;
	v = chromVPictureR;

	for (i = 0; i < numChromPixels; i++)
	{
		tmpuR[i] = u[i];
		tmpvR[i] = v[i];
	}

	/* interpolate chrominance in Y direction */

	convert420to422(u, 8 * mb_width, 16* mb_height);
	convert420to422(v, 8 * mb_width, 16* mb_height);

	sprintf (fname, "%s%03d.SIF", aParameter->outputSequence, aPictureNum);

	printf("%s\n",fname);

	if((fptr = fopen(fname, "wb")) == NULL)
	{
		printf("\nopen error filename: %s\n", fname);
		return (1);
	}

	/* store picture */

	y = luminPictureR;
	u = chromUPictureR;
	v = chromVPictureR;

	for (i = 0; i < vertical_size; i++)
	{
		for (j = 0, b = buffer; j < luminWidthHalf; j++)
		{
			*(b++) = *(u++);
			*(b++) = *(y++);
			*(b++) = *(v++);
			*(b++) = *(y++);
		}

		fwrite(buffer, sizeof(BYTE), luminWidth2, fptr);
		y += 16 * mb_width - 2 * luminWidthHalf;
		u += 8 * mb_width - luminWidthHalf;
		v += 8 * mb_width - luminWidthHalf;
	}

	fclose(fptr);

	u = chromUPictureR;
	v = chromVPictureR;

	for (i = 0; i < numChromPixels; i++)
	{
		u[i] = tmpuR[i];
		v[i] = tmpvR[i];
	}

	free(tmpuR);
	free(tmpvR);
	free(buffer);

	return (0);
}

int     Write_TGA(
	int                     aPictureNum,
	BYTE            *y,
	BYTE            *u,
	BYTE            *v,
	tParameter      *aParameter)
{
	int             i, j;
	static unsigned char    tga24[14]={0,0,2,0,0,0,0, 0,0,0,0,0,24,32};

	FILE    *fptr;
	char    fname[kMaxCharBufLen];
	BYTE    *u2;
	BYTE    *v2;
	int             r, g, b;

	u2 = (BYTE *) malloc((16*mb_width) * (16*mb_height) * sizeof(BYTE));
	v2 = (BYTE *) malloc((16*mb_width) * (16*mb_height) * sizeof(BYTE));

	/* interpolate chrominance in both directions */

	convert420to444(u, u2, 8 * mb_width, 8 * mb_height);
	convert420to444(v, v2, 8 * mb_width, 8 * mb_height);

	sprintf (fname, "%s%03d.tga", aParameter->outputSequence, aPictureNum);

	printf("%s\n",fname);

	if((fptr = fopen(fname, "wb")) == NULL)
	{
		printf("\nopen error filename: %s\n", fname);
		return (1);
	}

	/* write Targa header */
	for (i = 0; i < 12; i++)
		putc(tga24[i], fptr);

	putc(horizontal_size, fptr);
	putc(horizontal_size >> 8, fptr);
	putc(vertical_size, fptr);
	putc(vertical_size >> 8,fptr);
	putc(tga24[12], fptr);
	putc(tga24[13], fptr);
	
	/* store picture */

	for (j = 0; j < vertical_size; j++)
	{
		for (i = 0; i < horizontal_size; i++)
		{

			yuvtorgb(*y++, *u2++, *v2++, &r, &g, &b);

			putc(b, fptr);
			putc(g, fptr);
			putc(r, fptr);
		}

		y  += 16 * mb_width - horizontal_size;
		u2 += 16 * mb_width - horizontal_size;
		v2 += 16 * mb_width - horizontal_size;
	}

	fclose(fptr);

	free(u2);
	free(v2);

	return (0);
}

int     Write_PPM(
	int                     aPictureNum,
	BYTE            *y,
	BYTE            *u,
	BYTE            *v,
	tParameter      *aParameter)
{
	int             i, j;

	FILE    *fptr;
	char    fname[kMaxCharBufLen];
	BYTE    *u2;
	BYTE    *v2;
	int             r, g, b;

	u2 = (BYTE *) malloc((16*mb_width) * (16*mb_height) * sizeof(BYTE));
	v2 = (BYTE *) malloc((16*mb_width) * (16*mb_height) * sizeof(BYTE));

	/* interpolate chrominance in both directions */

	convert420to444(u, u2, 8 * mb_width, 8 * mb_height);
	convert420to444(v, v2, 8 * mb_width, 8 * mb_height);

	sprintf (fname, "%s%03d.ppm", aParameter->outputSequence, aPictureNum);

	printf("%s\n",fname);

	if((fptr = fopen(fname, "wb")) == NULL)
	{
		printf("\nopen error filename: %s\n", fname);
		return (1);
	}

	/* write PPM header */
	fprintf(fptr,"P6\n%d %d\n255\n",horizontal_size,vertical_size);
	
	/* store picture */

	for (j = 0; j < vertical_size; j++)
	{
		for (i = 0; i < horizontal_size; i++)
		{

			yuvtorgb(*y++, *u2++, *v2++, &r, &g, &b);

			putc(r, fptr);
			putc(g, fptr);
			putc(b, fptr);
		}

		y  += 16 * mb_width - horizontal_size;
		u2 += 16 * mb_width - horizontal_size;
		v2 += 16 * mb_width - horizontal_size;
	}

	fclose(fptr);

	free(u2);
	free(v2);

	return (0);
}

void    yuvtorgb(
	int             y,
	int             u,
	int             v,
	int             *pr,
	int             *pg,
	int             *pb)
{
	int             r, g, b;
	double  yf, uf, vf, rmy, gmy, bmy, rf, gf, bf;

	/* scale to 0..1 (Y) / -1..+1 (U,V) */

	yf = (y - 16) / (double)(235-16);
	uf = (u - 128) / (double)(240-16);
	vf = (v - 128) / (double)(240-16);

	/* compute color differences R-Y, G-Y, B-Y */

	rmy = ((1.0-0.299)/0.5) * vf; /* 1.402*vf */
	bmy = ((1.0-0.114)/0.5) * uf; /* 1.772*uf */
	gmy = -(0.299*rmy + 0.114*bmy)/0.587;

	/* compute color primaries R, G, B */

	rf = yf + rmy;
	bf = yf + bmy;
	gf = yf + gmy;

	/* scale to 0..255 */

	r = floor(256.0 * rf + 0.5);
	if (r < 0)
		r = 0;
	if (r > 255)
		r = 255;
	g = floor(256.0 * gf + 0.5);
	if (g < 0)
		g = 0;
	if (g > 255)
		g = 255;
	b = floor(256.0 * bf + 0.5);
	if (b < 0)
		b = 0;
	if (b > 255)
		b = 255;

	*pr = r;
	*pg = g;
	*pb = b;

}

int     Write_YUV(
	int             pictureNum,
	BYTE            *y_recon,
	BYTE            *cb_recon,
	BYTE            *cr_recon,
	tParameter      *parameter)
{
	FILE    *fptr;
	char    fname[kMaxCharBufLen];
	BYTE    *buffer;
	BYTE    *b;
	BYTE    *y;
	BYTE    *u;
	BYTE    *v;
	register line, col;
	int     lw, lh, cw, ch;

	lw = horizontal_size;
	lh = vertical_size;
	cw = lw >> 1;   /* chroma width is always half luma for 4:2:0 */
	ch = lh >> 1;   /* ditto for height */

	/* allocate "line buffer" of greatest common multiple dimensions */
	buffer = (BYTE *) malloc(lw * sizeof(BYTE));

	sprintf (fname, "%s%03d.yuv", parameter->outputSequence, pictureNum);
	printf("%s\n",fname);

	if((fptr = fopen(fname, "wb")) == NULL)
	{
		printf("\nopen error filename: %s\n", fname);
		return (-1);
	}

	/* store picture */

	y = y_recon;
	
	for (line = 0; line <  lh; line++)
	{
		for (col = 0, b = buffer; col < lw; col++)
		{
			*(b++) = *(y++);
		}
		fwrite(buffer, sizeof(BYTE), lw, fptr);
		y += (16 * mb_width) - lw;

	}

	u = cb_recon;
		
	for (line = 0; line < ch; line++)
	{
		for (col = 0, b = buffer; col < cw; col++)
		{
			*(b++) = *(u++);
		}

		fwrite(buffer, sizeof(BYTE), cw, fptr);
		u += (8 * mb_width) - cw;

	}
	       

	v = cr_recon;

	for (line = 0; line < ch; line++)
	{
		for (col = 0, b = buffer; col < cw; col++)
		{
			*(b++) = *(v++);
		}

		fwrite(buffer, sizeof(BYTE), cw, fptr);
		v += (8 * mb_width) - cw;

	}

	free(buffer);
	fclose(fptr);
	return (0);
}
