/* File: convert.c */

/*
 *      $Log:   convert.c,v $
 * Revision 1.1.1.1  93/03/29  11:27:55  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:33:17  au
 * Initial revision
 *
 */

#include <stdio.h>
/*#include <string.h>*/
#include "consts.h"
#include "mpeg1.h"
#include "encoder.pro"

/* transfer function constants */
static  int hf0[7] ={-29, 0, 88, 138, 88, 0, -29};
#define sf0             256
#define sf0Half         128     /*  sf0 / 2  */
#define Nf0L            -3 
#define Nf0H            3
#define Nf0Z            3     /*implies that 138 multiplies Oth line of fl0 */

static int hf1[4] ={1, 7, 7, 1};
#define sf1             16
#define sf1Half         8       /* sf1 / 2 */
#define Nf1L            -2
#define Nf1H            1
#define Nf1Z            2   /* implies that 2st 7 multiplies Oth line of fl1 */

void    convert422to420(
	int             fieldNum,
	BYTE            *chromPicture,
	int             chromWidth,
	int             chromHeight,
	int             direction,
	tParameter      *parameter)
{
	int             i;
	int             j;
	static int      *a, *b;
	void            filter();

	if(!a)
	{
		a = Iarray(parameter->luminHeight);
		b = Iarray(parameter->luminHeight);
	}

	switch(direction)
	{
		case FORWARD:

			for(j = 0; j < chromWidth; j++)
			{
				for(i = 0; i < chromHeight; i++)
				{
					*(a + i) = *(chromPicture + i * chromWidth + j);
				}

				filter(fieldNum, chromHeight, a, b, FORWARD, parameter);

				for(i = 0; i < chromHeight / 2; i++)
				{
					*(chromPicture + i * chromWidth +j) = *(b + i);
				}
			}

			break;

		case INVERSE:

			for(j = 0; j < chromWidth; j++)
			{
				for(i = 0; i < chromHeight / 2; i++)
				{
					*(a + i) = *(chromPicture + i * chromWidth + j);
				}

				filter(fieldNum, chromHeight / 2, a, b, INVERSE,parameter);

				for(i = 0; i < chromHeight; i++)
				{
					*(chromPicture + i * chromWidth + j) = *(b + i);
				}
			}

			break;
	}
}

void    filter(
	int     fieldNum,
	int     nn,
	int     *FF,
	int     *GG,
	int     direction,
        tParameter    *parameter)
{
	static int *F, *G;
	int i, k, l, n, M, L, sgn;
	int     tmp;

	if (!F)
	{
		F = Iarray(parameter->luminHeight);
		G = Iarray(parameter->luminHeight);
	}

	switch(direction)
	{
		case FORWARD:

			n = nn / 2;
			M = nn;

			for(i = 0; i < M; i++)
			{
				F[i] = FF[i];
				GG[i] = 0;       /* OK since dimentioned same */
			}

			switch(fieldNum)
			{
				case FIELD_1:

					for(i = 0; i < n; i++)
					{
						k = 2 * i;

						G[i] = 0;
						G[n+i] = 0;

						for(l = Nf0L; l < (Nf0H + 1); l++)
						{
							L = k - l;

							if(L < 0)
								L += M;
							else if (L > (M - 1))
								L -= M;

							G[i] += hf0[l + Nf0Z] * F[L];
						}

						sgn = (G[i] < 0) ? -1 : 1;

						if (G[i] < 0)
							GG[i] = 0;
						else
						{
							tmp = (G[i] + sf0Half) / sf0;

							GG[i] = (tmp > 255) ? 255 : tmp;
						}
					}

					break;  /* end Field 0 */

				case FIELD_2:

					for(i = 0; i < n; i++)
					{
						k = 2 * i;
						G[i] = 0;
						G[n+i] = 0;

						for(l = Nf1L; l < (Nf1H + 1); l++)
						{
							L = k - l;
							if(L < 0)
								L += M;
							else if(L > (M - 1))
								L -= M;
							G[i] += hf1[l + Nf1Z] * F[L];
						}

						sgn = (G[i] < 0) ? -1 : 1;

						GG[i] = (short)(sgn * ((sgn * G[i] + sf1Half) / sf1));
					}  /* end of Field 1 Forward direction */

					break;
			}

			break;

		case INVERSE:

			n = nn;
			M = 2 * n;

			for(i = 0; i < M; i++)
			{
				F[i] = FF[i];
				G[i] = 0;
			}

			switch(fieldNum)
			{
				case FIELD_1:

					F[n] = F[n - 1];  /* for 'interpolation' of last point */

					for(i = 0; i < n; i++)
					{
						k = 2 * i;
						GG[k] = (short)F[i];
						G[k + 1] = F[i] + F[i + 1];

						/* normalize and round */

						sgn = (G[k + 1] < 0) ? -1 : 1;
						GG[k + 1] = (short)(sgn * ((sgn * G[k + 1] + 1) / 2));
					}

					break;

				case FIELD_2:

					F[n] = F[0];  /* for 'wrap filter wrap around' */

					for(i = 0; i < n; i++)
					{
						k = 2 * i;
						G[k] = 3 * F[i] + F[i + 1];
						G[k + 1] = F[i] + 3 * F[i + 1];
					}

					/* normalize and round */

					for(i = 0; i < 2 * n; i++)
					{
						sgn = (G[i] < 0) ? -1 : 1;
						GG[i] = (short)(sgn * ((sgn * G[i] + 2) / 4));
					}

					break;
			} /* end INVERSE */

			break;
	}

	return;
}

void    convert420to422(
	BYTE    *chromPicture,
	int     chromWidth,
	int     chromHeight)
{
	int     i;
	int     j;
	int     *a;
	int     *b;

	a = (int *) malloc(chromHeight * sizeof(int));
	b = (int *) malloc(chromHeight * sizeof(int));

	for(j = 0; j < chromWidth; j++)
	{
		for(i = 0; i < chromHeight / 2; i++)
			*(a + i) = *(chromPicture + i * chromWidth + j);

		filter2(chromHeight / 2, a, b);

		for(i = 0; i < chromHeight; i++)
			*(chromPicture + i * chromWidth + j) = *(b + i);
	}

	free(a);
	free(b);
}

void    filter2(
	int     n,
	int     *F,
	int     *G)
{
	int i, k, M;
	int     tmp;


	M = 2 * n;
	
	F[n] = F[n - 1];  /* for 'interpolation' of last point */

	for(i = 0; i < n; i++)
	{
		k = 2 * i;
		G[k] = F[i];

		tmp = F[i] + F[i + 1];

		/* normalize and round */
		G[k + 1] = (tmp >= 0) ? ((tmp + 1) / 2) : -((-tmp + 1) / 2);
	}
	
	return;
}

void convert420to444(
	BYTE    *chrom420,
	BYTE    *chrom444,
	int     chromWidth,
	int     chromHeight)
{
	int     chromWidth2, i, j;
	BYTE    *src, *dst;

	/* bilinear interpolation filter */

	chromWidth2 = 2 * chromWidth;

	/* horizontal direction */

	for (j = 0; j < chromHeight; j++)
	{
		src = chrom420 + j * chromWidth;
		dst = chrom444 + j * chromWidth2;
		dst[0] = src[0];
		for (i = 1; i < chromWidth; i++)
		{
			dst[2*i-1] = (3 * src[i-1] + src[i] + 2) >> 2;
			dst[2*i]   = (src[i-1] + 3 * src[i] + 2) >> 2;
		}
		dst[chromWidth2-1] = src[chromWidth-1];
	}

	/* vertical direction, (in-place, bottom-up) */

	for (i = 0; i < chromWidth2; i++)
	{
		src = chrom444 + i + (chromHeight-2) * chromWidth2;
		dst = chrom444 + i + (2*(chromHeight-2)+1) * chromWidth2;
		dst[2*chromWidth2] = src[chromWidth2];
		for (j = chromHeight - 2; j >= 0; j--)
		{
			dst[chromWidth2] = (src[0] + 3 * src[chromWidth2] + 2) >> 2;
			dst[0] = (3 * src[0] + src[chromWidth2] + 2) >> 2;
			dst -= 2 * chromWidth2;
			src -= chromWidth2;
		}
	}
}
