/* file: dct.c */

/*
 *      $Log:   dct.c,v $
 * Revision 1.1.1.1  93/03/29  11:28:00  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:30:43  au
 * Initial revision
 *
 */

#include <math.h>
#include "consts.h"

#define PI      3.141592654

double          Forward_Cosine_Matrix[kDCTSize][kDCTSize];
double          Inverse_Cosine_Matrix[kDCTSize][kDCTSize];


/************************************************************************
*
* Module name: Init_DCT
*
* Initializes the DCT scale and twiddle factor matrix
*
************************************************************************/

void Init_DCT(void)
{
	extern double   Forward_Cosine_Matrix[][kDCTSize];
	extern double   Inverse_Cosine_Matrix[][kDCTSize];

	/* Local variables */

	double  normalization_vector[kDCTSize];
	short freq, time;

	/* Execute */

	/* DC term has different normalization scale than AC terms */
	normalization_vector[0] = sqrt((double)(1.0 / kDCTSize));

	/* Compute normalization scales for fundamentals and harmonics */
	for(freq = 1; freq < kDCTSize; freq++)
		normalization_vector[freq] = sqrt((double)(2.0 / kDCTSize));

	for(freq = 0; freq < kDCTSize; freq++)
	{
		for(time = 0; time < kDCTSize; time++)
		{
			Forward_Cosine_Matrix[freq][time] = normalization_vector[freq] 
				* cos((double)((2*time +1) * freq * PI)/(2*kDCTSize));

			Inverse_Cosine_Matrix[freq][time] = normalization_vector[time] 
				* cos((double)((2*freq +1) * time * PI)/(2*kDCTSize));
		}
	}
}




/* forward DCT */

void dct8x8(
	float   sample_block[],     /* input block -- spatial domain */
	float   coefficient_block[])      /* output block -- frequency domain*/
{
	extern double   Forward_Cosine_Matrix[][kDCTSize];  /* cosine matrix */

	/* Local variables */

	short   row, col, freq, time, index, index1;
	float   intermediate[kDCTLen];

	/*      Transform rows  */

	for (row = 0, index = 0; row < kDCTSize; row++, index += kDCTSize)
	{
		for(freq = 0; freq < kDCTSize; freq++)
		{
			intermediate[index + freq] = 0;

			for(time = 0; time < kDCTSize; time++)
			{
				intermediate[index + freq] += sample_block[index + time] 
					* Forward_Cosine_Matrix[freq][time];
			}
		}
	}
	

	/*  transform colums.  Transpose operation is integrated into  
	    the indexing of intermediate matrix */

	for (col = 0; col < kDCTSize; col++)  
	{
		for(freq = 0, index = col; freq < kDCTSize; freq++, index += kDCTSize)
		{
			coefficient_block[index] = 0;

			for(time = 0, index1 = col; time < kDCTSize; time++, index1 += kDCTSize)
			{
				coefficient_block[index] += intermediate[index1] 
					* Forward_Cosine_Matrix[freq][time];
			}
		}
	}
}


/************************************************************************
*
* Module name: IDCT
*
* Calculates an inverse discrete cosine transform
*
*************************************************************************/

void idct8x8(
	float   coefficient_block[],
	float   recon_block[])
{
	extern double   Inverse_Cosine_Matrix[][kDCTSize];

	/* Local variables */

	short   row, col, freq, time, index, index1;
	float   intermediate[kDCTLen];

	/* Execute */

	for(row = 0, index = 0; row < kDCTSize; row++, index += kDCTSize)
	{
		for(time = 0; time < kDCTSize; time++)
		{
			intermediate[index + time] = 0;

			for(freq = 0; freq < kDCTSize; freq++)
			{
				intermediate[index + time] += coefficient_block[index + freq] 
					* Inverse_Cosine_Matrix[time][freq];
			}
		}
	}

	for(col = 0; col < kDCTSize; col++)
	{
		for(time = 0, index = col; time < kDCTSize; time++, index += kDCTSize)
		{
			recon_block[index] = 0;

			for(freq = 0, index1 = col; freq < kDCTSize; freq++, index1 += kDCTSize)
			{
				recon_block[index] += intermediate[index1] 
					* Inverse_Cosine_Matrix[time][freq];
			}
		}
	}
}
