/* File:   decision.c  */
/* New constants need updating.  */

/*
 *      $Log:   decision.c,v $
 * Revision 1.1.1.1  93/03/29  11:27:59  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:30:43  au
 * Initial revision
 *
 */

#include <stdio.h>
#include "consts.h"
#include "mpeg1.h"
#include "encoder.pro"


/* construct intra macorblock */
int     ConstructI(
	BYTE    *constY,
	BYTE    *constU,
	BYTE    *constV,
	BYTE    *current,
	BYTE    *currentU,
	BYTE    *currentV,
	int     *diff,
	int     *diffU,
	int     *diffV,
	tParameter      *parameter)
{
	int     index1, index2;
	BYTE    *constY2, *constU2, *constV2;
	BYTE    *currentIndex, *currentIndexU, *currentIndexV;
	int     *diffIndex, *diffIndexU, *diffIndexV;

	/*      reconstruct with intra  */

    currentIndex = current;
	diffIndex = diff;
	constY2 = constY;

	/*      lumin   */
	
	for (index1 = 0; index1 < kMBHeight; index1++, 
			constY2 += parameter->luminWidthMB)
	{
		for (index2 = 0; index2 < kMBWidth; index2++,
				constY2++, currentIndex++, diffIndex++)
		{
			*constY2 = *currentIndex;
			*diffIndex = *currentIndex;
		}
	}

	/*      chrom   */

	currentIndexU = currentU;
	currentIndexV = currentV;
	diffIndexU =    diffU;
	diffIndexV =    diffV;
	constU2 =       constU;
	constV2 =       constV;


	/* need to change kBHeight to something more clear */
	for (index1 = 0; index1 < kBHeight; index1++,
			constU2 += parameter->chromWidthB, 
				constV2 += parameter->chromWidthB)
	{
		for (index2 = 0; index2 < kBWidth; index2++,
				constU2++, constV2++, currentIndexU++, currentIndexV++,
				diffIndexU++, diffIndexV++)
		{
			*constU2 =      *currentIndexU;
			*constV2 =      *currentIndexV;
			*diffIndexU =   *currentIndexU;
			*diffIndexV =   *currentIndexV;
		}
	}

	return(kIntraType);             /*      intra   */
}


/* Make P picture macroblock type decision */

int     Construct(
	BYTE    predictY[],
	BYTE    predictU[],
	BYTE    predictV[],
	BYTE    *constY,
	BYTE    *constU,
	BYTE    *constV,
	BYTE    current[],
	BYTE    currentU[],
	BYTE    currentV[],
	short   MV[],
	int     *diff,
	int     *diffU,
	int     *diffV,
	tParameter     *parameter)
{
	int     mbType = 0;   
	
	/*  bit:
	 *  1 - MC(1)/No MC(0)
	 *  2 - Intra(1)/Inter(0)
	 *  5 - ForwardMV
	 */
	
	int     index1, index2;
	int     SEMC, SENoMC;
	BYTE    *constY2, *constU2, *constV2;
	BYTE    *currentIndex, *currentIndexU, *currentIndexV;
	BYTE    *frameIndex, *frameIndexU, *frameIndexV;
	int     *diffIndex, *diffIndexU, *diffIndexV;

	/* frame SE (Square Error) */

	SEMC = CalSE(current, predictY, MV, parameter);

	currentIndex = current;
	frameIndex = predictY;

	/* frame SE no MC  */

	SENoMC = 0;

	for (index1 = 0; index1 < kMBHeight; index1++,
			frameIndex += parameter->luminWidthMB)
	{
		for (index2 = 0; index2 < kMBWidth; index2++,
				currentIndex++, frameIndex++)
		{
			SENoMC += (*currentIndex - *frameIndex) *
					(*currentIndex - *frameIndex);
		}
	}

	/* MC/no MC check  */

	if (SENoMC > SEMC)
	{
		mbType |= kMCType;      /* MC */
		mbType |= kForwardMV;   /* forward MV */

		/* reconstruct MB  */


		FrameConstr(predictY, predictU, predictV, constY, constU, 
			constV,MV, current, currentU, currentV, diff, diffU, 
			diffV, parameter);
	}
	else    /* No MC */
	{
		constY2 = constY;
		constU2 = constU;
		constV2 = constV;

		frameIndex = predictY;
		currentIndex = current;
		diffIndex = diff;

		/* lumin */

		for (index1 = 0; index1 < kMBHeight; index1++,
				constY2 += parameter->luminWidthMB, 
					frameIndex += parameter->luminWidthMB)
		{
			for (index2 = 0; index2 < kMBWidth; index2++,
					constY2++, frameIndex++, currentIndex++,
					diffIndex++)
			{
				*constY2 = *frameIndex;
				*diffIndex = *currentIndex - *constY2;
			}
		}

		/*      chrom   */

		frameIndexU =   predictU;
		frameIndexV =   predictV;
		currentIndexU = currentU;
		currentIndexV = currentV;
		diffIndexU =    diffU;
		diffIndexV =    diffV;

		for (index1 = 0; index1 < kBHeight; index1++,
				constU2 += parameter->chromWidthB, 
				constV2 += parameter->chromWidthB,
				frameIndexU += parameter->chromWidthB, 
				frameIndexV += parameter->chromWidthB)
		{
			for (index2 = 0; index2 < kBWidth; index2++,
					constU2++, constV2++, frameIndexU++, frameIndexV++,
					currentIndexU++, currentIndexV++,
					diffIndexU++, diffIndexV++)
			{
				*constU2 = *frameIndexU;
				*constV2 = *frameIndexV;
				*diffIndexU = *currentIndexU - *constU2;
				*diffIndexV = *currentIndexV - *constV2;
			}
		}
	}

	if(IntraCheck(current, currentU, currentV, constY, constU, constV,
			diff, diffU, diffV, parameter))
	{
		mbType = kIntraType;
	}

	return (mbType);
}


/* Make intra/non-intra decision */
int     IntraCheck(
	BYTE    current[],
	BYTE    currentU[],
	BYTE    currentV[],
	BYTE    *constY,
	BYTE    *constU,
	BYTE    *constV,
	int     *diff,
	int     *diffU,
	int     *diffV,
	tParameter      *parameter)
{
	BYTE    *currentIndex;
	int     var, varor, sum, *diffIndex;
	short   index1;

	/* intra/inter coding check */

	currentIndex = current;
	diffIndex = diff;
	var = 0;
	varor = 0;
	sum = 0;

	for (index1 = 0; index1 < kMBLen; index1++, currentIndex++, diffIndex++)
	{
		sum += *currentIndex;
		varor += *currentIndex * *currentIndex;
		var += *diffIndex * *diffIndex;
	}

	var /= 256;             /*  var/256  */
	varor /= 256;           /*  varor/256  */
	sum /= 256;             /*  sum/256  */
	varor -= (sum * sum);   /*  varor/256 - sum/256 * sum/256  */
	
	if ((var > 64) && (var > varor))
	{
		/* reconstruct with intra  */

		return (ConstructI (constY, constU, constV, current, 
			currentU, currentV, diff, diffU, diffV, parameter));
	}

	return (0);     /*  non intra */
}

/* make B picture macroblock decision */
int     ConstructB(
	BYTE    predictY[],
	BYTE    predictU[],
	BYTE    predictV[],
	BYTE    predictY2[],
	BYTE    predictU2[],
	BYTE    predictV2[],
	BYTE    *constY,
	BYTE    *constU,
	BYTE    *constV,
	BYTE    current[],
	BYTE    currentU[],
	BYTE    currentV[],
	short   MV[],
	short   MVB[],
	int     *diff,
	int     *diffU,
	int     *diffV,
	tParameter      *parameter)
{             
	int     mbType = kMCType;
	int     SEMC, FSEMC, BSEMC, ISEMC;

	
	FSEMC = CalSE(current, predictY, MV, parameter);
	BSEMC = CalSE(current, predictY2, MVB, parameter);
	ISEMC = ICalSE(current, predictY, predictY2, MV, MVB, parameter);

	if ((FSEMC <= BSEMC) && (FSEMC <= ISEMC))
	{
		SEMC = 0;       /*      forward         */
	}
	else if ((BSEMC < FSEMC) && (BSEMC <= ISEMC))
	{
		SEMC = 1;       /*      backward        */
	}
	else
	{
		SEMC = 2;       /*      interpolate     */
	}

	switch (SEMC)
	{
		case 0:         /*      forward         */

			mbType |= kForwardMV;

			FrameConstr(predictY, predictU, predictV, constY, 
					constU, constV, MV, current, currentU, 
					currentV, diff, diffU, diffV, parameter);
			break;

		case 1:         /*      backward        */

			mbType |= kBackwardMV;

			FrameConstr(predictY2, predictU2, predictV2, constY, 
					constU, constV, MVB, current, currentU, 
					currentV, diff, diffU, diffV, parameter);
			break;

		case 2:         /*      interpolate     */

			mbType |= kForwardMV;
			mbType |= kBackwardMV;

			FrameIConstr(predictY, predictU, predictV, predictY2, 
					predictU2, predictV2, constY, constU,
					constV, MV, MVB, current, currentU, 
					currentV, diff, diffU, diffV, parameter);
			break;
	}

	/*      intra/inter coding check        */

	if(IntraCheck(current, currentU, currentV, constY, constU, constV,
			diff, diffU, diffV, parameter))
	{
		mbType = kIntraType;
	}

	return (mbType);
}


/* Calculate Squared Error between two macroblocks */

int     CalSE(
	BYTE    *current,
	BYTE    *predictY,
	short   MV[],
	tParameter      *parameter)
{
	BYTE    *frameIndex,*currentIndex;
	int     offset, diffByte;
	int     SEFrame;
	short   f0off2, f0off3, f0off4, index1, index2, mvx, mvy;

	/*      calculate SEs   */

	mvx = MV[0] / 2;
	mvy = MV[1] / 2;
	offset = (mvy * parameter->luminWidth) + mvx;
	frameIndex = (BYTE *) (predictY + offset);

	GetLoc(MV, &f0off2, &f0off3, &f0off4, parameter->luminWidth);

	currentIndex = current;

	/*      frame SE (Square Error) */

	SEFrame = 0;

	for (index1 = 0; index1 < kMBHeight; index1++,
			frameIndex += parameter->luminWidthMB)
	{
		for (index2 = 0; index2 < kMBWidth; index2++,
				currentIndex++, frameIndex++)
		{
			/*      add 2 for integer rounding      */

			diffByte = ((int) *frameIndex +
					(int) *(frameIndex + f0off2) +
					(int) *(frameIndex + f0off3) +
					(int) *(frameIndex + f0off4) + 2) / 4;
			diffByte = *currentIndex - diffByte;

			SEFrame += diffByte * diffByte;
		}
	}

	return(SEFrame);
}


/* Calculate Interpolated (Bi-directional) Square Error */
int     ICalSE(
	BYTE    *current,
	BYTE    *predictY,
	BYTE    *predictYB,
	short   MV[],
	short   MVB[],
	tParameter      *parameter)
{
	BYTE    *frameIndex, *frameIndexB, *currentIndex;
	short   index1, index2, mvx, mvy;
	short   f0off2, f0off3, f0off4;
	short   f0off2B, f0off3B, f0off4B;
	int     offset, diffByte, SEFrame;

	/*      calculate SEs   */

	mvx = MV[0] / 2;
	mvy = MV[1] / 2;
	offset = (mvy * parameter->luminWidth) + mvx;
	frameIndex = (BYTE *) (predictY + offset);

	GetLoc(MV, &f0off2, &f0off3, &f0off4, parameter->luminWidth);

	mvx = MVB[0] / 2;
	mvy = MVB[1] / 2;
	offset = (mvy * parameter->luminWidth) + mvx;
	frameIndexB = (BYTE *) (predictYB + offset);

	GetLoc(MVB, &f0off2B, &f0off3B, &f0off4B, parameter->luminWidth);

	currentIndex = current;

	/*      frame SE (Square Error) */

	SEFrame = 0;

	for (index1 = 0; index1 < kMBHeight; index1++,
			frameIndex += parameter->luminWidthMB, 
				frameIndexB += parameter->luminWidthMB)
	{
		for (index2 = 0; index2 < kMBWidth; index2++,
				currentIndex++, frameIndex++, frameIndexB++)
		{
			/*      add 4 for integer rounding      */

			diffByte = ((int) *frameIndex +
					(int) *(frameIndex + f0off2) +
					(int) *(frameIndex + f0off3) +
					(int) *(frameIndex + f0off4) +
					(int) *frameIndexB +
					(int) *(frameIndexB + f0off2B) +
					(int) *(frameIndexB + f0off3B) +
					(int) *(frameIndexB + f0off4B) + 4) / 8;
			diffByte = *currentIndex - diffByte;

			SEFrame += diffByte * diffByte;
		}
	}

	return (SEFrame);
}
