/*  file: genbits.c */

/* 
 *      $Log:   ratectrl.c,v $
 * Revision 1.4  93/09/03  01:15:00  hana
 * VBV buffer operation 
 * 
 * Revision 1.2  93/06/15  15:36:31  oosa
 * 
 * 
 * Revision 1.1  1993/03/10  20:30:43  au
 * Initial revision
 *
 */

#include <stdio.h>
#include "consts.h"
#include "mpeg1.h"
#include "flc.h"
#include "bitcount.h"
#include "global.h"
#include "encoder.pro"


/* write sequence header and update its cost to ratecontrol */
void    Write_Sequence_Header(
	tParameter *parameter,
	tRateControl *rateControl)

{
	extern BYTE tQuantIntra[], tQuant[];

	int     index,
		index1;

	rateControl->headerBits += bSequenceHeaderCode;
	writeBits(cSequenceHeaderCode, bSequenceHeaderCode, 1);
	
	rateControl->headerBits += bHorizontalSize;
	writeBits(parameter->horizontal_size, bHorizontalSize, 0);
	
	rateControl->headerBits += bVerticalSize;
	writeBits(parameter->vertical_size, bVerticalSize, 0);
	
	rateControl->headerBits += bPelAspectRatio;
	writeBits(parameter->pel_aspect_ratio, bPelAspectRatio, 0);
	
	rateControl->headerBits += bPictureRate;
	writeBits(parameter->picture_rate, bPictureRate, 0);

	rateControl->headerBits += bBitRate;
	
	/* convert internal type to bitstream representation */
	index1 = parameter->bit_rate;
	index = index1 / cBitRate;
	if (index1 % cBitRate)
		index++;
	writeBits(index, bBitRate, 0);
	
	rateControl->headerBits += bMarkerBit;
	writeBits(cMarkerBit, bMarkerBit, 0);
	
	/* convert internal buffer size representation to bitstream type */
	rateControl->headerBits += bVbvBufferSize;
	index = parameter->vbv_buffer_size / 16384 ; /* 16384 = 16 * 1024 */
	writeBits(index, bVbvBufferSize, 0);
	
	rateControl->headerBits += bConstrainedParameterFlag;
	writeBits(parameter->constrained_parameters_flag, bConstrainedParameterFlag, 0);
	
	rateControl->headerBits += bLoadIntraQuantizerMatrix;
	writeBits(parameter->load_intra_quantizer_matrix, bLoadIntraQuantizerMatrix, 0);

	if(parameter->load_intra_quantizer_matrix) {
		/*      loadIntraQuantizerMatrix = TRUE */
	
		rateControl->headerBits += 64 * bIntraQuantizer;
		for (index = 0; index < 64; index++)
			writeBits(tQuantIntra[index], bIntraQuantizer, 0);
	}               

	rateControl->headerBits += bLoadNonIntraQuantizerMatrix;
	writeBits(parameter->load_non_intra_quantizer_matrix, bLoadNonIntraQuantizerMatrix, 0);

	if(parameter->load_non_intra_quantizer_matrix) {
		/*      loadNonIntraQuantizerMatrix = TRUE      */
	  
		rateControl->headerBits += 64 * bNonIntraQuantizer;
		for (index = 0; index < 64; index++)
			writeBits(tQuant[index], bNonIntraQuantizer, 0);
	}
}


/* write GOP header and update its cost to ratecontrol */
short   Write_GOP_Header(
		tParameter      *parameter,
		tRateControl    *rateControl)
{
	int             pic, sec, min, hrs;
        int  fps;
	short           numBits;

	numBits = rateControl->Si;

	rateControl->Si += bGOPHeaderCode;
	writeBits(cGOPHeaderCode, bGOPHeaderCode, 1);
	
	/* we need special handling for 29.97 and 23.XXX cases */
	
fprintf(stderr,"Now calculating GOP time code\n");
	/* this line may need to be changed to the look-up value */
	fps = (int) parameter->float_picture_rate;

	pic = parameter->pictureNum % fps;
	sec = (parameter->pictureNum / fps) % 60;
	min = (parameter->pictureNum / (60 * fps)) % 60;
	hrs = (parameter->pictureNum / (3600 * fps)) % 60;
	
	parameter->time_code = (long) (parameter->drop_frame_flag << 24) | 
	      (hrs << 19) | (min << 13) | (1 << 12) | (sec << 6) | (pic);
	
	/* probably better to have several writebit calls... that way    
	we don't have to worry about Big vs. Little Endian */

	rateControl->Si += bTimeCode;
	writeBits(parameter->time_code, bTimeCode, 0);

printf(stderr,"GOP time code = %ld\n", parameter->time_code);

	rateControl->Si += bClosedGOP;
	writeBits(parameter->closed_GOP, bClosedGOP, 0);
	rateControl->Si += bBrokenLink;
	
	/* need to change broken link flag */
	writeBits(parameter->broken_link, bBrokenLink, 0);

	numBits = rateControl->Si - numBits;

	return (numBits);
}

/* write picture header and update its cost to rate control */
short   Write_Picture_Header(
	short           picture_coding_type,
	tRateControl    *rateControl,
	short           forward_f_code,
	short           backward_f_code,
	int             temporal_reference)   /* display picture number */
{
	short   numBits;

	switch(picture_coding_type)
	{
		case kBPicture:

			numBits = rateControl->Sb;

			rateControl->Sb += bPictureStartCode;
			writeBits(cPictureStartCode, bPictureStartCode, 1);
			rateControl->Sb += bTemporalReference;
			
			writeBits(temporal_reference%1024, bTemporalReference, 0);
			
			rateControl->Sb += bPictureCodingType;
			writeBits(kBPicture, bPictureCodingType, 0);
			rateControl->Sb += bVbvDelay;
			if(rateControl->vbv_delay>=0){
			    writeBits(rateControl->vbv_delay, bVbvDelay, 0);
				}
			else{
			    /* VBV underflow */
			    writeBits(0, bVbvDelay, 0);
				}
			rateControl->Sb += bFullPelForwardVector;
			writeBits(cFullPelForwardVector, bFullPelForwardVector, 0);
			rateControl->Sb += bForwardFCode;
			writeBits(forward_f_code, bForwardFCode, 0);
			rateControl->Sb += bFullPelBackwardVector;
			writeBits(cFullPelBackwardVector, bFullPelBackwardVector, 0);
			rateControl->Sb += bBackwardFCode;
			writeBits(backward_f_code, bBackwardFCode, 0);
			rateControl->Sb += bExtraBitPicture;
			writeBits(cExtraBitPicture, bExtraBitPicture, 0);

			numBits = rateControl->Sb - numBits;

			break;

		case kPPicture:

			numBits = rateControl->Sp;

			rateControl->Sp += bPictureStartCode;
			writeBits(cPictureStartCode, bPictureStartCode, 1);
			rateControl->Sp += bTemporalReference;
			writeBits(temporal_reference%1024, bTemporalReference, 0);
			rateControl->Sp += bPictureCodingType;
			writeBits(kPPicture, bPictureCodingType, 0);
			rateControl->Sp += bVbvDelay;
			if(rateControl->vbv_delay>=0){
			    writeBits(rateControl->vbv_delay, bVbvDelay, 0);
			}
			else{
			    /* VBV underflow */
			    writeBits(0, bVbvDelay, 0);
			}
			rateControl->Sp += bFullPelForwardVector;
			writeBits(cFullPelForwardVector, bFullPelForwardVector, 0);
			rateControl->Sp += bForwardFCode;
			writeBits(forward_f_code, bForwardFCode, 0);
			rateControl->Sp += bExtraBitPicture;
			writeBits(cExtraBitPicture, bExtraBitPicture, 0);

			numBits = rateControl->Sp - numBits;

			break;

		case kIPicture:

			numBits = rateControl->Si;

			rateControl->Si += bPictureStartCode;
			writeBits(cPictureStartCode, bPictureStartCode, 1);
			rateControl->Si += bTemporalReference;
			writeBits(temporal_reference%1024, bTemporalReference, 0);
			rateControl->Si += bPictureCodingType;
			writeBits(kIPicture, bPictureCodingType, 0);
			rateControl->Si += bVbvDelay;
			if(rateControl->vbv_delay>=0){
			    writeBits(rateControl->vbv_delay, bVbvDelay, 0);
			}
			else{
			    /* VBV underflow */
			    writeBits(0, bVbvDelay, 0);
			}
			rateControl->Si += bExtraBitPicture;
			writeBits(cExtraBitPicture, bExtraBitPicture, 0);

			numBits = rateControl->Si - numBits;

			break;
	}

	return (numBits);
}


short   Write_Slice_Header(
	short           picture_coding_type,
	tRateControl    *rateControl)
{
	short   numBits;

	rateControl->macroblock_address_increment = 1;

	switch (picture_coding_type)
	{
		case kBPicture:

			numBits = rateControl->Sb;

			rateControl->Sb += bSliceStartCode;
			rateControl->Sb += bQuantizerScale;
			rateControl->Sb += bExtraBitSlice;

			numBits = rateControl->Sb - numBits;

			break;

		case kPPicture:

			numBits = rateControl->Sp;

			rateControl->Sp += bSliceStartCode;
			rateControl->Sp += bQuantizerScale;
			rateControl->Sp += bExtraBitSlice;

			numBits = rateControl->Sp - numBits;

			break;

		case kIPicture:

			numBits = rateControl->Si;

			rateControl->Si += bSliceStartCode;
			rateControl->Si += bQuantizerScale;
			rateControl->Si += bExtraBitSlice;

			numBits = rateControl->Si - numBits;

			break;
	}

	return (numBits);
}


int     Write_MB(
	short           picture_coding_type,
	int             macroblock_type,
	short           coded_block_pattern,
	tRateControl    *rateControl,
	tParameter      *parameter,
	int             MB_row,         /* current position with picture */
	int             MB_col,         /* current position within picture */
	short           forwardMV[],
	short           backwardMV[],
	int             y_coef[],
	int             cb_coef[],
	int             cr_coef[],
	int             quantizer_scale,
	short           fRange,
	short           forward_f_code,
	short           bRange,
	short           backward_f_code,
	int             *new_macroblock_type)
{
	static int      previous_macroblock_type;
	static int      previous_quantizer_scale;
	static short    PMVFor[2];
	static short    PMVBack[2];
	int             diff_vector[4], diff_vectorB[4];
	int             MB_bit_count = 0;
	static int      dct_dc_y_past, dct_dc_cb_past, dct_dc_cr_past;
	short           modified_quantizer_scale = 0;
	short           same;

	if (! MB_row && MB_col == 0)      /*      new frame       */
		previous_macroblock_type = 0;

	if (! MB_col)
	{
		/*      new slice       */

		dct_dc_y_past = 128;
		dct_dc_cb_past = 128;
		dct_dc_cr_past = 128;
	}

	/*      determine number of bits required to code current MB    */

	switch (picture_coding_type)
	{
		case kIPicture:

			MB_bit_count += 1;    /*      MBA - 1 bit                             */
			writeBits(1, 1, 0);

			/*      code MB type - I frame  */

			if (macroblock_type & kModifiedMQ)
			{
				MB_bit_count += MB_typeI_length[1];          
				writeBits(MB_typeI_value[1], MB_typeI_length[1], 0);
				MB_bit_count += 5;           
				writeBits(quantizer_scale, 5, 0);
				*new_macroblock_type = MB_typeI_switch[1];
			}
			else 
			{
				MB_bit_count += MB_typeI_length[0];   
				writeBits(MB_typeI_value[0], MB_typeI_length[0], 0);
				*new_macroblock_type = MB_typeI_switch[0];
			}

			/*      code blocks     */

			Write_Intra_Coefficients(y_coef, cb_coef, cr_coef,
					&dct_dc_y_past, &dct_dc_cb_past, &dct_dc_cr_past,
					&MB_bit_count);

			rateControl->dji += MB_bit_count - rateControl->Tij;
			rateControl->Si += MB_bit_count;
			return(MB_bit_count);

		case kPPicture:

			if ((coded_block_pattern) || (! MB_col) || 
				(MB_col == parameter->numMBsPerSlice - 1) 
					|| (macroblock_type & kMCType))
			{
				if ((! MB_col) || (macroblock_type & kIntraType))
				{
					PMVFor[0] = 0;
					PMVFor[1] = 0;
				}

				while (rateControl->macroblock_address_increment > 33)
				{
					MB_bit_count += MBA_length[0];
					writeBits(MBA_value[0], MBA_length[0], 0);
					rateControl->macroblock_address_increment -= 33;
				}
				
				MB_bit_count += MBA_length[rateControl->macroblock_address_increment];
				writeBits(MBA_value[rateControl->macroblock_address_increment],
						MBA_length[rateControl->macroblock_address_increment], 0);

				rateControl->macroblock_address_increment = 1;

				/*      code mb type - P frame  */

				if (macroblock_type & kIntraType)
				{
					if (macroblock_type & kModifiedMQ)
					{
						/*      mb type - intra w/ mod MQ       */

						MB_bit_count += MB_typeP_length[6];
						writeBits(MB_typeP_value[6], MB_typeP_length[6], 0);
						modified_quantizer_scale = 1;
						*new_macroblock_type = MB_typeP_switch[6];
					}
					else
					{
						MB_bit_count += MB_typeP_length[3];   /*      mb type - intra                 */
						writeBits(MB_typeP_value[3], MB_typeP_length[3], 0);
						*new_macroblock_type = MB_typeP_switch[3];
					}
				}
				else
				{
					if (macroblock_type & kMCType)
					{
						if (coded_block_pattern)
						{
							if (macroblock_type & kModifiedMQ)
							{
								/*      mb type - MC,coded,mod MQ       */

								MB_bit_count += MB_typeP_length[4];
								writeBits(MB_typeP_value[4], MB_typeP_length[4], 0);
								modified_quantizer_scale = 1;
								*new_macroblock_type = MB_typeP_switch[4];
							}
							else    /*      no mod MQ       */
							{
								/*      mb type - MC,coded      */

								MB_bit_count += MB_typeP_length[0];
								writeBits(MB_typeP_value[0], MB_typeP_length[0], 0);
								*new_macroblock_type = MB_typeP_switch[0];
							}
						}
						else    /*      not coded       */
						{
							/*      mb type - MC                    */

							MB_bit_count += MB_typeP_length[2];
							writeBits(MB_typeP_value[2], MB_typeP_length[2], 0);
							*new_macroblock_type = MB_typeP_switch[2];
							if (MB_col)
								quantizer_scale = previous_quantizer_scale;
						}
					}
					else    /*      no MC   */
					{
						if (coded_block_pattern)
						{
							if (macroblock_type & kModifiedMQ)
							{
								/*      mb type - coded,mod MQ  */

								MB_bit_count += MB_typeP_length[5];
								writeBits(MB_typeP_value[5], MB_typeP_length[5], 0);
								modified_quantizer_scale = 1;
								*new_macroblock_type = MB_typeP_switch[5];
							}
							else
							{
								/*      mb type - coded                 */

								MB_bit_count += MB_typeP_length[1];
								writeBits(MB_typeP_value[1], MB_typeP_length[1], 0);
								*new_macroblock_type = MB_typeP_switch[1];
							}

							PMVFor[0] = 0;
							PMVFor[1] = 0;
						}
						else
						{
							/*      mb type - no MC, not coded */
							/*
								first or last MB in slice, not skipped
								-> MC with MV=0, not coded
							*/

/*  3/21/94: the following two lines were added by Mr. Hirofumi Nisikawa 
	     (Mitsubishi Electric Corp) and Kinya Oosa (Nippon Steel).  */

							
							forwardMV[0] = 0;
							forwardMV[1] = 0;
							
							macroblock_type |= kMCType;
							MB_bit_count += MB_typeP_length[2];
							writeBits(MB_typeP_value[2], MB_typeP_length[2], 0);
							*new_macroblock_type = MB_typeP_switch[2];
							if (MB_col)
								quantizer_scale = previous_quantizer_scale;
						}
					}
				}
			}
			else    /*      skipped */
			{
				PMVFor[0] = 0;
				PMVFor[1] = 0;

				rateControl->macroblock_address_increment++;
				rateControl->djp += MB_bit_count - rateControl->Tpj;
				rateControl->Sp += MB_bit_count;
				return(MB_bit_count);
			}

			previous_quantizer_scale = quantizer_scale;

			if ((macroblock_type & kIntraType) &&
					(! (previous_macroblock_type & kIntraType)))
			{
				dct_dc_y_past = 128;
				dct_dc_cb_past = 128;
				dct_dc_cr_past = 128;
			}

			break;

		case kBPicture:

			/*      check if mvs are the same       */

			same = FALSE;

			if ((macroblock_type & kMCType) && (previous_macroblock_type & kMCType))
			{
				/*      check mvs are the same  */

				if ((macroblock_type & kForwardMV) && (macroblock_type & kBackwardMV))
				{
					if ((PMVFor[0] == forwardMV[0]) &&
							(PMVFor[1] == forwardMV[1]) &&
							(PMVBack[0] == backwardMV[0]) &&
							(PMVBack[1] == backwardMV[1]))
						same = TRUE;
				}
				else if (macroblock_type & kForwardMV)
				{
					if ((PMVFor[0] == forwardMV[0]) &&
							(PMVFor[1] == forwardMV[1]))
						same = TRUE;
				}
				else    /*      macroblock_type & kBackwardMV    */
				{
					if ((PMVBack[0] == backwardMV[0]) &&
							(PMVBack[1] == backwardMV[1]))
						same = TRUE;
				}
			}

			if (same)       /*      check if macroblock_type the same        */
				same = CheckMbType(previous_macroblock_type, macroblock_type, coded_block_pattern);

			if ((! same) || (! MB_col) || (MB_col == (parameter->numMBsPerSlice - 1)) ||
					coded_block_pattern)
			{
				if ((! MB_col) || (macroblock_type & kIntraType))
				{
					PMVFor[0] = 0;
					PMVFor[1] = 0;
					PMVBack[0] = 0;
					PMVBack[1] = 0;
				}

				while (rateControl->macroblock_address_increment > 33)
				{
					MB_bit_count += MBA_length[0];
					writeBits(MBA_value[0], MBA_length[0], 0);
					rateControl->macroblock_address_increment -= 33;
				}
				
				MB_bit_count += MBA_length[rateControl->macroblock_address_increment];
				writeBits(MBA_value[rateControl->macroblock_address_increment],
						MBA_length[rateControl->macroblock_address_increment], 0);

				rateControl->macroblock_address_increment = 1;

				/*      code mb type - B frame  */

				if (macroblock_type & kIntraType)
				{
					if (macroblock_type & kModifiedMQ)
					{
						/*      mb type - intra w/ mod MQ       */

						MB_bit_count += MB_typeB_length[10];  
						writeBits(MB_typeB_value[10], MB_typeB_length[10], 0);
						modified_quantizer_scale = 1;
						*new_macroblock_type = MB_typeB_switch[10];
					}
					else    /*      no mod MQ       */
					{
						/*      mb type - intra                         */

						MB_bit_count += MB_typeB_length[6];   
						writeBits(MB_typeB_value[6], MB_typeB_length[6], 0);
						*new_macroblock_type = MB_typeB_switch[6];
					}
				}
				else    /*      non intra       */
				{
					if (macroblock_type & kForwardMV)
					{
						if (macroblock_type & kBackwardMV)
						{
							/*      interpolated    */

							if (coded_block_pattern)
							{
								if (macroblock_type & kModifiedMQ)
								{
									/*      interp,coded,mod MQ     */

									MB_bit_count += MB_typeB_length[7];   
									writeBits(MB_typeB_value[7], MB_typeB_length[7], 0);
									modified_quantizer_scale = 1;
									*new_macroblock_type = MB_typeB_switch[7];
								}
								else    /*      no MQ   */
								{
									/*      interp,coded    */

									MB_bit_count += MB_typeB_length[1];   
									writeBits(MB_typeB_value[1], MB_typeB_length[1], 0);
									*new_macroblock_type = MB_typeB_switch[1];
								}
							}
							else    /*      not coded       */
							{
								/*      interp          */

								MB_bit_count += MB_typeB_length[0];   
								writeBits(MB_typeB_value[0], MB_typeB_length[0], 0);
								*new_macroblock_type = MB_typeB_switch[0];
							}
						}
						else    /*      forward only    */
						{
							if (coded_block_pattern)
							{
								if (macroblock_type & kModifiedMQ)
								{
									/*      for,coded,mod MQ        */

									MB_bit_count += MB_typeB_length[8];   
									writeBits(MB_typeB_value[8], MB_typeB_length[8], 0);
									modified_quantizer_scale = 1;
									*new_macroblock_type = MB_typeB_switch[8];
								}
								else    /*      no MQ   */
								{
									/*      for,coded       */

									MB_bit_count += MB_typeB_length[5];   
									writeBits(MB_typeB_value[5], MB_typeB_length[5], 0);
									*new_macroblock_type = MB_typeB_switch[5];
								}
							}
							else    /*      not coded       */
							{
								/*      for             */

								MB_bit_count += MB_typeB_length[4];   
								writeBits(MB_typeB_value[4], MB_typeB_length[4], 0);
								*new_macroblock_type = MB_typeB_switch[4];
							}
						}
					}
					else    /*      (macroblock_type & kBackwardMV)  */
					{
						if (coded_block_pattern)
						{
							if (macroblock_type & kModifiedMQ)
							{
								/*      back,coded,mod MQ       */

								MB_bit_count += MB_typeB_length[9];   
								writeBits(MB_typeB_value[9], MB_typeB_length[9], 0);
								modified_quantizer_scale = 1;
								*new_macroblock_type = MB_typeB_switch[9];
							}
							else    /*      no MQ   */
							{
								/*      back,coded      */

								MB_bit_count += MB_typeB_length[3];   
								writeBits(MB_typeB_value[3], MB_typeB_length[3], 0);
								*new_macroblock_type = MB_typeB_switch[3];
							}
						}
						else    /*      not coded       */
						{
							/*      back            */

							MB_bit_count += MB_typeB_length[2];   
							writeBits(MB_typeB_value[2], MB_typeB_length[2], 0);
							*new_macroblock_type = MB_typeB_switch[2];
						}
					}
				}
			}
			else    /*      skipped */
			{
				/*      current MVs == prev MVs */

				rateControl->macroblock_address_increment++;
				rateControl->djb += MB_bit_count - rateControl->Tbj;
				rateControl->Sb += MB_bit_count;
				return(MB_bit_count);
			}

			if ((macroblock_type & kIntraType) &&
					(! (previous_macroblock_type & kIntraType)))
			{
				dct_dc_y_past = 128;
				dct_dc_cb_past = 128;
				dct_dc_cr_past = 128;
			}

			break;
	}

	previous_macroblock_type = *new_macroblock_type;

	if (modified_quantizer_scale)
	{
		MB_bit_count += 5;            /*      new MQ  */
		writeBits(quantizer_scale, 5, 0);
	}

	if ((! (macroblock_type & kIntraType)) && (macroblock_type & kMCType))
	{
		switch (picture_coding_type)
		{
			case kBPicture:

				/*      find diff_vector        */

				if (macroblock_type & kForwardMV)
				{
					diff_vector[0] = forwardMV[0] - PMVFor[0];
					diff_vector[1] = forwardMV[1] - PMVFor[1];

					PMVFor[0] = forwardMV[0];
					PMVFor[1] = forwardMV[1];

					/*      code foward mv  */

					Write_MV(diff_vector, 2, forward_f_code,
							fRange, &MB_bit_count, parameter, picture_coding_type);
				}

				if (macroblock_type & kBackwardMV)
				{
					diff_vectorB[0] = backwardMV[0] - PMVBack[0];
					diff_vectorB[1] = backwardMV[1] - PMVBack[1];

					PMVBack[0] = backwardMV[0];
					PMVBack[1] = backwardMV[1];

					/*      code backward mv        */

					Write_MV(diff_vectorB, 2, backward_f_code,
							bRange, &MB_bit_count, parameter, picture_coding_type);
				}

				break;

			case kPPicture:

				/*      find diff_vector        */

				diff_vector[0] = forwardMV[0] - PMVFor[0];
				diff_vector[1] = forwardMV[1] - PMVFor[1];

				PMVFor[0] = forwardMV[0];
				PMVFor[1] = forwardMV[1];

				/*      code foward mv  */

				Write_MV(diff_vector, 2, forward_f_code,
						fRange, &MB_bit_count, parameter, picture_coding_type);

				break;
		}
	}

	if ((! (macroblock_type & kIntraType)) && (macroblock_type & kCodedType))
	{
		/*      code CBP        */

		MB_bit_count += CBP_length[coded_block_pattern];
		writeBits(CBP_value[coded_block_pattern], CBP_length[coded_block_pattern], 0);
	}

	/*      code coeffs     */

	if (macroblock_type & kIntraType)
		Write_Intra_Coefficients(y_coef, cb_coef, cr_coef,
				&dct_dc_y_past, &dct_dc_cb_past, &dct_dc_cr_past, &MB_bit_count);
	else if (coded_block_pattern)
		Write_Non_Intra_Coefficients(y_coef, cb_coef, cr_coef,
				&MB_bit_count, coded_block_pattern);

	switch (picture_coding_type)
	{
		case kPPicture:

			rateControl->djp += MB_bit_count - rateControl->Tpj;
			rateControl->Sp += MB_bit_count;

			break;

		case kBPicture:

			rateControl->djb += MB_bit_count - rateControl->Tbj;
			rateControl->Sb += MB_bit_count;

			break;
	}

	return(MB_bit_count);
}

/* generate and write macroblock motion vector data to bitstream */

void    Write_MV(
	int             *diff_vector,
	short           numDiffVectors,
	short           f_code,
	short           range,
	int             *MB_bit_count,
	tParameter      *parameter,
	short           picture_coding_type)
{
	int             scale_factor, residual, vlc_code_magnitude;
	short   index;

	scale_factor = 1 << (f_code - 1);

	for (index = 0; index < numDiffVectors; index++)
	{
		if (diff_vector[index] < -range)
			diff_vector[index] += 2 * range;
		else if (diff_vector[index] > (range - 1))
			diff_vector[index] -= 2 * range;

		if (diff_vector[index] == 0)
		{
			residual = 0;
			vlc_code_magnitude = 0;
		}
		else
		{
			residual = (abs(diff_vector[index]) - 1) % scale_factor;
			vlc_code_magnitude =
					(abs(diff_vector[index]) - residual) / scale_factor;

			if (scale_factor != 1)
				vlc_code_magnitude++;
		}

		/*      encode vlc_code_magnitude and sign of diff_vector[index]        */

		*MB_bit_count += MV_length[vlc_code_magnitude];

		if (diff_vector[index] < 0)
			writeBits(MV_value[vlc_code_magnitude] + 1,
					MV_length[vlc_code_magnitude], 0);
		else
			writeBits(MV_value[vlc_code_magnitude],
					MV_length[vlc_code_magnitude], 0);

		if (f_code != 1 && vlc_code_magnitude != 0)
		{
			*MB_bit_count += f_code - 1;  /*      encode residual */
			writeBits(residual, f_code - 1, 0);
		}
	}
}


void    Write_Intra_Coefficients(
	int             *y_coef,
	int             *cb_coef,
	int             *cr_coef,
	int             *dct_dc_y_past,
	int             *dct_dc_cb_past,
	int             *dct_dc_cr_past,
	int             *MB_bit_count)
{
	unsigned int    tempInt;
	short                   size, line, pix, row, col, offset, currentIndex, run;
	int                             diff_DC;

	offset = kMBWidth - kDCTSize;

	for (line = 0; line < kMBHeight; line += kDCTSize)
	{
		for (pix = 0; pix < kMBWidth; pix += kDCTSize)
		{
			currentIndex = line * kMBWidth + pix;

			/*      code DC coefficient     */

			diff_DC = abs(y_coef[currentIndex] - *dct_dc_y_past);

			/*      determine size  */

			size = 0;

			if (diff_DC != 0)
			{
				while (diff_DC >= (1 << size))
					size++;

				*MB_bit_count += DC_Y_length[size];            /*      VLC code                */
				writeBits(DC_Y_value[size], DC_Y_length[size], 0);

				diff_DC = y_coef[currentIndex] - *dct_dc_y_past;

				if (diff_DC < 0)
				{
					tempInt = -diff_DC;
					tempInt = ~tempInt;
					tempInt = (tempInt << (32 - size)) >> (32 - size);
				}
				else
					tempInt = diff_DC;

				*MB_bit_count += size;                                        /*      additional bits */
				writeBits(tempInt, size, 0);

				*dct_dc_y_past = y_coef[currentIndex];
			}
			else
			{
				*MB_bit_count += DC_Y_length[size];            /*      VLC code                */
				writeBits(DC_Y_value[size], DC_Y_length[size], 0);
			}

			/*      code AC coeffs  */

			col = 1;
			currentIndex++;
			run = 0;

			for (row = 0; row < kDCTSize; row++,
					currentIndex += offset)
			{
				for ( ; col < kDCTSize; col++, currentIndex++)
				{
					if (y_coef[currentIndex])
					{
						Write_AC_VLC(y_coef[currentIndex],
								&run, MB_bit_count, 0);
					}
					else    /*      zero    */
					{
						run++;
					}
				}

				col = 0;
			}

			*MB_bit_count += AC_length[0][0];  /*  EOB code  */
			writeBits(AC_value[0][0], AC_length[0][0], 0);
		}
	}

	/*      chrom - U       */

	/*      code DC coefficient     */

	diff_DC = abs(cb_coef[0] - *dct_dc_cb_past);

	/*      determine size  */

	size = 0;

	while (diff_DC >= (1 << size))
		size++;

	*MB_bit_count += DC_C_length[size];   /*  VLC code  */
	writeBits(DC_C_value[size], DC_C_length[size], 0);

	diff_DC = cb_coef[0] - *dct_dc_cb_past;

	if (diff_DC < 0)
	{
		tempInt = -diff_DC;
		tempInt = ~tempInt;
		tempInt = (tempInt << (32 - size)) >> (32 - size);
	}
	else
		tempInt = diff_DC;

	*MB_bit_count += size;                                        /*      additional bits */
	writeBits(tempInt, size, 0);

	*dct_dc_cb_past = cb_coef[0];

	/*      code AC coeffs  */

	col = 1;
	currentIndex = 1;
	run = 0;

	for (row = 0; row < kDCTSize; row++)
	{
		for ( ; col < kDCTSize; col++, currentIndex++)
		{
			if (cb_coef[currentIndex])
			{
				Write_AC_VLC(cb_coef[currentIndex],
						&run, MB_bit_count, 0);
			}
			else    /*      zero    */
			{
				run++;
			}
		}

		col = 0;
	}

	*MB_bit_count += AC_length[0][0];         /*      EOB code        */
	writeBits(AC_value[0][0], AC_length[0][0], 0);

	/*      chrom - V       */

	/*      code DC coefficient     */

	diff_DC = abs(cr_coef[0] - *dct_dc_cr_past);

	/*      determine size  */

	size = 0;

	while (diff_DC >= (1 << size))
		size++;

	*MB_bit_count += DC_C_length[size];  /*  VLC code  */
	writeBits(DC_C_value[size], DC_C_length[size], 0);

	diff_DC = cr_coef[0] - *dct_dc_cr_past;

	if (diff_DC < 0)
	{
		tempInt = -diff_DC;
		tempInt = ~tempInt;
		tempInt = (tempInt << (32 - size)) >> (32 - size);
	}
	else
		tempInt = diff_DC;

	*MB_bit_count += size;          /* additional bits */
	writeBits(tempInt, size, 0);

	*dct_dc_cr_past = cr_coef[0];

	/*      code AC coeffs  */

	col = 1;
	currentIndex = 1;
	run = 0;

	for (row = 0; row < kDCTSize; row++)
	{
		for ( ; col < kDCTSize; col++, currentIndex++)
		{
			if (cr_coef[currentIndex])
			{
				Write_AC_VLC(cr_coef[currentIndex],
						&run, MB_bit_count, 0);
			}
			else    /*      zero    */
			{
				run++;
			}
		}

		col = 0;
	}

	*MB_bit_count += AC_length[0][0];         /*      EOB code        */
	writeBits(AC_value[0][0], AC_length[0][0], 0);
}


void    Write_Non_Intra_Coefficients(
	int     *y_coef,
	int     *cb_coef,
	int     *cr_coef,
	int     *MB_bit_count,
	short   coded_block_pattern)
{
	short   line, pix, row, col, offset, currentIndex, run, code, first;

	offset = kMBWidth - kDCTSize;

	code = 0x20;

	for (line = 0; line < kMBHeight; line += kDCTSize)
	{
		for (pix = 0; pix < kMBWidth; pix += kDCTSize)
		{
			if (! (coded_block_pattern & code))
			{
				code >>= 1;
				continue;
			}

			code >>= 1;

			currentIndex = line * kMBWidth + pix;

			run = 0;
			first = 1;

			for (row = 0; row < kDCTSize; row++,
					currentIndex += offset)
			{
				for (col = 0; col < kDCTSize; col++, currentIndex++)
				{
					if (y_coef[currentIndex])
					{
						Write_AC_VLC(y_coef[currentIndex],
								&run, MB_bit_count, first);

						first = 0;
					}
					else    /*      zero    */
					{
						run++;
					}
				}
			}

			*MB_bit_count += AC_length[0][0];         /*      EOB code        */
			writeBits(AC_value[0][0], AC_length[0][0], 0);
		}
	}

	/*      chrom - U       */

	if (coded_block_pattern & 0x02)
	{
		run = 0;
		currentIndex = 0;
		first = 1;

		for (row = 0; row < kDCTSize; row++)
		{
			for (col = 0; col < kDCTSize; col++, currentIndex++)
			{
				if (cb_coef[currentIndex])
				{
					Write_AC_VLC(cb_coef[currentIndex],
							&run, MB_bit_count, first);
					first = 0;
				}
				else    /*      zero    */
				{
					run++;
				}
			}
		}

		*MB_bit_count += AC_length[0][0];         /*      EOB code        */
		writeBits(AC_value[0][0], AC_length[0][0], 0);
	}

	/*      chrominance - Cr       */

	if (coded_block_pattern & 0x01)
	{
		run = 0;
		currentIndex = 0;
		first = 1;

		for (row = 0; row < kDCTSize; row++)
		{
			for (col = 0; col < kDCTSize; col++, currentIndex++)
			{
				if (cr_coef[currentIndex])
				{
					Write_AC_VLC(cr_coef[currentIndex],
							&run, MB_bit_count, first);
					first = 0;
				}
				else    /*      zero    */
				{
					run++;
				}
			}
		}

		*MB_bit_count += AC_length[0][0];   /*      EOB code        */
		writeBits(AC_value[0][0], AC_length[0][0], 0);
	}
}


int     Write_AC_VLC(
	int     diff,
	short   *run,
	int     *MB_bit_count,
	short   first)
{
	unsigned int    mag, tempInt;
	short           signBit;
	int             numBits;

	numBits = *MB_bit_count;
	mag = abs(diff);
	signBit = (diff < 0) ? 1 : 0;

	if (first && (mag == 1) && (*run == 0))
	{
		writeBits(kACVlcFirstP + signBit, kACVlcFirst, 0);
		*MB_bit_count += kACVlcFirst;
	}
	else
	{
		*MB_bit_count += AC_length[*run][mag];

		if (AC_length[*run][mag] == 20)
		{
			tempInt = kACVlc_Escape20;
			tempInt |= (*run << 8);
			tempInt |= AC_value[EscapeOffset + signBit][mag];
			writeBits(tempInt, 20, 0);
		}
		else if (AC_length[*run][mag] == 28)
		{
			tempInt = kACVlc_Escape28;
			tempInt |= (*run << 16);

			if (signBit)
				tempInt |= 0x8000;

			tempInt |= AC_value[EscapeOffset + signBit][mag];
			writeBits(tempInt, 28, 0);
		}
		else
		{
			writeBits(AC_value[*run][mag] + signBit, 
				AC_length[*run][mag], 0);
		}
	}

	*run = 0;

	return(*MB_bit_count - numBits);
}


short   CheckMbType(
	int     previous_macroblock_type,
	int     macroblock_type,
	short   coded_block_pattern)
{
	int     new_macroblock_type;

	/*      both previous_macroblock_type and macroblock_type are not intra        */

	if (macroblock_type & kForwardMV)
	{
		if (macroblock_type & kBackwardMV)
		{
			/*      interpolated    */

			if (coded_block_pattern)
			{
				if (macroblock_type & kModifiedMQ)
				{
					/*      interp,coded,mod MQ     */

					new_macroblock_type = MB_typeB_switch[7];
				}
				else    /*      no MQ   */
				{
					/*      interp,coded    */

					new_macroblock_type = MB_typeB_switch[1];
				}
			}
			else    /*      not coded       */
			{
				/*      interp          */

				new_macroblock_type = MB_typeB_switch[0];
			}
		}
		else    /*      forward only    */
		{
			if (coded_block_pattern)
			{
				if (macroblock_type & kModifiedMQ)
				{
					/*      for,coded,mod MQ        */

					new_macroblock_type = MB_typeB_switch[8];
				}
				else    /*      no MQ   */
				{
					/*      for,coded       */

					new_macroblock_type = MB_typeB_switch[5];
				}
			}
			else    /*      not coded       */
			{
				/*      for             */

				new_macroblock_type = MB_typeB_switch[4];
			}
		}
	}
	else if (macroblock_type & kBackwardMV)
	{
		if (coded_block_pattern)
		{
			if (macroblock_type & kModifiedMQ)
			{
				/*      back,coded,mod MQ       */

				new_macroblock_type = MB_typeB_switch[9];
			}
			else    /*      no MQ   */
			{
				/*      back,coded      */

				new_macroblock_type = MB_typeB_switch[3];
			}
		}
		else    /*      not coded       */
		{
			/*      back            */

			new_macroblock_type = MB_typeB_switch[2];
		}
	}
	else    /*      no motion compensation   */
	{
		/*      could be skipped        */

		new_macroblock_type = MB_typeB_switch[11];
	}

	if (new_macroblock_type == previous_macroblock_type)
		return (1);

	return (0);
}
