/* File: initial.c */
/* This file reads in the user parameter script file (encoder.ini) */
  
/*
 *      $Log:   initial.c,v $
 * Revision 1.4  93/09/03  01:15:00  hana
 * VBV buffer operation 
 * 
 * Revision 1.1.1.1  93/03/29  11:27:55  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:33:17  au
 * Initial revision
 *
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "consts.h"
#include "mpeg1.h"
#include "encoder.h"
#include "encoder.pro"

void    Initial(
	tParameter      *parameter,
	tData           *Data,
	tRateControl    *rateControl,
	char            *parameterFName)

{
	extern FILE     *gBitStream;

	FILE    *fptr;
	int     i, j;
	char    lineBuf[kMaxCharBufLen];
	double  vbvBufferStart;


	/* get user input data */

	if ((fptr = fopen (parameterFName, "r")) == 0)
	{
		fprintf (stderr, "input parameter file '%s' not found\n", parameterFName);
		exit(-1);
	}

	/*  file data (string) */
	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%s", parameter->inputSequence);

	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%s", parameter->outputSequence);

	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%d", &parameter->sourceFileFormat);

	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%d", &parameter->reconFileFormat);

	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%s", parameter->statusFName);

	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%s", parameter->bitStreamFName);
	
	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%d", &parameter->firstPicture);

	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%d", &parameter->lastPicture);

	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%d", &parameter->horizontal_size);
	
fprintf(stderr,"hor = %d  ", parameter->horizontal_size);

	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%d", &parameter->vertical_size);

fprintf(stderr,"ver = %d\n", parameter->vertical_size);
	
	/* read in pel aspect ratio */
	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%d", &parameter->pel_aspect_ratio);

	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%d", &parameter->picture_rate);
	if(parameter->picture_rate < 1 || parameter->picture_rate>8)
	{
	  fprintf(stderr, "Invalid picture rate code (%d)\n", parameter->picture_rate);
	  exit(-1);
	}
	parameter->float_picture_rate=pictureRateTable[parameter->picture_rate];
	
	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%lf", &parameter->bit_rate);

	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%lf", &parameter->maxPictureDelay);

	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%d", &parameter->maxVbvBufferSize);

	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%d", &parameter->constrained_parameters_flag);

	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%d", &parameter->N);
	
	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%d", &parameter->M);
	
	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%d", &parameter->searchDistance);
	
	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%d", &parameter->halfPel);
	
	fgets (lineBuf, kMaxCharBufLen, fptr);
	sscanf (lineBuf, "%d", &parameter->fileMV);  /* 0 = FALSE */

	fclose (fptr);
	

	/* change bitRate in kbit/s to that in bit/s */
	parameter->bit_rate *= 1000.;
	
	/* convert picture rate from one of the 8 legal values indicated
	in the MPEG-1 specification (clause XXXX) to a floating point
	value */
	/* The amount of initial insertion to VBV buffer */
	/*  = half of bit per frame                      */
	
	vbvBufferStart = 0.5 * parameter->bit_rate / parameter->float_picture_rate;
	
	/* calculate vbv_buffer_size */ /* 16384 = 16*1024 */
	parameter->vbv_buffer_size =    
		parameter->maxPictureDelay * parameter->bit_rate 
		/ parameter->float_picture_rate
		+ vbvBufferStart;

	parameter->vbv_buffer_size =
		(parameter->vbv_buffer_size / 16384 + 1) * 16384;


	if(parameter->vbv_buffer_size > parameter->maxVbvBufferSize)
			parameter->vbv_buffer_size = parameter->maxVbvBufferSize;

	parameter->forward_f_code = Iarray(parameter->M);
	parameter->backward_f_code = Iarray(parameter->M);
	parameter->fRange = Iarray(parameter->M);
	parameter->bRange = Iarray(parameter->M);

	/*      range are multiplied by 2 for half pel  */

	/*      P Picture       */

	parameter->fRange[0] = parameter->searchDistance *
			parameter->M * 2;
	parameter->forward_f_code[0] = 1;
	while (parameter->fRange[0] > (8 << parameter->forward_f_code[0]))
		parameter->forward_f_code[0]++;
	parameter->fRange[0] = 8 << parameter->forward_f_code[0];
	parameter->backward_f_code[0] = 0;
	parameter->bRange[0] = 0;

	/*      B Pictures      */

	for (i = parameter->M - 1, j = 1; i > 0; i--, j++)
	{
		parameter->fRange[j] = parameter->searchDistance * j * 2;
		parameter->forward_f_code[j] = 1;
		while (parameter->fRange[j] > (8 << parameter->forward_f_code[j]))
			parameter->forward_f_code[j]++;
		parameter->fRange[j] = 8 << parameter->forward_f_code[j];
		parameter->bRange[j] = parameter->searchDistance * i * 2;
		parameter->backward_f_code[j] = 1;
		while (parameter->bRange[j] > (8 << parameter->backward_f_code[j]))
			parameter->backward_f_code[j]++;
		parameter->bRange[j] = 8 << parameter->backward_f_code[j];
	}

	Init_AC_VLC();

	if ((gBitStream = fopen (parameter->bitStreamFName, "wb")) == NULL)
	{
		printf ("error opening bitstream file '%s' ... Terminating ...\n",
				parameter->bitStreamFName);
		exit(-1);
	}

	rateControl->headerBits = 0;
	rateControl->totalGOPBits = 0;

	rateControl->Xi = 160.0 * parameter->bit_rate / 115.0;
	rateControl->Xp = 60.0 * parameter->bit_rate / 115.0;
	rateControl->Xb = 42.0 * parameter->bit_rate / 115.0;
	rateControl->R = 0;

	/*      Note: check number of frames in first GOP       */

	rateControl->G = (parameter->bit_rate /
			parameter->float_picture_rate) * parameter->N;
	rateControl->r = 2.0 * parameter->bit_rate / parameter->float_picture_rate;
	rateControl->d0i = 10.0 * rateControl->r / 31;
	rateControl->d0p = rateControl->d0i;                  /*      d0i * Kp, Kp = 1.0      */
	rateControl->d0b = rateControl->d0i * 1.4;    /*      d0i * Kb, Kb = 1.4      */
	rateControl->Tipb = parameter->bit_rate /
		(8 * parameter->float_picture_rate);
	rateControl->avg_act = 400.0;

	/* initial vbv_buffer filling for B_(-1) virtually */
	rateControl->vbvOccupy = parameter->vbv_buffer_size - vbvBufferStart -
		parameter->bit_rate / parameter->float_picture_rate;
		      
	/*      print user input data into status file  */

	Init_DCT();
}


/* initialize bitstream and user parameters */

int InitParameter(tParameter *parameter)
{
	

	if (parameter->horizontal_size % 2)
		{
		printf("Error! horizontal picture size must be an even 
			value due to 2:1 chroma decimation factor!\n");
		exit (-1);
		}
	
	/* coded picture: round up to nearest macroblock multiple */
	parameter->mb_width = (parameter->horizontal_size + 15) / 16;
	parameter->mb_height = (parameter->vertical_size + 15) / 16;

fprintf(stderr,"mb_width = %d, mb_height = %d\n", parameter->mb_width, parameter->mb_height);
	
	printf("coded picture is %d samples wider than source picture\n",
		parameter->mb_width*16 - parameter->horizontal_size);  
	printf("coded picture is %d samples taller than source picture\n",        
		parameter->mb_height*16 - parameter->vertical_size);

/* we need a distinction between coded and source picture size */
	
	/* coded picture dimensions */

	parameter->luminWidth = parameter->mb_width * kMBWidth;

	parameter->luminHeight = parameter->mb_height * kMBHeight;
	parameter->chromWidth = parameter->luminWidth / 2;         
	/* assumes 4:2:0 */
	parameter->chromHeight = parameter->luminHeight / 2;

	parameter->numLuminPixels = 
		parameter->luminWidth * parameter->luminHeight;
	
	parameter->numChromPixels =
		parameter->chromWidth * parameter->chromHeight;

	parameter->numSlices = parameter->mb_height;
	parameter->numMBsPerSlice = parameter->mb_width;
	parameter->numMacroBlocks = parameter->mb_height * 
		parameter->mb_width;
	
	/* note:  these figures are based on source dimensions which 
	could be up to 15 pels smaller than coded dimensions */

	parameter->luminHeightMB = parameter->luminHeight - kMBHeight;
	parameter->luminWidthMB  = parameter->luminWidth - kMBWidth;
	parameter->luminWidthDCT = parameter->luminWidth - kDCTSize;
	parameter->chromWidthB = parameter->chromWidth - kBWidth;
	parameter->chromWidthDCT = parameter->chromWidth - kDCTSize;

	parameter->mBLuminRowOffset = ((kMBHeight - 1) * parameter->luminWidth);
	parameter->mBChromRowOffset = ((kBHeight - 1) * parameter->chromWidth);

        parameter->load_intra_quantizer_matrix = 0;
        parameter->load_non_intra_quantizer_matrix = 0;

return (0);
}


/* this function should be called before sequence header is written 
	to bitstream and after InitParameter() has been called */


int InitData(tData *Data, tParameter *parameter)   
{
	int numLuminPixels, numChromPixels;
	
/* this can be updated to use the new tParameter elements */
	numLuminPixels = (16*parameter->mb_width) * 
				(16*parameter->mb_height);
	numChromPixels = (8*parameter->mb_width) * 
				(16*parameter->mb_height);
    
	
	/* Original picture array memory allocation */

	Data->luminPicture0 = Barray(numLuminPixels);
	Data->luminPicture1 = Barray(numLuminPixels);
	Data->luminPicture2 = Barray(numLuminPixels);

	Data->chromUPicture0 = Barray(numChromPixels);
	Data->chromUPicture1 = Barray(numChromPixels);
	Data->chromUPicture2 = Barray(numChromPixels);

	Data->chromVPicture0 = Barray(numChromPixels);
	Data->chromVPicture1 = Barray(numChromPixels);
	Data->chromVPicture2 = Barray(numChromPixels);

	/* Reconstructed picture array memory allocation */

	Data->luminPicture0R = Barray(numLuminPixels);
	Data->luminPicture1R = Barray(numLuminPixels);
	Data->luminPicture2R = Barray(numLuminPixels);

	Data->chromUPicture0R = Barray(numChromPixels);
	Data->chromUPicture1R = Barray(numChromPixels);
	Data->chromUPicture2R = Barray(numChromPixels);

	Data->chromVPicture0R = Barray(numChromPixels);
	Data->chromVPicture1R = Barray(numChromPixels);
	Data->chromVPicture2R = Barray(numChromPixels);
	
	/*      assign working pointers */

	Data->firstPredictY = Data->luminPicture0;
	Data->firstPredictU = Data->chromUPicture0;
	Data->firstPredictV = Data->chromVPicture0;
	Data->secondPredictY = Data->luminPicture1;
	Data->secondPredictU = Data->chromUPicture1;
	Data->secondPredictV = Data->chromVPicture1;
	Data->predictY = Data->luminPicture2;
	Data->predictU = Data->chromUPicture2;
	Data->predictV = Data->chromVPicture2;

	Data->firstPredictYR = Data->luminPicture0R;
	Data->firstPredictUR = Data->chromUPicture0R;
	Data->firstPredictVR = Data->chromVPicture0R;
	Data->secondPredictYR = Data->luminPicture1R;
	Data->secondPredictUR = Data->chromUPicture1R;
	Data->secondPredictVR = Data->chromVPicture1R;
	Data->predictYR = Data->luminPicture2R;
	Data->predictUR = Data->chromUPicture2R;
	Data->predictVR = Data->chromVPicture2R;
	
return(0);
}





BYTE    *Barray(
	int     size)
{
	BYTE    *a;

	a = (BYTE *) malloc((unsigned) size * sizeof(BYTE));

	if (!a)
	{
		printf("\nBarray: Failure in allocating arrary memory\n");
		exit(-1);
	}

	return a;
}

short *Sarray(
	int     size)
{
	short   *a;

	a = (short *) malloc((unsigned) size * sizeof(short));

	if (!a)
	{
		printf("\nSarray: Failure in allocating arrary memory\n");
		exit(-1);
	}

	return a;
}

int *Iarray(
	int     size)
{
	int   *a;

	a = (int *) malloc((unsigned) size * sizeof(int));

	if (!a)
	{
		printf("\nIarray: Failure in allocating arrary memory\n");
		exit(-1);
	}

	return a;
}


int Print_Parameters(
 tData *data, 
 tParameter *parameter, 
 tRateControl *rateControl)
{

	FILE    *fptr;

	if((fptr =fopen(parameter->statusFName, "w")) == NULL)
	{
		printf("\nopen error filename: %s\n", parameter->statusFName);
		exit(-1);
	}

   /* this section needs to be incorperated into a tracelevel routine */
	fprintf(fptr, "***** %s *****\n", parameter->statusFName);
	fprintf(fptr, "Input sequence : %s\n", parameter->inputSequence);
	fprintf(fptr, "Output sequence : %s\n", parameter->outputSequence);
	fprintf(fptr, "firstPicture = %d\tlastPicture = %d\n",
			parameter->firstPicture, parameter->lastPicture);
	fprintf(fptr, "Coding rate = %f kb/s\n", parameter->bit_rate/1000);
	fprintf(fptr, "VBV buffer size = %d bits\n", parameter->vbv_buffer_size);
	fprintf(fptr, "Picture rate = %f\tgroupOfPicture = %d\t",
			parameter->float_picture_rate, parameter->N);
	fprintf(fptr, "HalfPel = %d\n", parameter->halfPel);
	fprintf(fptr, "kLuminWidth = %d\tkLuminHeight = %d\tkChromWidth = %d\tkChromHeight = %d\n", parameter->luminWidth, parameter->luminHeight, parameter->chromWidth, parameter->chromHeight);
	fprintf(fptr, "Xi = %.2f\tXp = %.2f\tXb = %.2f\n",
			rateControl->Xi, rateControl->Xp, rateControl->Xb);
	fprintf(fptr, "G = %d\tr = %.2f\tTipb = %d\n",
			rateControl->G, rateControl->r, rateControl->Tipb);
	fprintf(fptr, "d0i = %d\td0p = %d\td0b = %d\n",
			rateControl->d0i, rateControl->d0p, rateControl->d0b);
        fprintf(fptr, "M = %d, N = %d\n", parameter->M, parameter->N);
	fclose(fptr);

return (0);
}


