/* File:  motion.c */

/*
 *      $Log:   motion.c,v $
 * Revision 1.1.1.1  93/03/29  11:28:00  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:30:43  au
 * Initial revision
 *
*/

/* Self-reminder: don't forget to update encoder.pro with new modules */

#include <stdio.h>
#include <stdlib.h>
#include "consts.h"
#include "mpeg1.h"
#include "encoder.pro"

/* Exhaustive Minimum Absolute Difference block matching search */

short   GetMotionVector(
 BYTE  *recon_picture,      /* pointer to upper left corner of ref buffer */
 BYTE  *original_picture,   /* pointer to upper left corner of orig picture 
                               co-incident with the current picture under the 
			       encoding process */
 BYTE  *current_picture,     /* pointer to upper left corner of cur picture */
 short MV_array[][2],
 int   search_radius,
 int   half_pel_flag,                /* half pel flag */
 int   fileMV,
 char  *fName,
 tParameter      *parameter)
{
	FILE            *fptr;
/* The line below negates the usefulness of file MV  -- Chad  3/22/94 */        
	int             noFile = 1;


	int     MB_row,         /* current macroblock row */
		MB_col,         /* current macroblock column */
		MB_origin_row,  /* current sample row */
		MB_origin_col,  /* current sample column */
		MB_number;      /* current macroblock number within picture */

	int     MV_count;       /* number of motion vectors, horizontal and 
				   vertical, in picture */
	int     ret = 0;
	
	int     hor,            /* horizontal vector component */
		ver;            /* vertical vector component */

	BYTE    *current_MB; 
	
	
	MV_count = parameter->numMacroBlocks * 2;

	if (fileMV)             /*      use file MV     */
	{
		if ((fptr = fopen (fName, "rb")) != NULL)
		{
			noFile = 0;

			fread(MV_array, sizeof(short), MV_count, fptr);
			fclose(fptr);
		}
	}
	
	if (noFile)
	{       /* find match for all macroblocks within target picture */
		
		for (MB_row = 0, MB_origin_row = 0, MB_number = 0; 
				MB_row < parameter->numSlices; MB_row++, 
				MB_origin_row += kMBHeight)
		{       /* compute pointer address of current MB */
			current_MB = (BYTE *) (current_picture + (MB_origin_row *parameter->luminWidth));
			
			/* macroblock row */
			for (MB_col = 0, MB_origin_col = 0; MB_col < 
					parameter->numMBsPerSlice; MB_col++, 
					MB_origin_col+= kMBWidth, MB_number++, current_MB 
					+= kMBWidth)
			{
				
				ret = FullSearch(
					&hor,
					&ver, 
					search_radius,                
					half_pel_flag,
					recon_picture,
                                        current_picture,
					original_picture,

/* 3/21/94 A bug was reported by Mr. Hirofumi Nisikawa and Kinya Oosa.  Current 
		has been changed to currrent1 */ 

					current_MB,
					MB_origin_row,
					MB_origin_col,
					parameter);
				
				
				MV_array[MB_number][0] = hor;
				MV_array[MB_number][1] = ver;
			
			}
			/* add tracelevel here */
			printf("Row %d of %d motion search completed.\n", 
				MB_row, parameter->numSlices);
		}
		
		if ((fptr = fopen (fName, "wb")) != NULL)
		{
			fwrite(MV_array, sizeof(short), MV_count, fptr);
			fclose(fptr);
		}

		else return (-1);
	}

	return (ret);
}


int FullSearch(
	int                     *hor,
	int                     *ver,
	short                   search_radius,                
	int                     half_pel_flag,
	BYTE                    *recon_picture,
        BYTE                    *current_picture,
	BYTE                    *original_picture,
	BYTE                    *current_MB,
	int                     MB_origin_row,
	int                     MB_origin_col,
	tParameter              *parameter)

/* need to update bufMV array to either be a passed pointer or a
return value */
		
	
	{
	
	int     AE_F = 0;           /* absolute error value */
	unsigned int    minF;

	BYTE    *prediction_MB;

	int     dis_radius,             /* distance along radius */
		search_quadrant,        /* sector */
		dis_diameter,           /* distance along diameter */
		diameter,               /* search diameter */
		dis_x_origin,           /* displacement from search origin */
		dis_y_origin,      
		mxF,                    /* buffer location of x MV element */
		myF,
		mv[2];
	

	/* compute address of reference macroblock with no motion */
	prediction_MB = original_picture + (MB_origin_row * parameter->luminWidth) + MB_origin_col; 
	
	/* compute distortion of target center (no motion compensation) */
	AE_F = MAE(current_MB, prediction_MB, parameter);

	minF = AE_F;    /* default vector is zero vector */
	mv[0] = 0;
	mv[1] = 0;

	/* now perform spiral search around origin of current macroblock */
	for (dis_radius = 1; dis_radius <= search_radius; dis_radius++)
	{
		/* relative displacement along search axis */
		dis_x_origin = dis_radius; 
		dis_y_origin = -dis_radius;     /* start in quandrant III */
		
		/* total search width is equal to twice search distance (f_code)
		    like diameter is twice radius */
		diameter = dis_radius << 1;

		/* seach each of the four Cartesian co-ordinate sectors */
		for (search_quadrant = 0; search_quadrant < 4; search_quadrant++)
		{
			/* total linear search span */
			for (dis_diameter = 0; dis_diameter < diameter; dis_diameter++)
			{       
				/* relative displacement from target center */
				myF = MB_origin_row + dis_y_origin;    
				mxF = MB_origin_col + dis_x_origin;

/* need to change kLuminWidthMB into more readable constant */
				
				/* make sure search window boundaries are within legal picture boundaries */
				if ((mxF >= 0) && (mxF < (parameter->luminWidthMB + 1)) 
					&& (myF >= 0) && 
						(myF < (parameter->luminHeightMB + 1)))
				{
				  /*      frame   */

					/* compute prediction address pointer */
					prediction_MB = original_picture + 
						(myF * parameter->luminWidth) + mxF;
					
					AE_F = MAE(current_MB, prediction_MB, parameter);

					/* update lowest distortion vector */
					if (AE_F < minF)
					{
						minF = AE_F;
						mv[0] = dis_x_origin;
						mv[1] = dis_y_origin;
					}
				}

				/* increment search displacement */
				switch (search_quadrant)
				{
				case 0:
					dis_y_origin++;
					break;
				case 1:
					dis_x_origin--;
					break;
				case 2:
					dis_y_origin--;
					break;
				case 3:
					dis_x_origin++;
					break;
				}
			}
		}
	}

	*hor = 2 * mv[0];          /*      frame MV half pel       */
	*ver = 2 * mv[1];          /*      frame MV half pel       */

	/*      half pel search */

	if (half_pel_flag)
	{
		minF = HalfPel(recon_picture, current_picture, hor, ver, MB_origin_row,
			MB_origin_col, minF, 1, parameter);
	}


return (0);
}


/* compute absolute error between prediction and target macroblock */

int MAE (BYTE           *current_MB, 
	 BYTE           *prediction_MB,
	 tParameter     *parameter)                                
{        
	int x, y;        
	int absolute_error = 0;

	for (y = 0; y < kMBHeight; y++, current_MB += parameter->luminWidthMB, 
		prediction_MB += parameter->luminWidthMB)
	{
		for (x = 0; x < kMBWidth; x++, current_MB++, prediction_MB++)
		{
			absolute_error += abs(*current_MB - *prediction_MB);
		}
	}

	return (absolute_error);
}


/* Half pel refinement as described in ISO/IEC 11172-2 Annex D.?.?.? */

/* implementation note:  simplify this routine using calls to MAE()  */
/* I know this routine can be made much more concise ! -- Chad */

unsigned int    HalfPel(
	BYTE                    *original_picture,
	BYTE                    *current_picture,
	int                     *hor,
	int                     *ver,
	int                     MB_origin_row,
	int                     MB_origin_col,
	unsigned int            minF,
	short                   offset,
	tParameter              *parameter)
{
	BYTE    *org, *org1, *curr;
	short   dx, dy;
		short   i,
				row,
				col;
		
		int             offset1,
				sum,
				diff,
				MB_origin_colA,
				MB_origin_rowA;


	MB_origin_colA = MB_origin_col + *hor / 2;
	MB_origin_rowA = MB_origin_row + *ver / 2;

	offset1 = offset * parameter->luminWidth - kMBWidth;
	dx = 0;
	dy = 0;

	/* 8 neighbour pel search on */
	for (i = 0; i < 8; i++)
	{
		switch (i)
		{
			case 0:         /*      upper left      */
			
				if ((MB_origin_rowA <= 0) || (MB_origin_colA <= 0))
					break;

				org1 = (BYTE *) (original_picture + MB_origin_rowA * 
					parameter->luminWidth + MB_origin_colA -1);
				org = (BYTE *) (org1 - offset * 
					parameter->luminWidth);
				curr = (BYTE *) (current_picture + MB_origin_row * 
					parameter->luminWidth + MB_origin_col);

				for (row = 0, sum = 0; row < kMBHeight; row += offset,
						org += offset1, org1 += offset1, curr += offset1)
				{
					for (col = 0; col < kMBWidth; col++, org++, org1++, curr++)
					{
						/*      add 2 for integer rounding      */

						diff = ((int) *org + (int) *(org + 1) +
								(int) *org1 + (int) *(org1 + 1) + 2) / 4;

						sum += abs (*curr - diff);
					}
				}

				if (sum < minF)   
				{
					minF = sum;
					dx = -1;
					dy = -1;
				}

				break;

			case 1:         /*      upper   */
			
				if (MB_origin_rowA <= 0)
					break;

				org1 = (BYTE *) (original_picture + MB_origin_rowA * parameter->luminWidth + MB_origin_colA);
				org = (BYTE *) (org1 - offset * parameter->luminWidth);
				curr = (BYTE *) (current_picture + MB_origin_row * parameter->luminWidth + MB_origin_col);

				for (row = 0, sum = 0; row < kMBHeight; row += offset,
						org += offset1, org1 += offset1, curr += offset1)
				{
					for (col = 0; col < kMBWidth; col++, org++, org1++, curr++)
					{
						/*      add 1 for integer rounding      */

						diff = ((int) *org + (int) *org1 + 1) / 2;

						sum += abs (*curr - diff);
					}
				}

				if (sum < minF)
				{
					minF = sum;
					dx = 0;
					dy = -1;
				}
			
				break;

			case 2:         /*      upper right     */
			
				if ((MB_origin_rowA <= 0) || (MB_origin_colA >= (parameter->luminWidth - kMBWidth - 1)))
					break;

				org1 = (BYTE *) (original_picture + MB_origin_rowA * parameter->luminWidth + MB_origin_colA);
				org = (BYTE *) (org1 - offset * parameter->luminWidth);
				curr = (BYTE *) (current_picture + MB_origin_row * parameter->luminWidth + MB_origin_col);

				for (row = 0, sum = 0; row < kMBHeight; row += offset,
						org += offset1, org1 += offset1, curr += offset1)
				{
					for (col = 0; col < kMBWidth; col++, org++, org1++, curr++)
					{
						/*      add 2 for integer rounding      */

						diff = ((int) *org + (int) *(org + 1) +
								(int) *org1 + (int) *(org1 + 1) + 2) / 4;

						sum += abs (*curr - diff);
					}
				}

				if (sum < minF)
				{
					minF = sum;
					dx = 1;
					dy = -1;
				}
			
				break;

			case 3:         /*      left    */
			
				if (MB_origin_colA <= 0)
					break;

				org = (BYTE *) (original_picture + MB_origin_rowA * parameter->luminWidth + MB_origin_colA - 1);
				curr = (BYTE *) (current_picture + MB_origin_row * parameter->luminWidth + MB_origin_col);

				for (row = 0, sum = 0; row < kMBHeight; row += offset,
						org += offset1, curr += offset1)
				{
					for (col = 0; col < kMBWidth; col++, org++, curr++)
					{
						/*      add 1 for integer rounding      */

						diff = ((int) *org + (int) *(org + 1) + 1) / 2;

						sum += abs (*curr - diff);
					}
				}

				if (sum < minF)
				{
					minF = sum;
					dx = -1;
					dy = 0;
				}
			
				break;

			case 4:         /*      right   */
			
				if (MB_origin_colA >= (parameter->luminWidth - kMBWidth - 1))
					break;

				org = (BYTE *) (original_picture + MB_origin_rowA * parameter->luminWidth + MB_origin_colA);
				curr = (BYTE *) (current_picture + MB_origin_row * parameter->luminWidth + MB_origin_col);

				for (row = 0, sum = 0; row < kMBHeight; row += offset,
						org += offset1, curr += offset1)
				{
					for (col = 0; col < kMBWidth; col++, org++, curr++)
					{
						/*      add 1 for integer rounding      */

						diff = ((int) *org + (int) *(org + 1) + 1) / 2;

						sum += abs (*curr - diff);
					}
				}

				if (sum < minF)
				{
					minF = sum;
					dx = 1;
					dy = 0;
				}
			
				break;

			case 5:         /*      lower left      */
			
				if ((MB_origin_rowA >= (parameter->luminHeight - kMBHeight - 1)) ||
						(MB_origin_colA <= 0))
					break;

				org = (BYTE *) (original_picture + MB_origin_rowA * parameter->luminWidth + MB_origin_colA - 1);
				org1 = (BYTE *) (org + offset * parameter->luminWidth);
				curr = (BYTE *) (current_picture + MB_origin_row * parameter->luminWidth + MB_origin_col);

				for (row = 0, sum = 0; row < kMBHeight; row += offset,
						org += offset1, org1 += offset1, curr += offset1)
				{
					for (col = 0; col < kMBWidth; col++, org++, org1++, curr++)
					{
						/*      add 2 for integer rounding      */

						diff = ((int) *org + (int) *(org + 1) +
								(int) *org1 + (int) *(org1 + 1) + 2) / 4;

						sum += abs (*curr - diff);
					}
				}

				if (sum < minF)
				{
					minF = sum;
					dx = -1;
					dy = 1;
				}
			
				break;

			case 6:         /*      lower   */
			
				if (MB_origin_rowA >= (parameter->luminHeight - kMBHeight - 1))
					break;

				org = (BYTE *) (original_picture + MB_origin_rowA * parameter->luminWidth + MB_origin_colA);
				org1 = (BYTE *) (org + offset * parameter->luminWidth);
				curr = (BYTE *) (current_picture + MB_origin_row * parameter->luminWidth + MB_origin_col);

				for (row = 0, sum = 0; row < kMBHeight; row += offset,
						org += offset1, org1 += offset1, curr += offset1)
				{
					for (col = 0; col < kMBWidth; col++, org++, org1++, curr++)
					{
						/*      add 1 for integer rounding      */

						diff = ((int) *org + (int) *org1 + 1) / 2;

						sum += abs (*curr - diff);
					}
				}

				if (sum < minF)
				{
					minF = sum;
					dx = 0;
					dy = 1;
				}
			
				break;

			case 7:         /*      lower right     */
			
				if ((MB_origin_rowA >= (parameter->luminHeight - kMBHeight - 1)) ||
						(MB_origin_colA >= (parameter->luminWidth - kMBWidth - 1)))
					break;

				org = (BYTE *) (original_picture + MB_origin_rowA * parameter->luminWidth + MB_origin_colA);
				org1 = (BYTE *) (org + offset * parameter->luminWidth);
				curr = (BYTE *) (current_picture + MB_origin_row * parameter->luminWidth + MB_origin_col);

				for (row = 0, sum = 0; row < kMBHeight; row += offset,
						org += offset1, org1 += offset1, curr += offset1)
				{
					for (col = 0; col < kMBWidth; col++, org++, org1++, curr++)
					{
						/*      add 2 for integer rounding      */

						diff = ((int) *org + (int) *(org + 1) +
								(int) *org1 + (int) *(org1 + 1) + 2) / 4;

						sum += abs (*curr - diff);
					}
				}

				if (sum < minF)
				{
					minF = sum;
					dx = 1;
					dy = 1;
				}
			
				break;
		}
	}

	*hor += dx;
	*ver += dy;

	return (minF);
}
