/* File: mpeg1.h */

/*
 *      $Log:   mpeg1.h,v $
 * Revision 1.4  93/09/03  1:15:00  hana
 * VBV buffer operation 
 * 
 * Revision 1.1.1.1  93/03/29  11:27:53  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:32:40  au
 * Initial revision
 *
 */

typedef struct t_Data           /*      picture data    */
{
	/*      working pointers        */
	BYTE *firstPredictY;    /*  1st prediction picture  */
	BYTE *firstPredictU;    /*  1st prediction picture  */
	BYTE *firstPredictV;    /*  1st prediction picture  */
	BYTE *secondPredictY;   /*  2nd prediction picture  */
	BYTE *secondPredictU;   /*  2nd prediction picture  */
	BYTE *secondPredictV;   /*  2nd prediction picture  */
	BYTE *predictY;         /*  B picture       */
	BYTE *predictU;         /*  B picture       */
	BYTE *predictV;         /*  B picture       */
	BYTE *predictY2;        /*  2nd B picture   */
	BYTE *predictU2;        /*  2nd B picture   */
	BYTE *predictV2;        /*  2nd B picture   */

	BYTE *firstPredictYR;   /*  reconstructed 1st prediction picture */
	BYTE *firstPredictUR;   /*  reconstructed 1st prediction picture */
	BYTE *firstPredictVR;   /*  reconstructed 1st prediction picture */
	BYTE *secondPredictYR;  /*  reconstructed 2nd prediction picture */
	BYTE *secondPredictUR;  /*  reconstructed 2nd prediction picture */
	BYTE *secondPredictVR;  /*  reconstructed 2nd prediction picture */
	BYTE *predictYR;        /*  reconstructed B picture */
	BYTE *predictUR;        /*  reconstructed B picture */
	BYTE *predictVR;        /*  reconstructed B picture */

	/*      storage arrays  */

	BYTE *luminPicture0;    /*      luminance of picture 0  */
	BYTE *luminPicture1;    /*      luminance of picture 1  */
	BYTE *luminPicture2;    /*      luminance of picture 2  */
	BYTE *chromUPicture0;   /*      chrominance of picture 0        */
	BYTE *chromUPicture1;   /*      chrominance of picture 1        */
	BYTE *chromUPicture2;   /*      chrominance of picture 2        */
	BYTE *chromVPicture0;   
	BYTE *chromVPicture1;
	BYTE *chromVPicture2;

	BYTE *luminPicture0R;   /* reconstructed luminance of picture 0  */
	BYTE *luminPicture1R;   /* reconstructed luminance of picture 1  */
	BYTE *luminPicture2R;   /* reconstructed luminance of picture 2  */
	BYTE *chromUPicture0R;  /* reconstructed chrominance of picture 0  */
	BYTE *chromUPicture1R;  /* reconstructed chrominance of picture 1  */
	BYTE *chromUPicture2R;  /* reconstructed chrominance of picture 2  */
	BYTE *chromVPicture0R;
	BYTE *chromVPicture1R;
	BYTE *chromVPicture2R;
} tData;

typedef struct t_Parameter              /*      user parameter  */
{
		char    inputSequence[kMaxCharBufLen];
		char    outputSequence[kMaxCharBufLen];
		char    statusFName[kMaxCharBufLen];
		char    bitStreamFName[kMaxCharBufLen];

                int     sourceFileFormat; /* file type for input pictures */
                int     reconFileFormat;  /* file type for output pictures */
		int     firstPicture;
		int     lastPicture;
	int     temporal_reference;       /* GOP picture number */
	int     pictureNum;               /* sequence picture number */
		int     horizontal_size;  /* src&display picture width sample units */
		int     mb_width;         /* coded picture width in macroblock units */
		int     vertical_size;    /* display picture height in sample units */
		int     mb_height;        /* coded picture height in macroblocks units */
		int     pel_aspect_ratio; /* one of 14 legal values */
		int     picture_rate;     /* one of 8 legal values */
		double  float_picture_rate;  /* one of 8 legal values */
	       
		double  bit_rate;           /* bit/sec */
		int     vbv_buffer_size;    /* bits, used to check vbv_delay */
		int     constrained_parameters_flag;        

		int     N;    /* number of frames in GOP */
		int     M;    /* Distance between P pictures -- "M" factor */

		int     load_intra_quantizer_matrix;
		int     load_non_intra_quantizer_matrix;

        int     closed_GOP;
        int     broken_link;
	int     drop_frame_flag;
	long    time_code;

		int     luminWidth;  /* coded picture width in sample units */
		int     luminHeight; /* coded picture height in sample units */
		int     chromWidth; 
		int     chromHeight; 
		
		long    numLuminPixels; /* number of coded samples in picture */
		long    numChromPixels;
	
		int     numSlices;      /* number of slices in picture */
		int     numMBsPerSlice; /* number of macroblocks per slice */
		int     numMacroBlocks; /* number of macroblocks per picture */
		
	/* source picture dimensions in units of samples, clipped  
	   to not extend beyond the outermost macroblock or block origin */
	
		int     luminHeightMB;
		int     luminWidthMB; 
		int     luminWidthDCT;
		int     chromWidthB;
		int     chromWidthDCT;

		int     mBLuminRowOffset;
		int     mBChromRowOffset; 
		
		double  maxPictureDelay;
		int     maxVbvBufferSize;
	
		int     searchDistance;
		int     *forward_f_code;  /*  0-P, 1-1st B, 2-2nd B Picture   */
		int     *backward_f_code; /*  0-P, 1-1st B, 2-2nd B Picture   */
		int     *fRange;  /* 0-P Picture, 1-1st B Picture, 2-2nd B Picture   */
		int     *bRange; /*  0-P Picture, 1-1st B Picture, 2-2nd B Picture   */
		int     halfPel;  /* half pel search enable if set true */
		int     fileMV; /*  use file MV when possible if set true */

} tParameter;


typedef struct t_RateControl            /*      rate control    */
{
	int    headerBits;     /*  total number of header bits encoded   */
	int    totalGOPBits;   /*  total number of GOP bits encoded   */
	int    totalPBits;     /*  total number of current frame bits encoded */
	int    macroblock_address_increment;

	int    R;      /*      remaining # of bits assigned to the GOP */
	int    G;      /*      bit_rate * N / picture_rate     */
	double r;      /*      reaction parameter      */

	double avg_act;        /*      average actj of last frame   */
	double actjSum;        /*      sum of actj of current frame   */

	int     d0i;    /*  initial fullness of virtual buffers - I type    */
	int     d0p;    /*  initial fullness of virtual buffers - P type    */
	int     d0b;    /*  initial fullness of virtual buffers - B type    */
	int     dji;    /*  current fullness of virtual buffers - I type    */
	int     djp;    /*  current fullness of virtual buffers - P type    */
	int     djb;    /*  current fullness of virtual buffers - B type    */
	int     Tipb;   /*  bit_rate / (8 * picture_rate)   */
	int     Ti;     /*  target # bits for next frame in GOP - I type    */
	int     Tp;     /*  target # bits for next frame in GOP - P type    */
	int     Tb;     /*  target # bits for next frame in GOP - B type    */
	int     Tij;    /*  target # bits for each MB - I type  */
	int     Tpj;    /*  target # bits for each MB - P type  */
	int     Tbj;    /*  target # bits for each MB - B type  */
	int     bj;     /* # bits generated encoding all MBs up to MB j  */

	int     Np;     /*  # of P-frame remaining in current GOP  */
	int     Nb;     /*  # of B-frame remaining in current GOP  */

	double  Xi;     /*  global complexity measure - I type   */
	double  Xp;     /*  global complexity measure - P type  */
	double  Xb;     /*  global complexity measure - B type  */
	int     Si;     /*  # of bits generated by encoding current I frame */
	int     Sp;     /*  # of bits generated by encoding current P frame */
	int     Sb;     /*  # of bits generated by encoding current B frame */
	double  Qi;     /*  average quantization parameter - I type  */
	double  Qp;     /*  average quantization parameter - P type  */
	double  Qb;     /*  average quantization parameter - B type  */
	double  Qj;     /*  reference quantization parameter   */

	double  vbvOccupy; /* VBV buffer occupancy behave as B_n in VBV 
			       operation */
	int     vbv_delay;
} tRateControl;


