/* File: PerfIdct.c  */

/*
 *      $Log:   PerfIdct.c,v $
 * Revision 1.1.1.1  93/03/29  11:27:54  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:33:17  au
 * Initial revision
 *
 */

#include <stdio.h>
#include "consts.h"
#include "mpeg1.h"
#include "encoder.pro"

/* de-zz scan order offset */

static BYTE     dzz[64] =
{
     0,  1,  8, 16,  9,  2,  3, 10,
    17, 24, 32, 25, 18, 11,  4,  5,
    12, 19, 26, 33, 40, 48, 41, 34,
    27, 20, 13,  6,  7, 14, 21, 28,
    35, 42, 49, 56, 57, 50, 43, 36,
    29, 22, 15, 23, 30, 37, 44, 51,
    58, 59, 52, 45, 38, 31, 39, 46,
    53, 60, 61, 54, 47, 55, 62, 63
};


void    PerformIDct(
	int             mbType,
	int             *currentMBDiff,
	int             *currentMBDiffU,
	int             *currentMBDiffV)
{
	int             currentIndex, currentIndex1, currentIndex2, line, pix, row, col;
	float           dctBuf[kDCTLen], dctTempBuf[kDCTLen], dctTrBuf[kDCTLen];

	for (line = 0; line < kMBHeight; line += kDCTSize)
	{
		for (pix = 0; pix < kMBWidth; pix += kDCTSize)
		{
			currentIndex2 = line * kMBWidth + pix;

			for (row = 0, currentIndex = 0, currentIndex1 = currentIndex2;
					row < kDCTSize; row++, currentIndex1 += kMBWidthDCT)
			{
				for (col = 0; col < kDCTSize; col++,
						currentIndex++, currentIndex1++)
				{
					/*      perform de-zz scan      */

					dctTrBuf[dzz[currentIndex]] =
							currentMBDiff[currentIndex1];
				}
			}

			/*      perform inverse DCT     */

			idct8x8(dctTrBuf, dctBuf);

			/*      return re-transformed data              */

			for (row = 0, currentIndex = 0, currentIndex1 = currentIndex2;
					row < kDCTSize; row++, currentIndex1 += kMBWidthDCT)
			{
				for (col = 0; col < kDCTSize; col++,
						currentIndex++, currentIndex1++)
				{
					currentMBDiff[currentIndex1] =
							(dctBuf[currentIndex] > 0.0) ?
							(dctBuf[currentIndex] + 0.5) :
							(dctBuf[currentIndex] - 0.5);
				}
			}
		}
	}

	/*      chrom   */

	/*      U       */

	for (row = 0; row < kDCTLen; row++)
	{
		/*      perform de-zz scan      */

		dctTrBuf[dzz[row]] = currentMBDiffU[row];
	}

	/*      perform IDCT    */

	idct8x8(dctTrBuf, dctBuf);

	/*      return re-transformed data              */

	for (row = 0; row < kDCTLen; row++)
	{
		currentMBDiffU[row] = (dctBuf[row] > 0.0) ?
				(dctBuf[row] + 0.5) : (dctBuf[row] - 0.5);
	}

	/*      V       */

	for (row = 0; row < kDCTLen; row++)
	{
		/*      perform de-zz scan      */

		dctTrBuf[dzz[row]] = currentMBDiffV[row];
	}

	/*      perform DCT     */

	idct8x8(dctTrBuf, dctBuf);

	/*      return re-transformed data              */

	for (row = 0; row < kDCTLen; row++)
	{
		currentMBDiffV[row] = (dctBuf[row] > 0.0) ?
				(dctBuf[row] + 0.5) : (dctBuf[row] - 0.5);
	}
}
