/* File: procPicture.c */

/* we may wish to generate a structure or add to tData{}  pointers
containing all the working picture buffers, target address (CurrentMBDiff)
etc.  This would reduce the size of many call parameters, e.g. 
Constr(), etc.  */

/*
 *      $Log:   procPicture.c,v $
 * Revision 1.4  93/09/03  01:15:00  hana
 * VBV buffer operation 
 * 
 * Revision 1.2  93/06/15  15:23:18  oosa
 * 
 * 
 * Revision 1.1  1993/03/10  20:30:43  au
 * Initial revision
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "consts.h"
#include "mpeg1.h"
#include "flc.h"
#include "encoder.pro"

void    ProcPicture(
	tData           *data,
	tParameter      *parameter,
	tRateControl    *rateControl,
	BYTE            *bufY,          /* source picture buffer */
	BYTE            *bufU,
	BYTE            *bufV,
	BYTE            *bufYR,         /* reconstruction picture buffer */
	BYTE            *bufUR,
	BYTE            *bufVR,
	int             picture_coding_type,
	int             pictureIndex,   /* P=0,B1=1,B2=2,... */
	short           forwardMV[][2], /* motion vector field */
	short           backwardMV[][2],
	int             macroblock_type[],
	short           coded_block_pattern[],
	int             MQuant[])
{
	BYTE    currentMB[kMBLen];
	BYTE    currentMBU[kBLen];
	BYTE    currentMBV[kBLen];
	BYTE    *predictMBY, *predictMBU, *predictMBV;     /* forward reference */
	BYTE    *predictMBYB, *predictMBUB, *predictMBVB;  /* backwards reference */
	BYTE    *reconY, *reconU, *reconV;                 /* reconstruction pictute */
	int     currentMBDiff[kMBLen];                     /* DFD macroblock */
	int     currentMBDiffU[kBLen];
	int     currentMBDiffV[kBLen];

	static int      totalBits = 0;
	
	/* target location pointers. bufIndex is the reconstruction cursor */
	int             MB_row, MB_col, bufIndex, bufIndexC, MB_index;
	
	int             tempInt, newMbType, Txj;
	int             fRange, forward_f_code, bRange, backward_f_code;

	/*      set fRange and forward_f_code   */

	forward_f_code = parameter->forward_f_code[pictureIndex];
	fRange = parameter->fRange[pictureIndex];

	/*      set bRange and backward_f_code  */

	backward_f_code = parameter->backward_f_code[pictureIndex];
	bRange = parameter->bRange[pictureIndex];

	/* perform pre-frame analysis and write GOP header */

fprintf(stderr,"Preframe analysis now beginning\n");

	PreFrame(picture_coding_type, parameter, rateControl,
			forward_f_code, backward_f_code);

	
	switch (picture_coding_type)
	{
		case kIPicture:

			Txj = rateControl->Tij;
			break;

		case kPPicture:

			Txj = rateControl->Tpj;
			break;

		case kBPicture:

			Txj = rateControl->Tbj;
			break;
	}

printf(stderr, "Picture_type = %d\n", picture_coding_type);

	/* process each slice */
	for (MB_row = 0, MB_index = 0, bufIndex = 0, bufIndexC = 0;
			MB_row < parameter->numSlices; MB_row++, bufIndex += 
			parameter->mBLuminRowOffset,bufIndexC += 
				parameter->mBChromRowOffset)

	{
		switch (picture_coding_type)
		{
			case kBPicture:

				rateControl->djb += Write_Slice_Header(picture_coding_type, rateControl);
				break;

			case kPPicture:

				rateControl->djp += Write_Slice_Header(picture_coding_type, rateControl);
				break;

			case kIPicture:

				rateControl->dji += Write_Slice_Header(picture_coding_type, rateControl);
				break;
		}
		
fprintf(stderr, "Written slice header \n");

		/* process each macroblock */
		for (MB_col = 0; MB_col < parameter->numMBsPerSlice;
				MB_col++, MB_index++, bufIndex += kMBWidth, bufIndexC += kBWidth)
		{
			
			/* copy source MB into currentMB and currentMBDiff */
			
			CopyToBuf(bufIndex, bufIndexC, bufY, bufU, bufV, currentMB,
					currentMBU, currentMBV, currentMBDiff,
					currentMBDiffU, currentMBDiffV, parameter);

			/* update luma reconstruction pointers by bufIndex */

			reconY = bufYR + bufIndex;
			predictMBY = data->firstPredictY + bufIndex;
			predictMBYB = data->secondPredictY + bufIndex;

			/* update chrominance pointers   */

			reconU = bufUR + bufIndexC;
			reconV = bufVR + bufIndexC;
			predictMBU = data->firstPredictU + bufIndexC;
			predictMBV = data->firstPredictV + bufIndexC;
			predictMBUB = data->secondPredictU + bufIndexC;
			predictMBVB = data->secondPredictV + bufIndexC;

			/*      compute rate control status     */

			DoPreMBRC(picture_coding_type, rateControl, currentMB, &MQuant[MB_index]);

			if (! MB_col)   
			{
				/*     Write slice header  */

				writeBits(cSliceStartCode + MB_row , bSliceStartCode, 1);
				writeBits(MQuant[MB_index], bQuantizerScale, 0);
				writeBits(cExtraBitSlice, bExtraBitSlice, 0);
			}

			/*      construct macroblock    */
			
			switch (picture_coding_type)
			{
				case kIPicture:

					macroblock_type[MB_index] = ConstructI(reconY, reconU, reconV,
							currentMB, currentMBU, currentMBV,
							currentMBDiff, currentMBDiffU, currentMBDiffV,
							parameter);
					break;
			
				case kPPicture:

					macroblock_type[MB_index] = Construct(
							predictMBY, predictMBU, predictMBV,
							reconY, reconU, reconV,
							currentMB, currentMBU, currentMBV,
							forwardMV[MB_index], currentMBDiff,
							currentMBDiffU, currentMBDiffV,
							parameter);
					break;

				default:        /*      B picture       */

					macroblock_type[MB_index] = ConstructB(
							predictMBY, predictMBU, predictMBV,
							predictMBYB, predictMBUB, predictMBVB,
							reconY, reconU, reconV,
							currentMB, currentMBU, currentMBV,
							forwardMV[MB_index], backwardMV[MB_index],
							currentMBDiff, currentMBDiffU, currentMBDiffV, parameter);
					break;
			}

			/*      perform DCT and Zig-Zag scanning        */

			PerformDct(macroblock_type[MB_index], currentMB, currentMBDiff,
					currentMBDiffU, currentMBDiffV, MB_index);

			/*      perform quantization and de-quantization        */

			if (macroblock_type[MB_index] & kIntraType)
				macroblock_type[MB_index] |= QuantizeI(&coded_block_pattern[MB_index],
						currentMBDiff, currentMBDiffU, currentMBDiffV,
						MQuant[MB_index]);
			else
				macroblock_type[MB_index] |= Quantize(&coded_block_pattern[MB_index],
						currentMBDiff, currentMBDiffU, currentMBDiffV,
						MQuant[MB_index]);

			if (macroblock_type[MB_index] & kCodedType)
			{
				/*      check for modified MQuant       */

				if (MB_col && (MQuant[MB_index] != MQuant[MB_index - 1]))
					macroblock_type[MB_index] |= kModifiedMQ;
			}
			else if (MB_col)         /*      not coded       */
				MQuant[MB_index] = MQuant[MB_index - 1];

			/*      update rate control status      */

			tempInt = Write_MB(picture_coding_type, macroblock_type[MB_index],
					coded_block_pattern[MB_index], rateControl, parameter, MB_row,
					MB_col, forwardMV[MB_index], backwardMV[MB_index],
					currentMBDiff, currentMBDiffU, currentMBDiffV,
					MQuant[MB_index], fRange, forward_f_code, bRange,
					backward_f_code, &newMbType);

			if (! tempInt)  /*      skipped MB      */
			{
				/*      if here => MB_col != 0, MB_index != 0     */

				MQuant[MB_index] = MQuant[MB_index - 1];
			}
			else
			{
				/*      reconstruct mb  */

				if (macroblock_type[MB_index] & kIntraType)
					IQuantizeI(coded_block_pattern[MB_index], currentMBDiff,
							currentMBDiffU, currentMBDiffV, MQuant[MB_index]);
				else
					IQuantize(coded_block_pattern[MB_index], currentMBDiff,
							currentMBDiffU, currentMBDiffV, MQuant[MB_index]);

				PerformIDct(macroblock_type[MB_index], currentMBDiff,
						currentMBDiffU, currentMBDiffV);

				/* Copy current MB difference into reconstruction buffer */
				
				CopyFromBuf(picture_coding_type, macroblock_type[MB_index],
						coded_block_pattern[MB_index], MB_row, MB_col, bufYR, bufUR,
						bufVR, currentMBDiff, currentMBDiffU, currentMBDiffV,
						MB_index, parameter);
			}

			/*      update quanitzation step size rate control data */

			switch(picture_coding_type)
			{
				case kBPicture:

					rateControl->Qb += MQuant[MB_index];
					break;

				case kPPicture:

					rateControl->Qp += MQuant[MB_index];
					break;

				case kIPicture:

					rateControl->Qi += MQuant[MB_index];
					break;
			}
		}
	}

	checkUpdateVbv(picture_coding_type, parameter->pictureNum, rateControl, parameter);
	
	PostFrame(picture_coding_type, rateControl, parameter);
	
	switch (picture_coding_type)
	{
		case kIPicture:

			totalBits += rateControl->Si;
			break;

		case kPPicture:

			totalBits += rateControl->Sp;
			break;

		case kBPicture:

			totalBits += rateControl->Sb;
			break;
	}

	/* Write reconstructed picture to file */
	if (WritePicture(parameter->pictureNum, picture_coding_type, bufYR, bufUR, 
			bufVR, bufY, bufU, bufV, parameter))
	{
		writeBits (0, 0, -1);
		exit (-1);
	}
}
