/* File:  procseq.c  */

#include <stdio.h>
#include <stdlib.h>
#include "consts.h"
#include "mpeg1.h"
#include "flc.h"
#include "encoder.pro"

void ProcSequence(
	tParameter      *parameter,
	tData           *data,
	tRateControl    *rateControl)
{
	
	int             picture_coding_type;

	BYTE            *bufY, *bufU, *bufV;    /* source picture */
	BYTE            *bufYR, *bufUR, *bufVR; /* reconstruction picture */
	static short    (*forwardMV)[2], (*backwardMV)[2];
	int             index1, index2;
	int		temporal_reference_IP;
	static int      *macroblock_type;                
	char            fName[16];

/*
*      macro block type:
*
*      bit:
*
*              1 - MC(1)/No MC (0)
*              2 - Intra(1)/Inter(0)
*              4 - Coded(1)/Not Coded(0)
*              5 - Forward MV
*              6 - Backward MV
*              7 - modified quantizer
*/
	static short *coded_block_pattern; /*coded block pattern*/
	static int *MQuant; /* macroblock quantization array */
	int     GOPnumber = -1;                 /* number of GOP's coded */

    /* allocate arrays */
    forwardMV = (short (*)[2])malloc(
		     parameter->numMacroBlocks * sizeof(short [2]));
    backwardMV = (short (*)[2])malloc(
		     parameter->numMacroBlocks * sizeof(short [2]));
    macroblock_type = Iarray(parameter->numMacroBlocks);
    coded_block_pattern = Sarray(parameter->numMacroBlocks);
    MQuant = Iarray(parameter->numMacroBlocks);

	/* write sequence header */
	Write_Sequence_Header(parameter, rateControl);

fprintf(stderr, "Sequence header written to buffer\n");

	/*      read first I source picture    */

	parameter->pictureNum = parameter->firstPicture;
	parameter->temporal_reference = temporal_reference_IP = 0; 
	    /* assume the first picture is I-picture */
	picture_coding_type = kIPicture;
	bufY = data->secondPredictY;
	bufU = data->secondPredictU;
	bufV = data->secondPredictV;
	bufYR = data->secondPredictYR;
	bufUR = data->secondPredictUR;
	bufVR = data->secondPredictVR;

	if (ReadPicture(parameter->pictureNum, bufY, bufU, bufV, parameter))
		exit(-1);

fprintf(stderr, "First picture read\n");

	/*      process each MB of first I picture      */

	ProcPicture(data, parameter, rateControl,
			bufY, bufU, bufV, bufYR, bufUR, bufVR,
			picture_coding_type, 0, forwardMV, backwardMV, macroblock_type,
			coded_block_pattern, MQuant);

fprintf(stderr,"Picture coding type= %d\n", picture_coding_type);

	/*      process sequence        */

	/* index1 points to the next I/P-picture */
	for (index1 = parameter->M;
			index1 <= (parameter->lastPicture - parameter->firstPicture);
			index1 += parameter->M)
	{
		/* Process BB...BP sequence */
		/* index2(!=0) points to the current B-picture */
		for (index2 = 0; index2 < parameter->M; index2++)
		{
		    
			if(index2 == 0)
			{
			    /* I,P-picture */
			    parameter->pictureNum = parameter->firstPicture + index1;
			}
			else
			{
			    /* B-picture */
			    parameter->pictureNum = parameter->firstPicture
				    + index1 - parameter->M + index2;
			}
			    
			/*      reassign reconstructed frame for next prediction                */

			switch (picture_coding_type)    /*      according to OLD picture_coding_type    */
			{
				case kIPicture:         /*      I and P pictures only   */
				case kPPicture:

					bufY = data->secondPredictY;
					bufU = data->secondPredictU;
					bufV = data->secondPredictV;

					data->secondPredictY = data->secondPredictYR;
					data->secondPredictU = data->secondPredictUR;
					data->secondPredictV = data->secondPredictVR;

					data->secondPredictYR = bufY;
					data->secondPredictUR = bufU;
					data->secondPredictVR = bufV;

					break;
			}

			if (index2 == 0)
			{
			    /* I,P-picture */

			    if (index1 % parameter->N == 0)
			    {
				picture_coding_type = kIPicture;
				GOPnumber++;
			    }
			    else
			    {
				picture_coding_type = kPPicture;
			    }
			}
			else
			{
				picture_coding_type = kBPicture;
			}

			/*      determine buffer location for picture storage   */

			switch (picture_coding_type)
			{
				case kIPicture:
				case kPPicture:
					
					/* source buffer */
					bufY = data->firstPredictY;
					bufU = data->firstPredictU;
					bufV = data->firstPredictV;

					/* rotate */
					data->firstPredictY = data->secondPredictY;
					data->firstPredictU = data->secondPredictU;
					data->firstPredictV = data->secondPredictV;

					data->secondPredictY = bufY;
					data->secondPredictU = bufU;
					data->secondPredictV = bufV;
					
					/* reconstruction buffer */
					bufYR = data->firstPredictYR;
					bufUR = data->firstPredictUR;
					bufVR = data->firstPredictVR;

					data->firstPredictYR = data->secondPredictYR;
					data->firstPredictUR = data->secondPredictUR;
					data->firstPredictVR = data->secondPredictVR;

					data->secondPredictYR = bufYR;
					data->secondPredictUR = bufUR;
					data->secondPredictVR = bufVR;

					break;

				default:

					bufY = data->predictY;
					bufU = data->predictU;
					bufV = data->predictV;

					bufYR = data->predictYR;
					bufUR = data->predictUR;
					bufVR = data->predictVR;

					break;
			}
			/* read source picture file into source buffer */
			
			/* update temporal reference */                        
			switch (picture_coding_type){
			    case kIPicture:
				temporal_reference_IP = parameter->M - 1;
				parameter->temporal_reference = temporal_reference_IP;
				break;
			    case kPPicture:
				temporal_reference_IP += parameter->M;
				parameter->temporal_reference = temporal_reference_IP;
				break;
			    case kBPicture:
				parameter->temporal_reference = temporal_reference_IP + index2 - parameter->M;
				break;
			}
				
			if (ReadPicture(parameter->pictureNum, bufY, bufU, bufV, parameter))
			{
				writeBits(0, 0, -1);    /*      flush buffer    */
				exit (-1);
			}

			/*      process picture */

			/*      get motion vectors      */

			switch (picture_coding_type)
			{
				case kPPicture:

					/*      I/P to P - picture distance = 3 */

					if (parameter->halfPel)
						sprintf (fName, "MVFH%03d.%d", parameter->pictureNum,
								parameter->searchDistance);
					else
						sprintf (fName, "MVF%03d.%d", parameter->pictureNum,
								parameter->searchDistance);

					GetMotionVector(data->firstPredictY, data->firstPredictYR,
							bufY, forwardMV, parameter->M * parameter->searchDistance,
							parameter->halfPel, parameter->fileMV, fName, parameter);

					break;

				case kBPicture:

					/*      I/P to B - picture distance = index2    */

					if (parameter->halfPel)
						sprintf (fName, "MVFH%03d.%d", parameter->pictureNum,
								parameter->searchDistance);
					else
						sprintf (fName, "MVF%03d.%d", parameter->pictureNum,
								parameter->searchDistance);

					GetMotionVector(data->firstPredictY, data->firstPredictYR,
							bufY, forwardMV, 
							index2 * parameter->searchDistance,
							parameter->halfPel, parameter->fileMV, fName, parameter);

					/*      B to I/P - picture distance = M - index2     */

					if (parameter->halfPel)
						sprintf (fName, "MVBH%03d.%d", parameter->pictureNum,
								parameter->searchDistance);
					else
						sprintf (fName, "MVB%03d.%d", parameter->pictureNum,
								parameter->searchDistance);

					GetMotionVector(data->secondPredictY, data->secondPredictYR,
							bufY, backwardMV, 
							(parameter->M - index2) * parameter->searchDistance,
							parameter->halfPel, parameter->fileMV, fName, parameter);
					break;

			}

			/*      process MBs     */

			ProcPicture(data, parameter, rateControl,
					bufY, bufU, bufV, bufYR, bufUR, bufVR,
					picture_coding_type, index2, forwardMV, backwardMV,
					macroblock_type, coded_block_pattern, MQuant);
		}
	}

	writeBits(cSequenceEndCode, bSequenceEndCode, 1);       /* sequence_end_code */

	writeBits(0, 0, -1);    /*      flush buffer    */

	exit(0);
}
