/* File: quantize.c */

/*
 *      $Log:   quant.h,v $
 * Revision 1.3      93/08/20  21:15:53  hana
 * MPEG1 encoder minor revision
 * 
 * Revision 1.1.1.1  93/03/29  11:27:53  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:32:40  au
 * Initial revision
 *
 */

/*
 *      Default intra quantizer matrix          
 *      On changing any part of this,           
 *       you should make
 *         cLoadIntraIntraQuantierMatrix = 0
 *       in flc.h .
 */

#include <stdio.h>
#include "consts.h"
#include "mpeg1.h"
#include "encoder.pro"


BYTE    tQuantIntra[64] =
{
	 8, 16, 16, 19, 16, 19, 22, 22,
	22, 22, 22, 22, 26, 24, 26, 27,
	27, 27, 26, 26, 26, 26, 27, 27,
	27, 29, 29, 29, 34, 34, 34, 29,
	29, 29, 27, 27, 29, 29, 32, 32,
	34, 34, 37, 38, 37, 35, 35, 34,
	35, 38, 38, 40, 40, 40, 48, 48,
	46, 46, 56, 56, 58, 69, 69, 83
};


BYTE    tQuant[64] =
{
	16, 16, 16, 16, 16, 16, 16, 16,
	16, 16, 16, 16, 16, 16, 16, 16,
	16, 16, 16, 16, 16, 16, 16, 16,
	16, 16, 16, 16, 16, 16, 16, 16,
	16, 16, 16, 16, 16, 16, 16, 16,
	16, 16, 16, 16, 16, 16, 16, 16,
	16, 16, 16, 16, 16, 16, 16, 16,
	16, 16, 16, 16, 16, 16, 16, 16
};

/* Quantize intra macroblock */
int     QuantizeI(
	short   *codedBPattern,
	int     *currentMBDiff,
	int     *currentMBDiffU,
	int     *currentMBDiffV,
	int     MQuant)
{
	extern BYTE     tQuantIntra[],
			tQuant[];

	short   line, pix, row, col, quantIndex;
	int     currentIndex;
	float   tFloat;
	int     tInt;


	/*      lumin   */

	for (line = 0; line < kMBHeight; line += kDCTSize)
	{
		for (pix = 0; pix < kMBWidth; pix += kDCTSize)
		{
			currentIndex = line * kMBWidth + pix;

			tFloat = currentMBDiff[currentIndex] / 8.0;
			currentMBDiff[currentIndex] =
					(tFloat >= 0.0) ? tFloat + 0.5 : tFloat - 0.5;

			/* clip and saturate */
			if (currentMBDiff[currentIndex] > 255)
				currentMBDiff[currentIndex] = 255;
			else if (currentMBDiff[currentIndex] < -255)
				currentMBDiff[currentIndex] = -255;

			/*      exclude dc coeff        */

			quantIndex = 1;
			col = 1;
			currentIndex++;

			for (row = 0; row < kDCTSize; row++, currentIndex += kMBWidthDCT)
			{
				for ( ; col < kDCTSize; col++, currentIndex++, quantIndex++)
				{
					tFloat = 8.0 * currentMBDiff[currentIndex] /
							(float)(MQuant * tQuantIntra[quantIndex]);
					tInt = (tFloat >= 0.0) ? tFloat + 0.5 : tFloat - 0.5;

					/* clip and saturate */
					if (tInt >= 255)
						currentMBDiff[currentIndex] = 255;
					else if (tInt <= -255)
						currentMBDiff[currentIndex] = -255;
					else
						currentMBDiff[currentIndex] = tInt;
				}

				col = 0;
			}
		}
	}

	/*      chrom   */

	tFloat = currentMBDiffU[0] / 8.0;
	currentMBDiffU[0] = (tFloat >= 0.0) ? tFloat + 0.5 : tFloat - 0.5;

	if (currentMBDiffU[0] > 255)
		currentMBDiffU[0] = 255;
	else if (currentMBDiffU[0] < -255)
		currentMBDiffU[0] = -255;

	currentIndex = 1;

	for (row = 1; row < kDCTLen; row++, currentIndex++)
	{
		/*      U       */

		tFloat = 8.0 * currentMBDiffU[currentIndex] /
				(float)(MQuant * tQuantIntra[currentIndex]);
		tInt = (tFloat > 0) ? tFloat + 0.5 : tFloat - 0.5;

		if (tInt >= 255)
			currentMBDiffU[currentIndex] = 255;
		else if (tInt <= -255)
			currentMBDiffU[currentIndex] = -255;
		else
			currentMBDiffU[currentIndex] = tInt;
	}

	tFloat = currentMBDiffV[0] / 8.0;
	currentMBDiffV[0] = (tFloat >= 0.0) ? tFloat + 0.5 : tFloat - 0.5;

	if (currentMBDiffV[0] > 255)
		currentMBDiffV[0] = 255;
	else if (currentMBDiffV[0] < -255)
		currentMBDiffV[0] = -255;

	currentIndex = 1;

	for (row = 1; row < kDCTLen; row++, currentIndex++)
	{
		/*      V       */

		tFloat = 8.0 * currentMBDiffV[currentIndex] /
				(float)(MQuant * tQuantIntra[currentIndex]);
		tInt = (tFloat > 0) ? tFloat + 0.5 : tFloat - 0.5;

		if (tInt >= 255)
			currentMBDiffV[currentIndex] = 255;
		else if (tInt <= -255)
			currentMBDiffV[currentIndex] = -255;
		else
			currentMBDiffV[currentIndex] = tInt;
	}

	*codedBPattern = 0x3F;          /*      all blocks coded        */

	return(kCodedType);
}



int     Quantize(
	short   *codedBPattern,
	int     *currentMBDiff,
	int     *currentMBDiffU,
	int     *currentMBDiffV,
	int     MQuant)
{
	extern BYTE     tQuantIntra[], tQuant[];

	short           code, line, pix, row, col, quantIndex;
	int             currentIndex;
	float           tFloat, tFloat2;
	int             tInt, tInt2;

	/*      lumin   */

	*codedBPattern = 0;

	code = 0x20;

	for (line = 0; line < kMBHeight; line += kDCTSize)
	{
		for (pix = 0; pix < kMBWidth; pix += kDCTSize)
		{
			currentIndex = line * kMBWidth + pix;

			quantIndex = 0;

			for (row = 0; row < kDCTSize; row++, currentIndex += kMBWidthDCT)
			{
				for (col = 0; col < kDCTSize; col++,
						currentIndex++, quantIndex++)
				{
					tFloat = 16.0 * currentMBDiff[currentIndex] /
							(float) tQuant[quantIndex];
					tInt = (tFloat >= 0.0) ? tFloat + 0.5 : tFloat - 0.5;

					if (MQuant % 2) /*      odd             */
						tInt2 = tInt / (2 * MQuant);
					else                            /*      even    */
					{
						if (tInt < 0)
							tInt2 = (tInt + 1) / (2 * MQuant);
						else    /*      > 0, case == 0 is truncated to 0 anyway */
							tInt2 = (tInt - 1) / (2 * MQuant);
					}

					if (tInt2 >= 255)
						currentMBDiff[currentIndex] = 255;
					else if (tInt2 <= -255)
						currentMBDiff[currentIndex] = -255;
					else
						currentMBDiff[currentIndex] = tInt2;

					if (currentMBDiff[currentIndex])
						*codedBPattern |= code;         /*      non-zero coeff  */
				}
			}

			code >>= 1;
		}
	}

	/*      chrom   */

	currentIndex = 0;

	for (row = 0; row < kDCTLen; row++, currentIndex++)
	{
		/*      U       */

		tFloat = 16.0 * currentMBDiffU[currentIndex] /
				(float) tQuant[currentIndex];
		tInt = (tFloat >= 0) ? tFloat + 0.5 : tFloat - 0.5;

		if (MQuant % 2) /*      odd             */
			tInt2 = tInt / (2 * MQuant);
		else                            /*      even    */
		{
			if (tInt < 0)
				tInt2 = (tInt + 1) / (2 * MQuant);
			else
				tInt2 = (tInt - 1) / (2 * MQuant);
		}

		if (tInt2 >= 255)
			currentMBDiffU[currentIndex] = 255;
		else if (tInt2 <= -255)
			currentMBDiffU[currentIndex] = -255;
		else
			currentMBDiffU[currentIndex] = tInt2;

		if (currentMBDiffU[currentIndex])
			*codedBPattern |= 0x02;         /*      non-zero coeff  */
	}

	currentIndex = 0;

	for (row = 0; row < kDCTLen; row++, currentIndex++)
	{
		/*      V       */

		tFloat = 16.0 * currentMBDiffV[currentIndex] /
				(float) tQuant[currentIndex];
		tInt = (tFloat >= 0) ? tFloat + 0.5 : tFloat - 0.5;

		if (MQuant % 2) /*      odd             */
			tInt2 = tInt / (2 * MQuant);
		else                            /*      even    */
		{
			if (tInt < 0)
				tInt2 = (tInt + 1) / (2 * MQuant);
			else
				tInt2 = (tInt - 1) / (2 * MQuant);
		}

		if (tInt2 >= 255)
			currentMBDiffV[currentIndex] = 255;
		else if (tInt2 <= -255)
			currentMBDiffV[currentIndex] = -255;
		else
			currentMBDiffV[currentIndex] = tInt2;

		if (currentMBDiffV[currentIndex])
			*codedBPattern |= 0x01;         /*      non-zero coeff  */
	}

	if (*codedBPattern)
		return(kCodedType);
	else
		return(0);
}


/* Inverse quantize intra macorblock */
void    IQuantizeI(
	short   codedBPattern,
	int     *currentMBDiff,
	int     *currentMBDiffU,
	int     *currentMBDiffV,
	int     MQuant)
{
	short   code, line, pix, row, col, quantIndex;
	int     currentIndex;
	int     tInt;

	/*      lumin   */

	for (line = 0; line < kMBHeight; line += kDCTSize)
	{
		for (pix = 0; pix < kMBWidth; pix += kDCTSize)
		{
			currentIndex = line * kMBWidth + pix;

			currentMBDiff[currentIndex] *= 8;

			if (currentMBDiff[currentIndex] < -2048)
				currentMBDiff[currentIndex] = -2048;
			else if (currentMBDiff[currentIndex] > 2047)
				currentMBDiff[currentIndex] = 2047;

			/*      exclude dc coeff        */

			currentIndex++;

			for (row = 0, col = 1, quantIndex = 1; row < kDCTSize;
					row++, col = 0, currentIndex += kMBWidthDCT)
			{
				for ( ; col < kDCTSize; col++, currentIndex++, quantIndex++)
				{
					if (currentMBDiff[currentIndex] != 0)
					{
						/*      rec = MQuant * 2 * QAC * WI / 16        */

						tInt = MQuant * currentMBDiff[currentIndex] *
								(int) tQuantIntra[quantIndex] / 8;

						if (! (tInt % 2))               /*      even    */
						{
							if (tInt < 0)
								tInt += 1;
							else if (tInt > 0)
								tInt -= 1;
						}

						if (tInt <= -2048)
							currentMBDiff[currentIndex] = -2048;
						else if (tInt >= 2047)
							currentMBDiff[currentIndex] = 2047;
						else
							currentMBDiff[currentIndex] = tInt;
					}
				}
			}
		}
	}

	/*      chrom   */

	currentMBDiffU[0] *= 8;

	if (currentMBDiffU[0] < -2048)
		currentMBDiffU[0] = -2048;
	else if (currentMBDiffU[0] > 2047)
		currentMBDiffU[0] = 2047;

	currentIndex = 1;

	for (row = 1; row < kDCTLen; row++, currentIndex++)
	{
		/*      U       */

		if (currentMBDiffU[currentIndex] != 0)
		{
			/*      rec = MQuant * 2 * QAC * WI / 16        */

			tInt = MQuant * currentMBDiffU[currentIndex] *
					(int) tQuantIntra[currentIndex] / 8;

			if (! (tInt % 2))               /*      even    */
			{
				if (tInt < 0)
					tInt += 1;
				else if (tInt > 0)
					tInt -= 1;
			}

			if (tInt <= -2048)
				currentMBDiffU[currentIndex] = -2048;
			else if (tInt >= 2047)
				currentMBDiffU[currentIndex] = 2047;
			else
				currentMBDiffU[currentIndex] = tInt;
		}
	}

	currentMBDiffV[0] *= 8;

	if (currentMBDiffV[0] < -2048)
		currentMBDiffV[0] = -2048;
	else if (currentMBDiffV[0] > 2047)
		currentMBDiffV[0] = 2047;

	currentIndex = 1;

	for (row = 1; row < kDCTLen; row++, currentIndex++)
	{
		/*      V       */

		if (currentMBDiffV[currentIndex] != 0)
		{
			/*      rec = MQuant * 2 * QAC * WI / 16        */

			tInt = MQuant * currentMBDiffV[currentIndex] *
					(int) tQuantIntra[currentIndex] / 8;

			if (! (tInt % 2))               /*      even    */
			{
				if (tInt < 0)
					tInt += 1;
				else if (tInt > 0)
					tInt -= 1;
			}

			if (tInt <= -2048)
				currentMBDiffV[currentIndex] = -2048;
			else if (tInt >= 2047)
				currentMBDiffV[currentIndex] = 2047;
			else
				currentMBDiffV[currentIndex] = tInt;
		}
	}
}

/* Inverse quantize non-intra macroblock */

void    IQuantize(
	short   codedBPattern,
	int     *currentMBDiff,
	int     *currentMBDiffU,
	int     *currentMBDiffV,
	int     MQuant)
{
	short   code, line, pix, row, col, quantIndex, offset1, offset2;
	int     currentIndex;
	float   tFloat, tFloat2;
	int     tInt, tInt2;

	/*      lumin   */

	code = 0x20;

	for (line = 0; line < kMBHeight; line += kDCTSize)
	{
		for (pix = 0; pix < kMBWidth; pix += kDCTSize)
		{
			if (! (codedBPattern & code))
			{
				code >>= 1;
				continue;
			}

			code >>= 1;

			currentIndex = line * kMBWidth + pix;

			for (row = 0, quantIndex = 0; row < kDCTSize; row++,
					currentIndex += kMBWidthDCT)
			{
				for (col = 0; col < kDCTSize; col++,
						currentIndex++, quantIndex++)
				{
					if (currentMBDiff[currentIndex])
					{
						if (currentMBDiff[currentIndex] > 0)
							tInt = (2 * currentMBDiff[currentIndex] + 1) *
									MQuant * (int) tQuant[quantIndex] / 16;
						else    /*      < 0             */
							tInt = (2 * currentMBDiff[currentIndex] - 1) *
									MQuant * (int) tQuant[quantIndex] / 16;

						if (! (tInt % 2))               /*      even    */
						{
							if (tInt < 0)
								tInt += 1;
							else if (tInt > 0)
								tInt -= 1;
						}

						if (tInt <= -2048)
							currentMBDiff[currentIndex] = -2048;
						else if (tInt >= 2047)
							currentMBDiff[currentIndex] = 2047;
						else
							currentMBDiff[currentIndex] = tInt;
					}
				}
			}
		}
	}

	/*      chrom   */

	if (codedBPattern & 0x2)
	{
		/*      U       */

		for (row = 0, currentIndex = 0; row < kDCTLen; row++, currentIndex++)
		{
			if (currentMBDiffU[currentIndex])
			{
				if (currentMBDiffU[currentIndex] > 0)
					tInt = (2 * currentMBDiffU[currentIndex] + 1) *
							MQuant * (int) tQuant[currentIndex] / 16;
				else
					tInt = (2 * currentMBDiffU[currentIndex] - 1) *
							MQuant * (int) tQuant[currentIndex] / 16;

				if (! (tInt % 2))               /*      even    */
				{
					if (tInt < 0)
						tInt += 1;
					else if (tInt > 0)
						tInt -= 1;
				}

				if (tInt <= -2048)
					currentMBDiffU[currentIndex] = -2048;
				else if (tInt >= 2047)
					currentMBDiffU[currentIndex] = 2047;
				else
					currentMBDiffU[currentIndex] = tInt;
			}
		}
	}

	if (codedBPattern & 0x1)
	{
		/*      V       */

		for (row = 0, currentIndex = 0; row < kDCTLen; row++, currentIndex++)
		{
			if (currentMBDiffV[currentIndex])
			{
				if (currentMBDiffV[currentIndex] > 0)
					tInt = (2 * currentMBDiffV[currentIndex] + 1) *
							MQuant * (int) tQuant[currentIndex] / 16;
				else
					tInt = (2 * currentMBDiffV[currentIndex] - 1) *
							MQuant * (int) tQuant[currentIndex] / 16;

				if (! (tInt % 2))               /*      even    */
				{
					if (tInt < 0)
						tInt += 1;
					else if (tInt > 0)
						tInt -= 1;
				}

				if (tInt <= -2048)
					currentMBDiffV[currentIndex] = -2048;
				else if (tInt >= 2047)
					currentMBDiffV[currentIndex] = 2047;
				else
					currentMBDiffV[currentIndex] = tInt;
			}
		}
	}
}
