/* File: PreFrame.c */

/* note to myself:  this file needs to be changed to a name that expresses
the essense of what preframe, postframe, macroblock, etc. do:

e.g.:   analysis.c,   step3.c,  etc.

*/

/*
 *      $Log:   PreFrame.c,v $
 * Revision 1.4  93/09/03  01:15:00  hana
 * VBV buffer operation 
 * 
 * Revision 1.1.1.1  93/03/29  11:28:00  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:30:43  au
 * Initial revision
 *
 */

#include <stdio.h>
#include "consts.h"
#include "mpeg1.h"
#include "flc.h"
#include "encoder.pro"

/* move this into rateControl later */
#define Kp      1.0
#define Kb      1.4

/* Perform pre-frame analysis as per TM Step 1 */
void    PreFrame(
	short           picture_coding_type,
	tParameter      *parameter,
	tRateControl    *rateControl,
	short           forward_f_code,
	short           backward_f_code)
{
	static int firstIPicture = TRUE;
  
	rateControl->totalPBits = 0;

fprintf(stderr,"Now beginning pre-frame analysis\n");
	switch (picture_coding_type)
	{
		case kBPicture1:                /*      B frame */
		case kBPicture2:

			rateControl->Sb = 0;
			rateControl->Qb = 0.0;
			rateControl->djb = rateControl->d0b;

			rateControl->Tb = rateControl->R / (rateControl->Nb +
					(rateControl->Np * Kb * rateControl->Xp /
					(Kp * rateControl->Xb)));
			if (rateControl->Tb < rateControl->Tipb)
				rateControl->Tb = rateControl->Tipb;
			
			/* calculate vbv_delay */
			rateControl->vbv_delay
				= calcVbvDelay(picture_coding_type, rateControl, parameter);
			
			/*      bit count - picture     */

			rateControl->djb += Write_Picture_Header(picture_coding_type, rateControl,
					forward_f_code, backward_f_code, 
						parameter->temporal_reference);

			rateControl->Tbj = (rateControl->Tb - rateControl->Sb) /
					parameter->numMacroBlocks;

			break;

		case kPPicture:         /*      P frame */

			rateControl->Sp = 0;
			rateControl->Qp = 0.0;
			rateControl->djp = rateControl->d0p;

			rateControl->Tp = rateControl->R / (rateControl->Np +
					(rateControl->Nb * Kp * rateControl->Xb /
					(Kb * rateControl->Xp)));
			if (rateControl->Tp < rateControl->Tipb)
				rateControl->Tp = rateControl->Tipb;

			/* calculate vbv_delay */
			rateControl->vbv_delay
				= calcVbvDelay(picture_coding_type, rateControl, parameter);

			/*      bit count - picture     */

			rateControl->djp += Write_Picture_Header(picture_coding_type, rateControl,
					forward_f_code, backward_f_code, 
						parameter->temporal_reference);

			rateControl->Tpj = (rateControl->Tp - rateControl->Sp) /
					parameter->numMacroBlocks;

			break;

		case kIPicture:         /*      I frame */

			if(firstIPicture) {
				rateControl->Si = rateControl->headerBits;
				firstIPicture = FALSE;
			}
			else 
				rateControl->Si = 0;
			rateControl->Qi = 0.0;
			rateControl->dji = rateControl->d0i;

fprintf(stderr, "Now calcuating rate control parameters for I picture\n");

			rateControl->R += rateControl->G;
			rateControl->Np = parameter->N / parameter->M - 1;
			rateControl->Nb = parameter->N - rateControl->Np - 1;

fprintf(stderr, "Np = %d, Nb = %d\n", rateControl->Np, rateControl->Nb);
			rateControl->Ti = rateControl->R / (1 + (rateControl->Np *
					rateControl->Xp / (Kp * rateControl->Xi)) +
					(rateControl->Nb * rateControl->Xb / (Kb *
					rateControl->Xi)));
			if (rateControl->Ti < rateControl->Tipb)
				rateControl->Ti = rateControl->Tipb;

			/*      bit count - group of picture    */

fprintf(stderr, "Now calling GOP_Header function\n");

			rateControl->dji += Write_GOP_Header(parameter,rateControl);   

			/* calculate vbv_delay */
			rateControl->vbv_delay
				= calcVbvDelay(picture_coding_type, rateControl, parameter);
			/*      bit count - picture     */

			rateControl->dji += Write_Picture_Header(picture_coding_type, rateControl,
					forward_f_code, backward_f_code, parameter->temporal_reference);

			rateControl->Tij = (rateControl->Ti - rateControl->Si) /
					parameter->numMacroBlocks;

			break;
	}

	rateControl->actjSum  = 0.0;
}


void    PostFrame(
	short           picture_coding_type,
	tRateControl    *rateControl,
	tParameter      *parameter)
{
	switch (picture_coding_type)
	{
		case kBPicture1:        /*      B frame         */
		case kBPicture2:

			rateControl->Qb /= parameter->numMacroBlocks;
			rateControl->Xb = rateControl->Sb * rateControl->Qb;
			rateControl->R -= rateControl->Sb;
			rateControl->d0b = rateControl->djb;
			rateControl->Nb--;
			break;

		case kPPicture:         /*      P frame         */

			rateControl->Qp /= parameter->numMacroBlocks;
			rateControl->Xp = rateControl->Sp * rateControl->Qp;
			rateControl->R -= rateControl->Sp;
			rateControl->d0p = rateControl->djp;
			rateControl->Np--;
			break;
		
		case kIPicture:         /*      I frame         */

			rateControl->Qi /= parameter->numMacroBlocks;
			rateControl->Xi = rateControl->Si * rateControl->Qi;
			rateControl->R -= rateControl->Si;
			rateControl->d0i = rateControl->dji;
			break;
	}

	rateControl->avg_act = rateControl->actjSum / parameter->numMacroBlocks;
}


void    DoPreMBRC(
	short           picture_coding_type,
	tRateControl    *rateControl,
	BYTE            *currentMB,
	int             *MQuant)
{
	short   line, pix, row, col, currentIndex, currentIndex1;
	float   sum, P_mean, var_sblk, min_var_sblk, actj, N_actj;

	/*      calc actj       */

	min_var_sblk = 1.0e10;  /*      initialize to something large   */

	/*      frame   */

	for (line = 0; line < kMBHeight; line += kDCTSize)
	{
		for (pix = 0; pix < kMBWidth; pix += kDCTSize)
		{
			currentIndex1 = line * kMBWidth + pix;

			/*      P_mean  */

			sum = 0.0;

			for (row = 0, currentIndex = currentIndex1; row < kDCTSize;
					row++, currentIndex += kMBWidthDCT)
			{
				for (col = 0; col < kDCTSize; col++, currentIndex++)
				{
					sum += currentMB[currentIndex];
				}
			}

			P_mean = sum / kDCTLen;

			/*      var_sblk        */

			sum = 0.0;

			for (row = 0, currentIndex = currentIndex1; row < kDCTSize;
					row++, currentIndex += kMBWidthDCT)
			{
				for (col = 0; col < kDCTSize; col++, currentIndex++)
				{
					var_sblk = ((float) currentMB[currentIndex] - P_mean);
					sum += var_sblk * var_sblk;
				}
			}

			var_sblk = sum / kDCTLen;

			if (var_sblk < min_var_sblk)
				min_var_sblk = var_sblk;
		}
	}

	actj = 1.0 + min_var_sblk;

	rateControl->actjSum += actj;

	N_actj = (2.0 * actj + rateControl->avg_act) /
			(actj + 2.0 * rateControl->avg_act);

	switch (picture_coding_type)
	{
		case kBPicture1:
		case kBPicture2:

			rateControl->Qj = rateControl->djb * 31.0 / rateControl->r;
			break;

		case kPPicture:

			rateControl->Qj = rateControl->djp * 31.0 / rateControl->r;
			break;

		case kIPicture:

			rateControl->Qj = rateControl->dji * 31.0 / rateControl->r;
			break;
	}
	
	/* step 3 ? */
	*MQuant = rateControl->Qj * N_actj + 0.5;

	if (*MQuant < 1)
		*MQuant = 1;
	else if (*MQuant > 31)
		*MQuant = 31;

}


int     calcVbvDelay(
	short           picture_coding_type,
	tRateControl    *rateControl,
	tParameter      *parameter)
{
	int     delay;
	float   bufferAst;

fprintf(stderr,"Now calculating vbvdelay\n");

	bufferAst = rateControl->vbvOccupy + (float)parameter->bit_rate /
		parameter->float_picture_rate;

	switch(picture_coding_type) { /* subtract bits for picture header */
		case kBPicture1:                /*      B frame */
		case kBPicture2:
			bufferAst -= (rateControl->Sb + bPictureStartCode);
			break;
		case kPPicture:         /*      P frame */
			bufferAst -= (rateControl->Sp + bPictureStartCode);
			break;
		case kIPicture:                 /*      I frame */
			bufferAst -= (rateControl->Si + bPictureStartCode);
			break;
	}

	delay = (int)(90000 * bufferAst / parameter->bit_rate);
	
fprintf(stderr,"vbvdelay = %d \n", delay);

	return(delay);
}


void    checkUpdateVbv(
	short           picture_coding_type,
	int             pictureNum,     
	tRateControl    *rateControl,
	tParameter      *parameter)
{
	float   bitsThisFrame,
		bitRatePerFrame;        

	switch(picture_coding_type) {
		case kBPicture1:                /*      B frame */
		case kBPicture2:
			bitsThisFrame = rateControl->Sb;
			break;
		case kPPicture:         /*      P frame */
			bitsThisFrame = rateControl->Sp;
			break;
		case kIPicture:                 /*      I frame */
			bitsThisFrame = rateControl->Si;
			break;
	}

	bitRatePerFrame = (float)parameter->bit_rate / parameter->float_picture_rate;

	/* check if d_n+1 > B_n + ( 2 R/P ) - B */
	if( bitsThisFrame <=
	   rateControl->vbvOccupy + 2.0 * bitRatePerFrame - parameter->vbv_buffer_size)
		fprintf(stderr, "VBV Buffer Over Flow : picNum = %d, coded bits = %d\n",
				pictureNum, (int)bitsThisFrame);
	/* check if d_n+1 <= B_n + ( R/P ) */
	else if(bitsThisFrame > rateControl->vbvOccupy + bitRatePerFrame)
		fprintf(stderr, "VBV Buffer Under Flow : picNum = %d, coded bits = %d\n",
				pictureNum, (int)bitsThisFrame);

	/* update vbv_buffer */
	rateControl->vbvOccupy += bitRatePerFrame;
	rateControl->vbvOccupy -= bitsThisFrame;

	return;
}
