/* File: ReadPict.c */

/*
 *      $Log:   ReadPict.c,v $
 * Revision 1.1.1.1  93/03/29  11:27:54  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:33:17  au
 * Initial revision
 *
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "consts.h"
#include "mpeg1.h"
#include "encoder.pro"

/* need to add .tga, port. bitmap, and .yuv */

short   ReadPicture(
	int             pictureNumber,
	BYTE            *y_buffer,
	BYTE            *cb_buffer,
	BYTE            *cr_buffer,
	tParameter      *parameter)
{
	int ret;

	switch(parameter->sourceFileFormat)
	{
	case 0:
		ret = Read_SIF(pictureNumber, y_buffer, cb_buffer, cr_buffer, parameter);
		break;
	case 1:
		ret = Read_TGA(pictureNumber, y_buffer, cb_buffer, cr_buffer, parameter);
		break;
	case 2:                
		ret = Read_PPM(pictureNumber, y_buffer, cb_buffer, cr_buffer, parameter);
		break;
	case 3:
		ret = Read_YUV(pictureNumber, y_buffer, cb_buffer, cr_buffer, parameter);
		break;

	}
	return (ret);
}

short   Read_SIF(
	int             pictureNumber,
	BYTE            *y_buffer,
	BYTE            *cb_buffer,
	BYTE            *cr_buffer,
	tParameter      *parameter)
{
	int ret;
	int     i, j;

	FILE    *fptr;
	char    fname[kMaxCharBufLen];
	static BYTE *buffer;
	BYTE    *b;
	BYTE    *y;
	BYTE    *u;
	BYTE    *v;

	if (!buffer)
	{
	  buffer = (BYTE *)malloc(parameter->luminWidth*2);
	}
	sprintf (fname, "%s%03d.SIF", parameter->inputSequence, pictureNumber);

	if ((fptr = fopen(fname, "rb")) == NULL)
	{
		printf("\nopen ERROR on filename: %s", fname);
		return (1);
	}

	y = y_buffer;
	u = cb_buffer;
	v = cr_buffer;

	for (i = 0; i < parameter->luminHeight; i++)
	{
		fread(buffer, sizeof(BYTE), parameter->luminWidth*2, fptr);

		for (j = 0, b = buffer; j < parameter->luminWidth/2; j++)
		{
			*(u++) = *(b++);
			*(y++) = *(b++);
			*(v++) = *(b++);
			*(y++) = *(b++);
		}
	}

	fclose(fptr);

	/* decimate chrominance in vertical direction */
	u = cb_buffer;
	v = cr_buffer;

	
	convert422to420(1, u, parameter->chromWidth, 
	     parameter->chromHeight*2, FORWARD, parameter);
	convert422to420(1, v, parameter->chromWidth, 
	     parameter->chromHeight*2, FORWARD, parameter);

	return (0);
}

short   Read_TGA(
	int             pictureNumber,
	BYTE            *y_buffer,
	BYTE            *cb_buffer,
	BYTE            *cr_buffer,
	tParameter      *parameter)
{

fprintf(stderr,"Sorry -- TGA file format has not been implemented yet.");

exit(-1);
}


short   Read_PPM(
	int             pictureNumber,
	BYTE            *y_buffer,
	BYTE            *cb_buffer,
	BYTE            *cr_buffer,
	tParameter      *parameter)
{


fprintf(stderr,"Sorry -- PPM file format has not been implemented yet.");
exit(-1);
}

/* Read Stanford-compatible raw 4:2:0 YUV file format */
short   Read_YUV(
	int             pictureNumber,
	BYTE            *y_buffer,
	BYTE            *cb_buffer,
	BYTE            *cr_buffer,
	tParameter      *parameter)
{
	int ret;
	int     i, j, col, line;

	FILE    *fptr;
	char    fname[kMaxCharBufLen];
	static  BYTE *buffer;
	BYTE    *b;
	BYTE    *y;
	BYTE    *u;
	BYTE    *v;

        int lw = parameter->luminWidth;
        int lh = parameter->luminHeight;
        int cw = parameter->chromWidth;
        int ch = parameter->chromHeight;
    
fprintf(stderr,"Reading YUV input picture\n");

	if (!buffer)
	{
	  buffer = (BYTE *)malloc(parameter->luminWidth);
	}

fprintf(stderr, "Malloced buffer\n");

	sprintf (fname, "%s%03d.yuv", parameter->inputSequence, pictureNumber);

	if ((fptr = fopen(fname, "rb")) == NULL)
	{
		fprintf(stderr,"\nopen ERROR on filename: %s", fname);
		return (1);
	}

fprintf(stderr, "Opened file\n");

	y = y_buffer;
	
	for (line = 0; line < lh; line++)
	{
		fread(buffer, sizeof(BYTE), lw, fptr);

		for (col = 0, b = buffer; col < lw; col++)
		{
                        *(y++) = *(b++);
		}
	}
       
	u = cb_buffer;


	for (line = 0; line < ch; line++)
	{
		fread(buffer, sizeof(BYTE), cw, fptr);

		for (col = 0, b = buffer; col < cw; col++)
		{
                        *(u++) = *(b++);
		}
	}

	v = cr_buffer;

	for (line = 0; line < ch; line++)
	{
		fread(buffer, sizeof(BYTE), cw, fptr);

		for (col = 0, b = buffer; col < cw; col++)
		{
                        *(v++) = *(b++);
		}
	}

	fclose(fptr);
	return (0);
}


