/*
 *      $Log:   reconstr.c,v $
 * Revision 1.1.1.1  93/03/29  11:27:54  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:33:17  au
 * Initial revision
 *
 */

#include <stdio.h>
#include "consts.h"
#include "mpeg1.h"
#include "encoder.pro"

void    FrameConstr(
	BYTE    *predictY,
	BYTE    *predictU,
	BYTE    *predictV,
	BYTE    *constY,
	BYTE    *constU,
	BYTE    *constV,
	short   MV[],
	BYTE    *current,
	BYTE    *currentU,
	BYTE    *currentV,
	int     *diff,
	int     *diffU,
	int     *diffV,
	tParameter      *parameter)

{
	int     offset, mvx, mvy, index1, index2;
	BYTE    *constY2, *constU2, *constV2;
	BYTE    *frameIndex, *frameIndexU, *frameIndexV;
	short   mv[2];
	BYTE    *currentIndex, *currentIndexU, *currentIndexV;
	int     *diffIndex, *diffIndexU, *diffIndexV;

	short   f0off2, f0off3, f0off4;

	/*      reconstruct MB  */

	constY2 = constY;

	/*      lumin   */

	mvx = MV[0] / 2;
	mvy = MV[1] / 2;
	offset = (mvy * parameter->luminWidth) + mvx;
	frameIndex = (BYTE *) (predictY + offset);

	GetLoc(MV, &f0off2, &f0off3, &f0off4, parameter->luminWidth);

	currentIndex = current;
	diffIndex = diff;

	for (index1 = 0; index1 < kMBHeight; index1++,
			constY2 += parameter->luminWidthMB, 
				frameIndex += parameter->luminWidthMB)
	{
		for (index2 = 0; index2 < kMBWidth; index2++, constY2++, 
			frameIndex++, currentIndex++, diffIndex++)
		{
			/*      add 2 for integer rounding      */

			*constY2 = ((int) *frameIndex  +
					(int) *(frameIndex + f0off2) +
					(int) *(frameIndex + f0off3) +
					(int) *(frameIndex + f0off4) + 2) / 4;

			*diffIndex = *currentIndex - *constY2;

		}
	}

	/*      chrom   */

	mv[0] = MV[0] / 2;
	mv[1] = MV[1] / 2;
	GetLoc(mv, &f0off2, &f0off3, &f0off4, parameter->chromWidth);

	mvx = MV[0] / 4;
	mvy = MV[1] / 4;
	offset = mvy * parameter->chromWidth + mvx;
	constU2 = constU;
	constV2 = constV;
	frameIndexU = (BYTE *) (predictU + offset);
	frameIndexV = (BYTE *) (predictV + offset);
	currentIndexU = currentU;
	currentIndexV = currentV;
	diffIndexU = diffU;
	diffIndexV = diffV;

	for (index1 = 0; index1 < kBHeight; index1++,
			constU2 += parameter->chromWidthB, 
			constV2 += parameter->chromWidthB,
					frameIndexU += parameter->chromWidthB, 
					frameIndexV += parameter->chromWidthB)
	{
		for (index2 = 0; index2 < kBWidth; index2++, constU2++, 
			constV2++, frameIndexU++, frameIndexV++, 
				currentIndexU++, currentIndexV++, 
					diffIndexU++, diffIndexV++)
		{
			/*      add 2 for integer rounding      */

			*constU2 = ((int) *frameIndexU +
					(int) *(frameIndexU + f0off2) +
					(int) *(frameIndexU + f0off3) +
					(int) *(frameIndexU + f0off4) + 2) / 4;
			*constV2 = ((int) *frameIndexV +
					(int) *(frameIndexV + f0off2) +
					(int) *(frameIndexV + f0off3) +
					(int) *(frameIndexV + f0off4) + 2) / 4;

			*diffIndexU = *currentIndexU - *constU2;
			*diffIndexV = *currentIndexV - *constV2;

		}
	}
}


/* Reconstructed interpolated (bi-directional) macroblock */
void    FrameIConstr(
	BYTE    *predictY,
	BYTE    *predictU,
	BYTE    *predictV,
	BYTE    *predictYB,
	BYTE    *predictUB,
	BYTE    *predictVB,
	BYTE    *constY,
	BYTE    *constU,
	BYTE    *constV,
	short   MV[],
	short   MVB[], 
	BYTE    *current,
	BYTE    *currentU,
	BYTE    *currentV,
	int     *diff,
	int     *diffU,
	int     *diffV,
	tParameter      *parameter)

{
	int     offset, mvx, mvy, index1, index2;
	BYTE    *constY2, *constU2, *constV2;
	BYTE    *frameIndex, *frameIndexU, *frameIndexV;
	BYTE    *frameIndexB, *frameIndexUB, *frameIndexVB;

	BYTE    *currentIndex, *currentIndexU, *currentIndexV;
	int     *diffIndex, *diffIndexU, *diffIndexV;

	short   mv[2];
	short   f0off2, f0off3, f0off4, f0off2B, f0off3B, f0off4B;

	int for_pel, back_pel;

	/*      reconstruct MB  */

	constY2 = constY;
	constU2 = constU;
	constV2 = constV;

	/*      lumin   */

	mvx = MV[0] / 2;
	mvy = MV[1] / 2;
	offset = (mvy * parameter->luminWidth) + mvx;
	frameIndex = (BYTE *) (predictY + offset);

	GetLoc(MV, &f0off2, &f0off3, &f0off4, parameter->luminWidth);

	mvx = MVB[0] / 2;
	mvy = MVB[1] / 2;
	offset = (mvy * parameter->luminWidth) + mvx;
	frameIndexB = (BYTE *) (predictYB + offset);

	GetLoc(MVB, &f0off2B, &f0off3B, &f0off4B, parameter->luminWidth);

	currentIndex = current;
	diffIndex = diff;

	for (index1 = 0; index1 < kMBHeight; index1++, constY2 += 
		parameter->luminWidthMB, frameIndex += parameter->luminWidthMB, 
			frameIndexB += parameter->luminWidthMB)
	{
		for (index2 = 0; index2 < kMBWidth; index2++, constY2++,
				frameIndex++, frameIndexB++, currentIndex++, 
				diffIndex++)
		{

			/*      add 2 for integer rounding      */
			for_pel = ((int) *frameIndex +
				   (int) *(frameIndex + f0off2) +
				   (int) *(frameIndex + f0off3) +
				   (int) *(frameIndex + f0off4) + 2) / 4;

			back_pel = ((int) *frameIndexB +
				    (int) *(frameIndexB + f0off2B) +
				    (int) *(frameIndexB + f0off3B) +
				    (int) *(frameIndexB + f0off4B) + 2) / 4;

			/*      add 1 for integer rounding      */
			*constY2 = (for_pel + back_pel + 1) / 2;

			*diffIndex = *currentIndex - *constY2;

		}
	}

	/*      chrom   */

	mv[0] = MV[0] / 2;
	mv[1] = MV[1] / 2;
	GetLoc(mv, &f0off2, &f0off3, &f0off4, parameter->chromWidth);
	mv[0] = MVB[0] / 2;
	mv[1] = MVB[1] / 2;
	GetLoc(mv, &f0off2B, &f0off3B, &f0off4B, parameter->chromWidth);

	mvx = MV[0] / 4;
	mvy = MV[1] / 4;
	offset = mvy * parameter->chromWidth + mvx;
	frameIndexU = (BYTE *) (predictU + offset);
	frameIndexV = (BYTE *) (predictV + offset);
	mvx = MVB[0] / 4;
	mvy = MVB[1] / 4;
	offset = mvy * parameter->chromWidth + mvx;
	frameIndexUB = (BYTE *) (predictUB + offset);
	frameIndexVB = (BYTE *) (predictVB + offset);

	currentIndexU = currentU;
	currentIndexV = currentV;
	diffIndexU = diffU;
	diffIndexV = diffV;

	for (index1 = 0; index1 < kBHeight; index1++, constU2 += 
		parameter->chromWidthB, constV2 += parameter->chromWidthB, 
		frameIndexU += parameter->chromWidthB, frameIndexV += 
		parameter->chromWidthB, frameIndexUB += parameter->chromWidthB,
		frameIndexVB += parameter->chromWidthB)
	{
		for (index2 = 0; index2 < kBWidth; index2++, constU2++, constV2++,
				frameIndexU++, frameIndexV++, frameIndexUB++, 
					frameIndexVB++, currentIndexU++, 
					currentIndexV++, diffIndexU++, diffIndexV++)

		{
			/*      add 2 for integer rounding      */
			for_pel = ((int) *frameIndexU +
				   (int) *(frameIndexU + f0off2) +
				   (int) *(frameIndexU + f0off3) +
				   (int) *(frameIndexU + f0off4) + 2) / 4;

			back_pel = ((int) *frameIndexUB +
				    (int) *(frameIndexUB + f0off2B) +
				    (int) *(frameIndexUB + f0off3B) +
				    (int) *(frameIndexUB + f0off4B) + 2) / 4;

			/*      add 1 for integer rounding      */
			*constU2 = (for_pel + back_pel + 1) / 2;

			for_pel = ((int) *frameIndexV +
				   (int) *(frameIndexV + f0off2) +
				   (int) *(frameIndexV + f0off3) +
				   (int) *(frameIndexV + f0off4) + 2) / 4;

			back_pel = ((int) *frameIndexVB +
				    (int) *(frameIndexVB + f0off2B) +
				    (int) *(frameIndexVB + f0off3B) +
				    (int) *(frameIndexVB + f0off4B) + 2) / 4;

			*constV2 = (for_pel + back_pel + 1) / 2;

			*diffIndexU = *currentIndexU - *constU2;
			*diffIndexV = *currentIndexV - *constV2;

		}
	}
}


/* convert motion vector into 1-D pointer picture buffer co-ordinates */
void    GetLoc(
	short   MV[],
	short   *f0off2,
	short   *f0off3,
	short   *f0off4,
	short   width)
{
	/*      frame   */

	if (MV[0] >= 0)
	{
		if (MV[1] >= 0)
		{
			switch (((MV[0] % 2) << 1) + (MV[1] % 2))
			{
				case 0:         /*      no half pel     */

					*f0off2 = 0;
					*f0off3 = 0;
					*f0off4 = 0;
					break;

				case 1:         /*      y half pel      */

					*f0off2 = 0;
					*f0off3 = width;
					*f0off4 = *f0off3;
					break;

				case 2:         /*      x half pel      */

					*f0off2 = 1;
					*f0off3 = 0;
					*f0off4 = *f0off2;
					break;

				case 3:         /*      x, y half pel   */

					*f0off2 = 1;
					*f0off3 = width;
					*f0off4 = *f0off3 + *f0off2;
					break;
			}
		}
		else    /*      MV[1] < 0       */
		{
			switch (((MV[0] % 2) << 1) + (-MV[1] % 2))
			{
				case 0:         /*      no half pel     */

					*f0off2 = 0;
					*f0off3 = 0;
					*f0off4 = 0;
					break;

				case 1:         /*      y half pel      */

					*f0off2 = 0;
					*f0off3 = -width;
					*f0off4 = *f0off3;
					break;

				case 2:         /*      x half pel      */

					*f0off2 = 1;
					*f0off3 = 0;
					*f0off4 = *f0off2;
					break;

				case 3:         /*      x, y half pel   */

					*f0off2 = 1;
					*f0off3 = -width;
					*f0off4 = *f0off3 + *f0off2;
					break;
			}
		}
	}
	else    /*      MV[0] < 0       */
	{
		if (MV[1] >= 0)
		{
			switch (((-MV[0] % 2) << 1) + (MV[1] % 2))
			{
				case 0:         /*      no half pel     */

					*f0off2 = 0;
					*f0off3 = 0;
					*f0off4 = 0;
					break;

				case 1:         /*      y half pel      */

					*f0off2 = 0;
					*f0off3 = width;
					*f0off4 = *f0off3;
					break;

				case 2:         /*      x half pel      */

					*f0off2 = -1;
					*f0off3 = 0;
					*f0off4 = *f0off2;
					break;

				case 3:         /*      x, y half pel   */

					*f0off2 = -1;
					*f0off3 = width;
					*f0off4 = *f0off3 + *f0off2;
					break;
			}
		}
		else    /*      MV[1] < 0       */
		{
			switch (((-MV[0] % 2) << 1) + (-MV[1] % 2))
			{
				case 0:         /*      no half pel     */

					*f0off2 = 0;
					*f0off3 = 0;
					*f0off4 = 0;
					break;

				case 1:         /*      y half pel      */

					*f0off2 = 0;
					*f0off3 = -width;
					*f0off4 = *f0off3;
					break;

				case 2:         /*      x half pel      */

					*f0off2 = -1;
					*f0off3 = 0;
					*f0off4 = *f0off2;
					break;

				case 3:         /*      x, y half pel   */

					*f0off2 = -1;
					*f0off3 = -width;
					*f0off4 = *f0off3 + *f0off2;
					break;
			}
		}
	}
}
