/* File: stats.c */

/*
 *      $Log:   stats.c,v $
 * Revision 1.1.1.1  93/03/29  11:27:56  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:33:17  au
 * Initial revision
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "consts.h"
#include "mpeg1.h"
#include "encoder.pro"

void    statistic(
	int             pictureNum,
	short           pictureType,
	tParameter      *parameter,
	BYTE            *original,
	BYTE            *originalU,
	BYTE            *originalV,
	BYTE            *reconst,
	BYTE            *reconstU,
	BYTE            *reconstV,
	int             numLuminPixels,
	int             numChromPixels)
{
	FILE            *statFptr, *fopen();
	double          Ysnr, Usnr, Vsnr;

	/*      calculates signal to noise ratio        */

	if ((statFptr = fopen (parameter->statusFName, "a")) == NULL)
	{
		printf ("error opening file '%s', terminating ...\n",
				parameter->statusFName);
		exit(-1);
	}

	snr(reconst, original, &Ysnr, reconstU, originalU, &Usnr,
			reconstV, originalV, &Vsnr , numLuminPixels, numChromPixels);

	fprintf(statFptr, "pictureNum = %d\tpictureType = %s\t", pictureNum,
			(pictureType == kIPicture) ? "I picture" :
			(pictureType == kPPicture) ? "P picture" : "B picture");
	fprintf(statFptr, "\tYsnr = %.2f\tUsnr = %.2f\tVsnr = %.2f\n",
			Ysnr, Usnr, Vsnr);

	fclose(statFptr);
}


void    snr(
	BYTE    *ycod,
	BYTE    *ysrc,
	double  *ysnr,
	BYTE    *ucod,
	BYTE    *usrc,
	double  *usnr,
	BYTE    *vcod,
	BYTE    *vsrc,
	double  *vsnr,
	int     numLuminPixels,
	int     numChromPixels)
{
	int     dif, i;
	double  acc = 0.0;

	for (i = 0; i < numLuminPixels; i++)
	{
		dif = ycod[i] - ysrc[i];
		acc += dif * dif;
	}

	acc /= numLuminPixels;
	*ysnr = 20 * log10(255 / sqrt(acc));

	acc = 0.0;

	for (i = 0; i < numChromPixels; i++)
	{
		dif = ucod[i] - usrc[i];
		acc += dif * dif;
	}

	acc /= numChromPixels;
	*usnr = 20 * log10(255 / sqrt(acc));

	acc = 0.0;

	for (i = 0; i < numChromPixels; i++)
	{
		dif = vcod[i] - vsrc[i];
		acc += dif * dif;
	}

	acc /= numChromPixels;
	*vsnr = 20 * log10(255 / sqrt(acc));
}
