/* File: transfer.c */

/*
 *      $Log:   transfer.c,v $
 * Revision 1.1.1.1  93/03/29  11:27:56  oosa
 * MPEG1 encoder/decoder initial revision
 * 
 * Revision 1.1  1993/03/10  20:33:17  au
 * Initial revision
 *
 */

#include <stdio.h>
#include "consts.h"
#include "mpeg1.h"
#include "encoder.pro"

/* copy one macroblock from source to destination */
void    CopyToBuf(
	int     bufIndex,
	int     bufIndexC,
	BYTE    bufY[],
	BYTE    bufU[],
	BYTE    bufV[],
	BYTE    *currentMB,
	BYTE    *currentMBU,
	BYTE    *currentMBV,
	int     *currentMBDiff,
	int     *currentMBDiffU,
	int     *currentMBDiffV,
	tParameter      *parameter)
{
	int     row, col, currentIndex;

	/*      lumin   */

	for (row = 0, currentIndex = 0; row < kMBHeight;
			row++, bufIndex += parameter->luminWidthMB)
	{
		for (col = 0; col < kMBWidth; col++, currentIndex++, bufIndex++)
		{
			currentMB[currentIndex] = bufY[bufIndex];
			currentMBDiff[currentIndex] = bufY[bufIndex];
		}
	}

	/*      chrom   */

	for (row = 0, currentIndex = 0; row < kBHeight;
			row++, bufIndexC += parameter->chromWidthB)
	{
		for (col = 0; col < kBWidth; col++, currentIndex++, bufIndexC++)
		{
			currentMBU[currentIndex] = bufU[bufIndexC];
			currentMBV[currentIndex] = bufV[bufIndexC];
			currentMBDiffU[currentIndex] = bufU[bufIndexC];
			currentMBDiffV[currentIndex] = bufV[bufIndexC];
		}
	}
}


/* copy entire picture */ 

void    CopyFromBuf(
	short           pictureType,
	int             mbType,
	short           codedBPattern,
	int             mbRow,
	int             mbCol,
	BYTE            *bufYR,
	BYTE            *bufUR,
	BYTE            *bufVR,
	int             currentMBDiff[],
	int             currentMBDiffU[],
	int             currentMBDiffV[],
	int             mbIndex,
	tParameter      *parameter)
{
	int             bufIndex, bufIndex1, currentIndex, temp;
	short           code, line, pix, row, col;

	if (mbType & kIntraType)        /*      intra   */
	{
		bufIndex = mbRow * kMBHeight * parameter->luminWidth + mbCol * kMBWidth;

		/*      lumin   */

		for (row = 0, currentIndex = 0; row < kMBHeight; row++,
				bufIndex += parameter->luminWidthMB)
		{
			for (col = 0; col < kMBWidth; col++, currentIndex++, bufIndex++)
			{
				bufYR[bufIndex] = (currentMBDiff[currentIndex] >= 255) ? 255 :
						((currentMBDiff[currentIndex] <= 0) ? 0 :
						currentMBDiff[currentIndex]);
			}
		}

		/*      chrom   */

		bufIndex = mbRow * kBHeight * parameter->chromWidth + mbCol * kBWidth;

		for (row = 0, currentIndex = 0; row < kBHeight; row++,
				bufIndex += parameter->chromWidthB)
		{
			for (col = 0; col < kBWidth; col++, currentIndex++, bufIndex++)
			{
				bufUR[bufIndex] = (currentMBDiffU[currentIndex] >= 255) ? 255 :
						((currentMBDiffU[currentIndex] <= 0) ? 0 :
						currentMBDiffU[currentIndex]);
				bufVR[bufIndex] = (currentMBDiffV[currentIndex] >= 255) ? 255 :
						((currentMBDiffV[currentIndex] <= 0) ? 0 :
						currentMBDiffV[currentIndex]);
			}
		}
	}
	else    /*      inter   */
	{
		/*      lumin   */

		code = 0x20;

		for (line = 0; line < kMBHeight; line += kDCTSize)
		{
			bufIndex1 = (mbRow * kMBHeight + line) * parameter->luminWidth +
					mbCol * kMBWidth;

			for (pix = 0; pix < kMBWidth; pix += kDCTSize)
			{
				if (! (codedBPattern & code))
				{
					code >>= 1;
					continue;
				}

				bufIndex = bufIndex1 + pix;

				code >>= 1;

				currentIndex = line * kMBWidth + pix;

				for (row = 0; row < kDCTSize;
						row++, currentIndex += kMBWidthDCT,
						bufIndex += parameter->luminWidthDCT)
				{
					for (col = 0; col < kDCTSize; col++, currentIndex++,
							bufIndex++)
					{
						temp = bufYR[bufIndex] + currentMBDiff[currentIndex];
						bufYR[bufIndex] =
								(temp >= 255) ? 255 : ((temp <= 0) ? 0 : temp);
					}
				}
			}
		}

		/*      chrom - U       */

		if (codedBPattern & 0x02)
		{
			bufIndex = mbRow * kBHeight * parameter->chromWidth + mbCol * kBWidth;
/* ? */
			for (row = 0, currentIndex = 0; row < kBHeight; row++,
					bufIndex += parameter->chromWidthDCT)
			{
				for (col = 0; col < kBWidth; col++, currentIndex++, bufIndex++)
				{
					temp = bufUR[bufIndex] + currentMBDiffU[currentIndex];
					bufUR[bufIndex] =
							(temp >= 255) ? 255 : ((temp <= 0) ? 0 : temp);
				}
			}
		}

		/*      chrom - V       */

		if (codedBPattern & 0x01)
		{
			bufIndex = mbRow * kBHeight * parameter->chromWidth + mbCol * kBWidth;

			for (row = 0, currentIndex = 0; row < kBHeight; row++,
					bufIndex += parameter->chromWidthDCT)
			{
				for (col = 0; col < kBWidth; col++, currentIndex++, bufIndex++)
				{
					temp = bufVR[bufIndex] + currentMBDiffV[currentIndex];
					bufVR[bufIndex] =
							(temp >= 255) ? 255 : ((temp <= 0) ? 0 : temp);
				}
			}
		}
	}
}
